@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2012-2014 Red Hat.
+ * Copyright (c) 2012-2015 Red Hat.
  * Copyright (c) 2000,2004,2005 Silicon Graphics, Inc.  All Rights Reserved.
  * 
  * This library is free software; you can redistribute it and/or modify it
@@ -15,12 +15,35 @@
 
 #include "pmapi.h"
 #include "impl.h"
+#ifdef HAVE_NET_IF_H
 #include <net/if.h>
+#endif
+#ifdef HAVE_IPHLPAPI_H
+#include <iphlpapi.h>
+#endif
 #define SOCKET_INTERNAL
 #include "internal.h"
 
 /* default connect timeout is 5 seconds */
-static struct timeval	canwait = { 5, 000000 };
+static struct timeval	conn_wait = { 5, 0 };
+static int		conn_wait_done;
+
+#ifdef PM_MULTI_THREAD
+static pthread_mutex_t	auxconnect_lock = PTHREAD_MUTEX_INITIALIZER;
+#else
+void			*auxconnect_lock;
+#endif
+
+#if defined(PM_MULTI_THREAD) && defined(PM_MULTI_THREAD_DEBUG)
+/*
+ * return true if lock == auxconnect_lock
+ */
+int
+__pmIsAuxconnectLock(void *lock)
+{
+    return lock == (void *)&auxconnect_lock;
+}
+#endif
 
 __pmHostEnt *
 __pmHostEntAlloc(void)
@@ -78,7 +101,13 @@ __pmSockAddrInit(__pmSockAddr *addr, int family, int address, int port)
 	addr->sockaddr.ipv6.sin6_family = family;
 	addr->sockaddr.ipv6.sin6_port = htons(port);
 	if (address == INADDR_LOOPBACK)
-	    addr->sockaddr.ipv6.sin6_addr.s6_addr[15] = 1;
+	    addr->sockaddr.ipv6.sin6_addr = in6addr_loopback;
+	else
+	    addr->sockaddr.ipv6.sin6_addr = in6addr_any;
+    }
+    else if (family == AF_UNSPEC) {
+	/* do nothing */
+	;
     }
     else
 	__pmNotifyErr(LOG_ERR,
@@ -117,6 +146,10 @@ __pmSockAddrGetPort(const __pmSockAddr *addr)
         return ntohs(addr->sockaddr.inet.sin_port);
     if (addr->sockaddr.raw.sa_family == AF_INET6)
         return ntohs(addr->sockaddr.ipv6.sin6_port);
+#if defined(HAVE_STRUCT_SOCKADDR_UN)
+    if (addr->sockaddr.raw.sa_family == AF_UNIX)
+	return 0;
+#endif
     __pmNotifyErr(LOG_ERR,
 		  "__pmSockAddrGetPort: Invalid address family: %d\n",
 		  addr->sockaddr.raw.sa_family);
@@ -150,12 +183,15 @@ void
 __pmSockAddrSetPath(__pmSockAddr *addr, const char *path)
 {
 #if defined(HAVE_STRUCT_SOCKADDR_UN)
-    if (addr->sockaddr.raw.sa_family == AF_UNIX)
-	strncpy(addr->sockaddr.local.sun_path, path, sizeof(addr->sockaddr.local.sun_path));
-    else
+    if (addr->sockaddr.raw.sa_family == AF_UNIX) {
+	int buflen = sizeof(addr->sockaddr.local.sun_path);
+	strncpy(addr->sockaddr.local.sun_path, path, buflen);
+	addr->sockaddr.local.sun_path[buflen-1] = '\0';
+    } else {
 	__pmNotifyErr(LOG_ERR,
 		"%s:__pmSockAddrSetPath: Invalid address family: %d\n",
 		__FILE__, addr->sockaddr.raw.sa_family);
+    }
 #else
     __pmNotifyErr(LOG_ERR, "%s:__pmSockAddrSetPath: AF_UNIX is not supported\n", __FILE__);
 #endif
@@ -215,6 +251,11 @@ __pmSockAddrCompare(const __pmSockAddr *addr1, const __pmSockAddr *addr2)
 		       sizeof(addr1->sockaddr.local.sun_path));
     }
 #endif
+    
+    if (addr1->sockaddr.raw.sa_family == AF_UNSPEC) {
+	/* if the address family is unspecified, assume not equal */
+	return 1;
+    }
 
     /* Unknown address family. */
     __pmNotifyErr(LOG_ERR,
@@ -254,6 +295,11 @@ __pmStringToSockAddr(const char *cp)
 	    /* Set the address family to 0, meaning "not set". */
 	    addr->sockaddr.raw.sa_family = 0;
 	}
+        else if (strcmp(cp, "INADDR_LOOPBACK") == 0) {
+	    addr->sockaddr.inet.sin_addr.s_addr = htonl(INADDR_LOOPBACK);
+	    /* Set the address family to 0, meaning "not set". */
+	    addr->sockaddr.raw.sa_family = 0;
+	}
 	else {
 	    int sts;
 	    /* Determine the address family. */
@@ -455,8 +501,8 @@ __pmInitSocket(int fd, int family)
     if (__pmSetSockOpt(fd, SOL_SOCKET, SO_LINGER, (char *)&nolinger,
 		   (__pmSockLen)sizeof(nolinger)) < 0) {
 	__pmNotifyErr(LOG_WARNING,
-		      "%s:__pmCreateSocket(%d): __pmSetSockOpt SO_LINGER: %s\n",
-		      __FILE__, fd, netstrerror_r(errmsg, sizeof(errmsg)));
+		      "%s:__pmInitSocket(fd=%d, family=%d): __pmSetSockOpt SO_LINGER: %s\n",
+		      __FILE__, fd, family, netstrerror_r(errmsg, sizeof(errmsg)));
     }
 
 #if defined(HAVE_STRUCT_SOCKADDR_UN)
@@ -468,8 +514,8 @@ __pmInitSocket(int fd, int family)
     if (__pmSetSockOpt(fd, IPPROTO_TCP, TCP_NODELAY, (char *)&nodelay,
 		       (__pmSockLen)sizeof(nodelay)) < 0) {
 	__pmNotifyErr(LOG_WARNING,
-		      "%s:__pmCreateSocket(%d): __pmSetSockOpt TCP_NODELAY: %s\n",
-		      __FILE__, fd, netstrerror_r(errmsg, sizeof(errmsg)));
+		      "%s:__pmInitSocket(fd=%d, family=%d): __pmSetSockOpt TCP_NODELAY: %s\n",
+		      __FILE__, fd, family, netstrerror_r(errmsg, sizeof(errmsg)));
     }
 
     return fd;
@@ -502,7 +548,7 @@ __pmCreateIPv6Socket(void)
      * kernels.  Setting then testing is the most reliable way we've found.
      */
     on = 1;
-    __pmSetSockOpt(fd, IPPROTO_IPV6, IPV6_V6ONLY, &on, sizeof(on));
+    (void)__pmSetSockOpt(fd, IPPROTO_IPV6, IPV6_V6ONLY, &on, sizeof(on));
     on = 0;
     sts = __pmGetSockOpt(fd, IPPROTO_IPV6, IPV6_V6ONLY, (void *)&on, &onlen);
     if (sts < 0 || on != 1) {
@@ -581,9 +627,11 @@ __pmBind(int fd, void *addr, __pmSockLen addrlen)
 
 #ifdef PCP_DEBUG
     if ((pmDebug & DBG_TRACE_CONTEXT) && (pmDebug & DBG_TRACE_DESPERATE)) {
+	char *sockname = __pmSockAddrToString(sock);
 	fprintf(stderr, "%s:__pmBind(fd=%d, family=%d, port=%d, addr=%s)\n",
 	    __FILE__, fd, __pmSockAddrGetFamily(sock), __pmSockAddrGetPort(sock),
-	    __pmSockAddrToString(sock));
+	    sockname);
+	free(sockname);
     }
 #endif
     if (sock->sockaddr.raw.sa_family == AF_INET)
@@ -697,36 +745,59 @@ __pmConnectRestoreFlags(int fd, int fdFlags)
     return fd;
 }
 
-const struct timeval *
+int
+__pmSetConnectTimeout(double timeout)
+{
+    if (timeout < 0.0)
+	return -EINVAL;
+
+    PM_LOCK(auxconnect_lock);
+    /* THREADSAFE - no locks acquired in __pmtimevalFromReal() */
+    __pmtimevalFromReal(timeout, &conn_wait);
+    conn_wait_done = 1;
+    PM_UNLOCK(auxconnect_lock);
+    return 0;
+}
+
+double
 __pmConnectTimeout(void)
 {
-    static int		first_time = 1;
+    char	*timeout_str, *end_ptr;
+    double	timeout;
 
-    /*
-     * get optional stuff from environment ...
-     * 	PMCD_CONNECT_TIMEOUT
-     *	PMCD_PORT
-     */
-    PM_INIT_LOCKS();
-    PM_LOCK(__pmLock_libpcp);
-    if (first_time) {
-	char	*env_str;
-	first_time = 0;
-
-	if ((env_str = getenv("PMCD_CONNECT_TIMEOUT")) != NULL) {
-	    char	*end_ptr;
-	    double	timeout = strtod(env_str, &end_ptr);
-	    if (*end_ptr != '\0' || timeout < 0.0)
+    /* get optional PMCD connection timeout from the environment */
+    PM_LOCK(auxconnect_lock);
+    if (!conn_wait_done) {
+	conn_wait_done = 1;
+	PM_LOCK(__pmLock_extcall);
+	timeout_str = getenv("PMCD_CONNECT_TIMEOUT");		/* THREADSAFE */
+	if (timeout_str != NULL) {
+	    timeout = strtod(timeout_str, &end_ptr);
+	    if (*end_ptr != '\0' || timeout < 0.0) {
+		char	*fromenv = strdup(timeout_str);
+		PM_UNLOCK(__pmLock_extcall);
+		PM_UNLOCK(auxconnect_lock);
 		__pmNotifyErr(LOG_WARNING, "%s:__pmAuxConnectPMCDPort: "
 			      "ignored bad PMCD_CONNECT_TIMEOUT = '%s'\n",
-			      __FILE__, env_str);
+			      __FILE__, fromenv);
+		PM_LOCK(auxconnect_lock);
+		free(fromenv);
+	    }
 	    else {
-		__pmtimevalFromReal(timeout, &canwait);
+		PM_UNLOCK(__pmLock_extcall);
+		/* THREADSAFE - no locks acquired in __pmtimevalFromReal() */
+		__pmtimevalFromReal(timeout, &conn_wait);
 	    }
 	}
+	else
+	    PM_UNLOCK(__pmLock_extcall);
     }
-    PM_UNLOCK(__pmLock_libpcp);
-    return (&canwait);
+
+    /* THREADSAFE - no locks acquired in __pmtimevalToReal() */
+    timeout = __pmtimevalToReal(&conn_wait);
+    PM_UNLOCK(auxconnect_lock);
+
+    return timeout;
 }
  
 int
@@ -791,11 +862,15 @@ __pmAuxConnectPMCD(const char *hostname)
 {
     static int		*pmcd_ports = NULL;
 
-    PM_INIT_LOCKS();
-    PM_LOCK(__pmLock_libpcp);
-    if (pmcd_ports == NULL)
+    PM_LOCK(auxconnect_lock);
+    if (pmcd_ports == NULL) {
+	/*
+	 * THREADSAFE - only __pmLock_extcall acquired during call to
+	 * __pmPMCDAddPorts()
+	 */
 	__pmPMCDAddPorts(&pmcd_ports, 0);
-    PM_UNLOCK(__pmLock_libpcp);
+    }
+    PM_UNLOCK(auxconnect_lock);
 
     /* __pmPMCDAddPorts discovers at least one valid port, if it returns. */
     return __pmAuxConnectPMCDPort(hostname, pmcd_ports[0]);
@@ -806,23 +881,43 @@ __pmAuxConnectPMCDPort(const char *hostname, int pmcd_port)
 {
     __pmSockAddr	*myAddr;
     __pmHostEnt		*servInfo;
-    int			fd = -1;	/* Fd for socket connection to pmcd */
+    int			fd;
     int			sts;
-    int			fdFlags = 0;
+    int			fdFlags[FD_SETSIZE];
+    __pmFdSet		allFds;
+    __pmFdSet		readyFds;
+    int			maxFd;
     void		*enumIx;
+    struct timeval	stv;
+    struct timeval	*pstv;
+    int			rc;
+    int			i;
 
     if ((servInfo = __pmGetAddrInfo(hostname)) == NULL) {
 #ifdef PCP_DEBUG
 	if (pmDebug & DBG_TRACE_CONTEXT) {
+	    const char	*errmsg;
+	    PM_LOCK(__pmLock_extcall);
+	    errmsg = hoststrerror();		/* THREADSAFE */
 	    fprintf(stderr, "%s:__pmAuxConnectPMCDPort(%s, %d) : hosterror=%d, ``%s''\n",
-		    __FILE__, hostname, pmcd_port, hosterror(), hoststrerror());
+		    __FILE__, hostname, pmcd_port, hosterror(), errmsg);
+	    PM_UNLOCK(__pmLock_extcall);
 	}
 #endif
 	return -EHOSTUNREACH;
     }
 
+    /*
+     * We want to respect the connect timeout that has been configured, but we
+     * may have more than one address to try. Do this by creating a socket for
+     * each address and then using __pmSelectWrite() to wait for one of them to
+     * respond. That way, the timeout is applied to all of the addresses
+     * simultaneously. First, create the sockets, add them to the fd set
+     * and try to connect.
+     */
     __pmConnectTimeout();
-
+    __pmFD_ZERO(&allFds);
+    maxFd = -1;
     enumIx = NULL;
     for (myAddr = __pmHostEntGetSockAddr(servInfo, &enumIx);
 	 myAddr != NULL;
@@ -844,46 +939,71 @@ __pmAuxConnectPMCDPort(const char *hostname, int pmcd_port)
 	}
 
 	/* Attempt to connect */
-	fdFlags = __pmConnectTo(fd, myAddr, pmcd_port);
+	fdFlags[fd] = __pmConnectTo(fd, myAddr, pmcd_port);
 	__pmSockAddrFree(myAddr);
-	if (fdFlags < 0) {
+	if (fdFlags[fd] < 0) {
 	    /*
 	     * Mark failure in case we fall out the end of the loop
 	     * and try next address
 	     */
-	    fd = -ECONNREFUSED;
+	    __pmCloseSocket(fd);
 	    continue;
 	}
 
-	/* FNDELAY and we're in progress - wait on select */
-	struct timeval stv = canwait;
-	struct timeval *pstv = (stv.tv_sec || stv.tv_usec) ? &stv : NULL;
-	__pmFdSet wfds;
-	int rc;
+	/* Add it to the fd set. */
+	__pmFD_SET(fd, &allFds);
+	if (fd > maxFd)
+	    maxFd = fd;
+    }
+    __pmHostEntFree(servInfo);
 
-	__pmFD_ZERO(&wfds);
-	__pmFD_SET(fd, &wfds);
-	sts = 0;
-	if ((rc = __pmSelectWrite(fd+1, &wfds, pstv)) == 1) {
-	    sts = __pmConnectCheckError(fd);
-	}
-	else if (rc == 0) {
-	    sts = ETIMEDOUT;
-	}
-	else {
-	    sts = (rc < 0) ? neterror() : EINVAL;
-	}
- 
-	/* Successful connection? */
-	if (sts == 0)
-	    break;
+    /* If we were unable to open any sockets, then give up. */
+    if (maxFd == -1)
+	return -ECONNREFUSED;
 
-	/* Unsuccessful connection. */
-	__pmCloseSocket(fd);
-	fd = -sts;
+    /* FNDELAY and we're in progress - wait on select */
+    __pmFD_COPY(&readyFds, &allFds);
+    /*
+     * THREADSAFE - conn_wait only changed in one-trip code in
+     * __pmConnectTimeout()
+     */
+    stv = conn_wait;
+    pstv = (stv.tv_sec || stv.tv_usec) ? &stv : NULL;
+    rc = __pmSelectWrite(maxFd+1, &readyFds, pstv);
+
+    /* Figure out what happened. */
+    if (rc == 0)
+	fd = -ETIMEDOUT;
+    else if (rc < 0)
+	fd = -neterror();
+    else {
+	/*
+	 * Examine the fd set and choose the first successfully connected
+	 * socket, if any.
+	 * Note that because rc > 0, at least one fd is ready and 'fd' will
+	 * definitely get set. However, initialize it to keep coverity happy.
+	 */
+	fd = -EINVAL;
+	for (i = 0; i <= maxFd; ++i) {
+	    if (__pmFD_ISSET(i, &readyFds)) {
+		/* Successful connection? */
+		sts = __pmConnectCheckError(i);
+		if (sts == 0) {
+		    fd = i;
+		    break;
+		}
+		fd = -sts;
+	    }
+	}
     }
 
-    __pmHostEntFree(servInfo);
+    /* Clean up the unused fds. */
+    for (i = 0; i <= maxFd; ++i) {
+	if (i != fd && __pmFD_ISSET(i, &allFds))
+	    __pmCloseSocket(i);
+    }
+
+    /* Unsuccessful? */
     if (fd < 0)
         return fd;
 
@@ -892,7 +1012,7 @@ __pmAuxConnectPMCDPort(const char *hostname, int pmcd_port)
      * flags and make sure this file descriptor is closed if exec() is
      * called
      */
-    return __pmConnectRestoreFlags(fd, fdFlags);
+    return __pmConnectRestoreFlags(fd, fdFlags[fd]);
 }
 
 /*
@@ -905,19 +1025,23 @@ __pmAuxConnectPMCDPort(const char *hostname, int pmcd_port)
 const char *
 __pmPMCDLocalSocketDefault(void)
 {
-    static char pmcd_socket[MAXPATHLEN];
+    static char 		pmcd_socket[MAXPATHLEN];
 
-    PM_INIT_LOCKS();
-    PM_LOCK(__pmLock_libpcp);
+    PM_LOCK(auxconnect_lock);
     if (pmcd_socket[0] == '\0') {
 	char *envstr;
-	if ((envstr = getenv("PMCD_SOCKET")) != NULL)
+	PM_LOCK(__pmLock_extcall);
+	if ((envstr = getenv("PMCD_SOCKET")) != NULL) {		/* THREADSAFE */
 	    snprintf(pmcd_socket, sizeof(pmcd_socket), "%s", envstr);
-	else
+	    PM_UNLOCK(__pmLock_extcall);
+	}
+	else {
+	    PM_UNLOCK(__pmLock_extcall);
 	    snprintf(pmcd_socket, sizeof(pmcd_socket), "%s%c" "pmcd.socket",
 		     pmGetConfig("PCP_RUN_DIR"), __pmPathSeparator());
+	}
     }
-    PM_UNLOCK(__pmLock_libpcp);
+    PM_UNLOCK(auxconnect_lock);
 
     return pmcd_socket;
 }
@@ -961,11 +1085,18 @@ __pmAuxConnectPMCDUnixSocket(const char *sock_path)
     /* Attempt to connect */
     fdFlags = __pmConnectTo(fd, myAddr, -1);
     __pmSockAddrFree(myAddr);
-    if (fdFlags < 0)
+    if (fdFlags < 0) {
+	__pmCloseSocket(fd);
 	return -ECONNREFUSED;
+    }
 
-    /* FNDELAY and we're in progress - wait on select */
-    stv = canwait;
+    /*
+     * FNDELAY and we're in progress - wait on select
+     *
+     * THREADSAFE - conn_wait only changed in one-trip code in
+     * __pmConnectTimeout()
+     */
+    stv = conn_wait;
     pstv = (stv.tv_sec || stv.tv_usec) ? &stv : NULL;
     __pmFD_ZERO(&wfds);
     __pmFD_SET(fd, &wfds);
@@ -1020,8 +1151,11 @@ __pmHostEntGetName(__pmHostEnt *he)
 	     addr = __pmHostEntGetSockAddr(he, &enumIx)) {
 	    he->name = __pmGetNameInfo(addr);
 #ifdef PCP_DEBUG
-	    if (pmDebug & DBG_TRACE_DESPERATE)
-		fprintf(stderr, "%s:__pmHostEntGetName: __pmGetNameInfo(%s) returns %s\n", __FILE__, __pmSockAddrToString(addr), he->name);
+	    if (pmDebug & DBG_TRACE_DESPERATE) {
+		char *sockname = __pmSockAddrToString(addr);
+		fprintf(stderr, "%s:__pmHostEntGetName: __pmGetNameInfo(%s) returns %s\n", __FILE__, sockname, he->name);
+		free(sockname);
+	    }
 #endif
 	    __pmSockAddrFree(addr);
 	    if (he->name != NULL)
@@ -1119,8 +1253,14 @@ __pmGetAddrInfo(const char *hostName)
 	hints.ai_flags = AI_ADDRCONFIG; /* Only return configured address types */
 #endif
 
+	PM_LOCK(__pmLock_extcall);
 	sts = getaddrinfo(hostName, NULL, &hints, &hostEntry->addresses);
+	PM_UNLOCK(__pmLock_extcall);
 	if (sts != 0) {
+#ifdef PCP_DEBUG
+	    if (pmDebug & DBG_TRACE_DESPERATE)
+		fprintf(stderr, "%s:__pmGetAddrInfo: getaddrinfo(%s, ...) -> %d %s\n", __FILE__, hostName, sts, gai_strerror(sts));
+#endif
 	    __pmHostEntFree(hostEntry);
 	    hostEntry = NULL;
 	}
@@ -1298,12 +1438,15 @@ __pmSecureClientHandshake(int fd, int flags, const char *hostname, __pmHashCtl *
      *
      * But we can still talk to a pmcd that requires credentials, provided
      * we are using unix domain sockets (the kernel provides the auth info
-     * to pmcd in this case, with no other special sauce required).
+     * to pmcd in this case, with no other special sauce required).  We're
+     * also able to communicate container attributes.
      */
-    if (flags == PDU_FLAG_CREDS_REQD)
-	if (__pmHashSearch(PCP_ATTR_UNIXSOCK, attrs) != NULL)
-	    return 0;
-    return -EOPNOTSUPP;
+    if (flags & PDU_FLAG_CREDS_REQD)
+	if (__pmHashSearch(PCP_ATTR_UNIXSOCK, attrs) == NULL)
+	    return -EOPNOTSUPP;
+    if (flags & (PDU_FLAG_SECURE|PDU_FLAG_AUTH|PDU_FLAG_COMPRESS))
+	return -EOPNOTSUPP;
+    return 0;
 }
 
 int
