@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2013-2015 Red Hat.
+ * Copyright (c) 2013-2016 Red Hat.
  * 
  * This library is free software; you can redistribute it and/or modify it
  * under the terms of the GNU Lesser General Public License as published
@@ -105,8 +105,14 @@ __pmPMCDAddPorts(int **ports, int nports)
     char *env;
     int  new_nports = nports;
 
-    if ((env = getenv("PMCD_PORT")) != NULL)
-	new_nports = __pmAddPorts(env, ports, nports);
+    PM_LOCK(__pmLock_extcall);
+    if ((env = getenv("PMCD_PORT")) != NULL)		/* THREADSAFE */
+	/*
+	 * THREADSAFE because __pmAddPorts acquires no locks (other than
+	 * on the fatal __pmNoMem() path)
+	 */
+	new_nports = __pmAddPorts(env, ports, nports);		/* THREADSAFE */
+    PM_UNLOCK(__pmLock_extcall);
 
     /*
      * Add the default port, if no new ports were added or if there was an
@@ -133,8 +139,14 @@ __pmProxyAddPorts(int **ports, int nports)
     char *env;
     int  new_nports = nports;
 
-    if ((env = getenv("PMPROXY_PORT")) != NULL)
-	new_nports = __pmAddPorts(env, ports, nports);
+    PM_LOCK(__pmLock_extcall);
+    if ((env = getenv("PMPROXY_PORT")) != NULL)		/* THREADSAFE */
+	/*
+	 * THREADSAFE because __pmAddPorts acquires no locks (other than
+	 * on the fatal __pmNoMem() path)
+	 */
+	new_nports = __pmAddPorts(env, ports, nports);		/* THREADSAFE */
+    PM_UNLOCK(__pmLock_extcall);
 
     /*
      * Add the default port, if no new ports were added or if there was an
@@ -161,8 +173,14 @@ __pmWebdAddPorts(int **ports, int nports)
     char *env;
     int  new_nports = nports;
 
-    if ((env = getenv("PMWEBD_PORT")) != NULL)
-	new_nports = __pmAddPorts(env, ports, nports);
+    PM_LOCK(__pmLock_extcall);
+    if ((env = getenv("PMWEBD_PORT")) != NULL)		/* THREADSAFE */
+	/*
+	 * THREADSAFE because __pmAddPorts acquires no locks (other than
+	 * on the fatal __pmNoMem() path)
+	 */
+	new_nports = __pmAddPorts(env, ports, nports);		/* THREADSAFE */
+    PM_UNLOCK(__pmLock_extcall);
 
     /*
      * Add the default port, if no new ports were added or if there was an
@@ -326,7 +344,9 @@ AddRequestPort(const char *address, int port)
 {
     ReqPortInfo	*rp;
 
-    if (address == NULL)
+    if (address == NULL && port == 0)
+	address = "INADDR_LOOPBACK";
+    else if (address == NULL)
 	address = "INADDR_ANY";
 
     if (nReqPorts == szReqPorts)
@@ -409,7 +429,8 @@ AddressFamily(int family)
  * Otherwise:
  * address is a string representing the Inet/IPv6 address that the port
  * is advertised for.  To allow connections to all this host's internet
- * addresses from clients use address = "INADDR_ANY".
+ * addresses from clients use address == "INADDR_ANY", or for localhost
+ * access only use address == "INADDR_LOOPBACK".
  * On input, 'family' is a pointer to the address family to use (AF_INET,
  * AF_INET6) if the address specified is empty.  If the spec is not
  * empty then family is ignored and is set to the actual address family
@@ -423,6 +444,9 @@ OpenRequestSocket(int port, const char *address, int *family,
     int			one, sts;
     __pmSockAddr	*myAddr;
     int			isUnix = 0;
+    int			try;
+    struct timeval	tick = { 0, 250*1000 };		/* 250msec */
+    char 		errmsg[PM_MAXERRMSGLEN];
 
     /*
      * Using this flag will eliminate the need for more conditional
@@ -447,21 +471,33 @@ OpenRequestSocket(int port, const char *address, int *family,
 	fd = __pmCreateUnixSocket();
     }
     else {
-	if ((myAddr = __pmStringToSockAddr(address)) == NULL) {
-	    __pmNotifyErr(LOG_ERR, "OpenRequestSocket(%d, %s) invalid address\n",
-			  port, address);
-	    goto fail;
-	}
-
 	/*
 	 * If the address is unspecified, then use the address family we
 	 * have been given, otherwise the family will be determined by
 	 * __pmStringToSockAddr.
 	 */
-	if (address == NULL || strcmp(address, "INADDR_ANY") == 0)
-	    __pmSockAddrSetFamily(myAddr, *family);
-	else
+	if (address == NULL || strcmp(address, "INADDR_ANY") == 0) {
+	    if ((myAddr = __pmSockAddrAlloc()) == NULL) {
+		__pmNoMem("OpenRequestSocket: can't allocate socket address",
+			  sizeof(*myAddr), PM_FATAL_ERR);
+	    }
+	    __pmSockAddrInit(myAddr, *family, INADDR_ANY, 0);
+	}
+	else if (strcmp(address, "INADDR_LOOPBACK") == 0) {
+	    if ((myAddr = __pmSockAddrAlloc()) == NULL) {
+		__pmNoMem("OpenRequestSocket: can't allocate socket address",
+			  sizeof(*myAddr), PM_FATAL_ERR);
+	    }
+	    __pmSockAddrInit(myAddr, *family, INADDR_LOOPBACK, 0);
+	}
+	else {
+	    if ((myAddr = __pmStringToSockAddr(address)) == NULL) {
+		__pmNotifyErr(LOG_ERR, "OpenRequestSocket(%d, %s) invalid address\n",
+			      port, address);
+		goto fail;
+	    }
 	    *family = __pmSockAddrGetFamily(myAddr);
+	}
 	__pmSockAddrSetPort(myAddr, port);
 
 	/* Create the socket. */
@@ -478,7 +514,7 @@ OpenRequestSocket(int port, const char *address, int *family,
 
     if (fd < 0) {
 	__pmNotifyErr(LOG_ERR, "OpenRequestSocket(%d, %s, %s) __pmCreateSocket: %s\n",
-		port, address, AddressFamily(*family), netstrerror());
+		port, address, AddressFamily(*family), netstrerror_r(errmsg, sizeof(errmsg)));
 	goto fail;
     }
 
@@ -489,7 +525,7 @@ OpenRequestSocket(int port, const char *address, int *family,
 		       (__pmSockLen)sizeof(one)) < 0) {
 	__pmNotifyErr(LOG_ERR,
 		      "OpenRequestSocket(%d, %s, %s) __pmSetSockOpt(SO_REUSEADDR): %s\n",
-		      port, address, AddressFamily(*family), netstrerror());
+		      port, address, AddressFamily(*family), netstrerror_r(errmsg, sizeof(errmsg)));
 	goto fail;
     }
 #else
@@ -497,7 +533,7 @@ OpenRequestSocket(int port, const char *address, int *family,
 		       (__pmSockLen)sizeof(one)) < 0) {
 	__pmNotifyErr(LOG_ERR,
 		      "OpenRequestSocket(%d, %s, %s) __pmSetSockOpt(EXCLUSIVEADDRUSE): %s\n",
-		      port, address, AddressFamily(*family), netstrerror());
+		      port, address, AddressFamily(*family), netstrerror_r(errmsg, sizeof(errmsg)));
 	goto fail;
     }
 #endif
@@ -507,18 +543,36 @@ OpenRequestSocket(int port, const char *address, int *family,
 		(__pmSockLen)sizeof(one)) < 0) {
 	__pmNotifyErr(LOG_ERR,
 		"OpenRequestSocket(%d, %s, %s) __pmSetSockOpt(SO_KEEPALIVE): %s\n",
-		port, address, AddressFamily(*family), netstrerror());
+		port, address, AddressFamily(*family), netstrerror_r(errmsg, sizeof(errmsg)));
 	goto fail;
     }
 
-    sts = __pmBind(fd, (void *)myAddr, __pmSockAddrSize());
+    /*
+     * Attempts to make daemon restart (especailly pmcd) have increased
+     * the probability of trying to reuse a socket before the last use
+     * has been completely torn down ... so be prepared to try this a
+     * few times (4 x 250msec)
+     */
+    sts = -1;
+    try = 0;
+    while (sts < 0 && try < 4) {
+	sts = __pmBind(fd, (void *)myAddr, __pmSockAddrSize());
+	if (sts >= 0)
+	    break;
+#ifdef PCP_DEBUG
+	if (pmDebug & DBG_TRACE_DESPERATE)
+	    fprintf(stderr, "OpenRequestSocket(%d, %s, %s) __pmBind try %d: %s\n",
+		port, address, AddressFamily(*family), try+1, netstrerror_r(errmsg, sizeof(errmsg)));
+#endif
+	__pmtimevalSleep(tick);
+	try++;
+    }
     __pmSockAddrFree(myAddr);
     myAddr = NULL;
     if (sts < 0) {
-	sts = neterror();
 	__pmNotifyErr(LOG_ERR, "OpenRequestSocket(%d, %s, %s) __pmBind: %s\n",
-		port, address, AddressFamily(*family), netstrerror());
-	if (sts == EADDRINUSE)
+		port, address, AddressFamily(*family), netstrerror_r(errmsg, sizeof(errmsg)));
+	if (neterror() == EADDRINUSE)
 	    __pmNotifyErr(LOG_ERR, "%s may already be running\n", pmProgname);
 	goto fail;
     }
@@ -533,7 +587,7 @@ OpenRequestSocket(int port, const char *address, int *family,
 	if (sts != 0) {
 	    __pmNotifyErr(LOG_ERR,
 		"OpenRequestSocket(%d, %s, %s) chmod(%s): %s\n",
-		port, address, AddressFamily(*family), address, strerror(errno));
+		port, address, AddressFamily(*family), address, netstrerror_r(errmsg, sizeof(errmsg)));
 	    goto fail;
 	}
     }
@@ -541,7 +595,7 @@ OpenRequestSocket(int port, const char *address, int *family,
     sts = __pmListen(fd, backlog);	/* Max. pending connection requests */
     if (sts < 0) {
 	__pmNotifyErr(LOG_ERR, "OpenRequestSocket(%d, %s, %s) __pmListen: %s\n",
-		port, address, AddressFamily(*family), netstrerror());
+		port, address, AddressFamily(*family), netstrerror_r(errmsg, sizeof(errmsg)));
 	goto fail;
     }
 
@@ -576,15 +630,18 @@ OpenRequestPorts(__pmFdSet *fdset, int backlog)
 
     for (i = 0; i < nReqPorts; i++) {
 	ReqPortInfo	*rp = &reqPorts[i];
-	int		portsOpened = 0;;
+	int		portsOpened = 0;
 
 	/*
-	 * If the spec is NULL or "INADDR_ANY", then we open one socket
-	 * for each address family (inet, IPv6).  Otherwise, the address
-	 * family will be determined by OpenRequestSocket.  Reporting of
-	 * all errors is left to OpenRequestSocket to avoid doubling up.
+	 * If the spec is NULL or "INADDR_ANY" or "INADDR_LOOPBACK", then
+	 * we open one socket for each address family (inet, IPv6).
+	 * Otherwise, the address family will be determined by
+	 * OpenRequestSocket.  Reporting of all errors is left to
+	 * OpenRequestSocket to avoid doubling up.
 	 */
-	if (rp->address == NULL || strcmp(rp->address, "INADDR_ANY") == 0) {
+	if (rp->address == NULL ||
+	    strcmp(rp->address, "INADDR_ANY") == 0 ||
+	    strcmp(rp->address, "INADDR_LOOPBACK") == 0) {
 	    family = AF_INET;
 	    if ((fd = OpenRequestSocket(rp->port, rp->address, &family,
 					backlog, fdset, &maximum)) >= 0) {
@@ -630,6 +687,7 @@ OpenRequestPorts(__pmFdSet *fdset, int backlog)
 	}
     }
 
+#ifndef IS_MINGW
     /* Open a local unix domain socket, if specified, and supported. */
     if (localSocketPath != NULL) {
 #if defined(HAVE_STRUCT_SOCKADDR_UN)
@@ -644,6 +702,7 @@ OpenRequestPorts(__pmFdSet *fdset, int backlog)
 		      pmProgname);
 #endif
     }
+#endif
 
     if (success)
 	return maximum;
@@ -842,6 +901,27 @@ __pmServerRequestPortString(int fd, char *buffer, size_t sz)
     return NULL;
 }
 
+unsigned int
+__pmServerGetFeaturesFromPDU(__pmPDU *pb)
+{
+    int         version;
+    int         challenge;
+    __pmPDUInfo         pduinfo;
+    int		sts;
+    unsigned int server_features=0;
+
+    version = __pmDecodeXtendError(pb, &sts, &challenge);
+
+    if( version >= 0 && version == PDU_VERSION2  && sts >=0 ){
+	pduinfo = __ntohpmPDUInfo(*(__pmPDUInfo *)&challenge);
+	server_features = pduinfo.features;
+    }
+    
+    return server_features;
+}
+
+static unsigned int server_features;
+
 #if !defined(HAVE_SECURE_SOCKETS)
 
 int
@@ -852,6 +932,15 @@ __pmSecureServerSetup(const char *db, const char *passwd)
     return 0;
 }
 
+int
+__pmSecureServerCertificateSetup(const char *db, const char *passwd, const char *cert_nickname)
+{
+    (void)db;
+    (void)passwd;
+    (void)cert_nickname;
+    return 0;
+}
+
 int
 __pmSecureServerIPCFlags(int fd, int flags)
 {
@@ -911,33 +1000,42 @@ __pmSecureServerHandshake(int fd, int flags, __pmHashCtl *attrs)
     return PM_ERR_IPC;
 }
 
+/*
+ * PM_SERVER_FEATURE_CERT_REQD is a special case that does not need a
+ * secure server, all other server features are not supported
+ */
 int
 __pmSecureServerHasFeature(__pmServerFeature query)
 {
-    /* CREDS_REQD is a special case that does not need a secure server */
-    if ((query & PDU_FLAG_CREDS_REQD) != 0)
-	return 1;
+    if (query == PM_SERVER_FEATURE_CERT_REQD){
+	return (server_features & (1 << query)) != 0;
+    }
     return 0;
 }
 
 int
 __pmSecureServerSetFeature(__pmServerFeature wanted)
 {
-    (void)wanted;
+    if (wanted == PM_SERVER_FEATURE_CERT_REQD){
+        server_features |= (1 << wanted);
+        return 1;
+    }
+
     return 0;
 }
 
 int
 __pmSecureServerClearFeature(__pmServerFeature clear)
 {
-    (void)clear;
+    if (clear == PM_SERVER_FEATURE_CERT_REQD){
+    	server_features &= ~(1<<clear);
+	return 1;
+    }
     return 0;
 }
 
 #endif /* !HAVE_SECURE_SOCKETS */
 
-static unsigned int server_features;
-
 int
 __pmServerClearFeature(__pmServerFeature clear)
 {
@@ -951,7 +1049,8 @@ __pmServerClearFeature(__pmServerFeature clear)
 int
 __pmServerSetFeature(__pmServerFeature wanted)
 {
-    if (wanted == PM_SERVER_FEATURE_DISCOVERY ||
+    if (wanted == PM_SERVER_FEATURE_LOCAL ||
+	wanted == PM_SERVER_FEATURE_DISCOVERY ||
 	wanted == PM_SERVER_FEATURE_CREDS_REQD ||
 	wanted == PM_SERVER_FEATURE_UNIX_DOMAIN) {
 	server_features |= (1 << wanted);
@@ -977,6 +1076,7 @@ __pmServerHasFeature(__pmServerFeature query)
     case PM_SERVER_FEATURE_IPV6:
 	sts = (strcmp(__pmGetAPIConfig("ipv6"), "true") == 0);
 	break;
+    case PM_SERVER_FEATURE_LOCAL:
     case PM_SERVER_FEATURE_DISCOVERY:
     case PM_SERVER_FEATURE_CONTAINERS:
     case PM_SERVER_FEATURE_CREDS_REQD:
