@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2012-2014 Red Hat.
+ * Copyright (c) 2012-2014,2017 Red Hat.
  * Copyright (c) 1995-2002,2004 Silicon Graphics, Inc.  All Rights Reserved.
  * 
  * This library is free software; you can redistribute it and/or modify it
@@ -14,7 +14,7 @@
  *
  * Thread-safe notes
  *
- * Do not need ctxp->c_pmcd->pc_lock lock around __pmSendCreds() call,
+ * Do not need ctxp->c_pmcd->pc_lock mutex around __pmSendCreds() call,
  * as the context has not been created, so no-one else could be using
  * the context's fd.
  */
@@ -83,6 +83,134 @@ negotiate_proxy(int fd, const char *hostname, int port)
     return ok;
 }
 
+/*
+ * Verify that the requested context flags (ctxflags) are
+ * viable given available functionality from pmcd at the
+ * other end of a connection (features).
+ * Returns appropriate PDU flags for the client to send to
+ * pmcd if we can go ahead with the connection, else error.
+ */
+static int
+check_feature_flags(int ctxflags, int features, int local_conn)
+{
+    int		pduflags = 0;
+
+    if (features & PDU_FLAG_CREDS_REQD)
+	/*
+	 * This is a mandatory connection feature - pmcd must be
+	 * sent user credential information one way or another -
+	 * i.e. via SASL2 authentication, or AF_UNIX peer creds.
+	 */
+	pduflags |= PDU_FLAG_CREDS_REQD;
+
+    if ((features & PDU_FLAG_CERT_REQD) && !local_conn) {
+	/*
+	 * This is a mandatory connection feature for remote connections.
+	 * pmcd must be sent a trusted certificate.
+	 */
+	pduflags |= PDU_FLAG_CERT_REQD;
+	if (!(ctxflags & PM_CTXFLAG_SECURE)) {
+	    /* PMCD requires a client cert, but we are not even setup for secure connections */
+	    return PM_ERR_NEEDCLIENTCERT;
+	}
+     }
+
+    if (ctxflags) {
+	/*
+	 * If an optional connection feature (e.g. encryption) is
+	 * desired, the pmcd that we're talking to must advertise
+	 * support for the feature.  And if it did, the client in
+	 * turn must request it be enabled (now, via pduflags).
+	 */
+	if (ctxflags & (PM_CTXFLAG_SECURE|PM_CTXFLAG_RELAXED)) {
+	    if (features & PDU_FLAG_SECURE) {
+		pduflags |= PDU_FLAG_SECURE;
+		/*
+		 * Determine whether the server can send an ACK for a
+		 * secure connection request. We can still connect
+		 * whether it does or not, but we need to know the
+		 * protocol.
+		 */
+		if (features & PDU_FLAG_SECURE_ACK)
+		    pduflags |= PDU_FLAG_SECURE_ACK;
+	    } else if (ctxflags & PM_CTXFLAG_SECURE) {
+		return -EOPNOTSUPP;
+	    }
+	}
+	if (ctxflags & PM_CTXFLAG_COMPRESS) {
+	    if (features & PDU_FLAG_COMPRESS)
+		pduflags |= PDU_FLAG_COMPRESS;
+	    else
+		return -EOPNOTSUPP;
+	}
+	if (ctxflags & PM_CTXFLAG_AUTH) {
+	    if (features & PDU_FLAG_AUTH)
+		pduflags |= PDU_FLAG_AUTH;
+	    else
+		return -EOPNOTSUPP;
+	}
+	if (ctxflags & PM_CTXFLAG_CONTAINER) {
+	    if (features & PDU_FLAG_CONTAINER)
+		pduflags |= PDU_FLAG_CONTAINER;
+	    else
+		return -EOPNOTSUPP;
+	}
+    }
+    return pduflags;
+}
+
+static int
+container_handshake(int fd, __pmHashCtl *attrs)
+{
+    __pmHashNode *node;
+    const char *name;
+    int length;
+
+    if ((node = __pmHashSearch(PCP_ATTR_CONTAINER, attrs)) == NULL)
+	return -ESRCH;
+    if ((name = (const char *)node->data) == NULL)
+	return -ESRCH;
+    length = strlen(name);
+
+    if (pmDebug & DBG_TRACE_CONTEXT)
+	fprintf(stderr, "%s:__pmConnectHandshake container=\"%s\" [%d]\n",
+		__FILE__, name, length);
+
+    return __pmSendAttr(fd, FROM_ANON, PCP_ATTR_CONTAINER, name, length);
+}
+
+static int
+attributes_handshake(int fd, int flags, const char *host, __pmHashCtl *attrs)
+{
+    int sts;
+
+    if ((sts = __pmSecureClientHandshake(fd, flags, host, attrs)) < 0)
+	return sts;
+
+    /* additional connection attributes, done after secure setup */
+    if (((flags & PDU_FLAG_CONTAINER) != 0) &&
+	((sts = container_handshake(fd, attrs)) < 0))
+	return sts;
+
+    return 0;
+}
+
+static int
+is_local_connection(const char *hostname, __pmHashCtl *attrs)
+{
+    /* 
+     * If attrs has PCP_ATTR_UNIXSOCK or PCP_ATTR_LOCAL
+     * OR
+     * If hostname contains localhost
+     *	We consider this a local connection and don't enforce CERT_REQD
+     */
+    if (__pmHashSearch(PCP_ATTR_UNIXSOCK, attrs) || 
+	__pmHashSearch(PCP_ATTR_LOCAL, attrs) || 
+	strstr(hostname, "localhost"))
+	return 1;
+    return 0;
+}
+
 /*
  * client connects to pmcd handshake
  */
@@ -123,88 +251,54 @@ __pmConnectHandshake(int fd, const char *hostname, int ctxflags, __pmHashCtl *at
 	if (version == PDU_VERSION2) {
 	    __pmPDUInfo		pduinfo;
 	    __pmVersionCred	handshake;
-	    int			pduflags = 0;
+	    int			pduflags;
+	    int			local_conn;
+
+	    local_conn = is_local_connection(hostname, attrs);
 
 	    pduinfo = __ntohpmPDUInfo(*(__pmPDUInfo *)&challenge);
+	    pduflags = sts = check_feature_flags(ctxflags, pduinfo.features, local_conn);
+	    if (sts < 0) {
+		__pmUnpinPDUBuf(pb);
+		return sts;
+	    }
 
-	    if (pduinfo.features & PDU_FLAG_CREDS_REQD)
-		/*
-		 * This is a mandatory connection feature - pmcd must be
-		 * sent user credential information one way or another -
-		 * i.e. via SASL2 authentication, or AF_UNIX peer creds.
-		 */
-		pduflags |= PDU_FLAG_CREDS_REQD;
-
-	    if (ctxflags) {
-		/*
-		 * If an optional connection feature (e.g. encryption) is
-		 * desired, the pmcd that we're talking to must advertise
-		 * support for the feature.  And if it did, the client in
-		 * turn must request it be enabled (now, via pduflags).
-		 */
-		if (ctxflags & (PM_CTXFLAG_SECURE|PM_CTXFLAG_RELAXED)) {
-		    if (pduinfo.features & PDU_FLAG_SECURE) {
-			pduflags |= PDU_FLAG_SECURE;
-			/*
-			 * Determine whether the server can send an ACK for a
-			 * secure connection request. We can still connect
-			 * whether it does or not, but we need to know the
-			 * protocol.
-			 */
-			if (pduinfo.features & PDU_FLAG_SECURE_ACK)
-			    pduflags |= PDU_FLAG_SECURE_ACK;
-		    } else if (ctxflags & PM_CTXFLAG_SECURE) {
-			__pmUnpinPDUBuf(pb);
-			return -EOPNOTSUPP;
-		    }
-		}
-		if (ctxflags & PM_CTXFLAG_COMPRESS) {
-		    if (pduinfo.features & PDU_FLAG_COMPRESS)
-			pduflags |= PDU_FLAG_COMPRESS;
-		    else {
-			__pmUnpinPDUBuf(pb);
-			return -EOPNOTSUPP;
-		    }
-		}
-		if (ctxflags & PM_CTXFLAG_AUTH) {
-		    if (pduinfo.features & PDU_FLAG_AUTH)
-			pduflags |= PDU_FLAG_AUTH;
-		    else {
-			__pmUnpinPDUBuf(pb);
-			return -EOPNOTSUPP;
-		    }
-		}
+	    if ((ok = __pmSetFeaturesIPC(fd, version, pduinfo.features)) < 0) {
+		__pmUnpinPDUBuf(pb);
+		return ok;
 	    }
 
 	    /*
 	     * Negotiate connection version and features (via creds PDU)
 	     */
-	    if ((ok = __pmSetVersionIPC(fd, version)) < 0) {
-		__pmUnpinPDUBuf(pb);
-		return ok;
-	    }
-
 	    memset(&handshake, 0, sizeof(handshake));
 	    handshake.c_type = CVERSION;
 	    handshake.c_version = PDU_VERSION;
 	    handshake.c_flags = pduflags;
-
 	    sts = __pmSendCreds(fd, (int)getpid(), 1, (__pmCred *)&handshake);
 
 	    /*
-	     * At this point we know caller wants to set channel options and
-	     * pmcd supports them so go ahead and update the socket now (this
+	     * At this point we know the caller wants to set channel options
+	     * and pmcd supports them so go ahead and update the socket (this
 	     * completes the SSL handshake in encrypting mode, authentication
-	     * via SASL, and/or enabling compression in NSS).
+	     * via SASL, enabling compression in NSS, and any other requested
+	     * connection attributes).
 	     */
 	    if (sts >= 0 && pduflags)
-		sts = __pmSecureClientHandshake(fd, pduflags, hostname, attrs);
+		sts = attributes_handshake(fd, pduflags, hostname, attrs);
 	}
 	else
 	    sts = PM_ERR_IPC;
     }
-    else if (sts != PM_ERR_TIMEOUT)
-	sts = PM_ERR_IPC;
+    else {
+	/*
+	 * Note: no need to call __pmCloseChannelbyContext() here because
+	 * we only get here from __pmConnectPMCD() where if this call fails
+	 * the socket is closed and the channel is never set up.
+	 */
+	if (sts != PM_ERR_TIMEOUT)
+	    sts = PM_ERR_IPC;
+    }
 
     if (pinpdu > 0)
 	__pmUnpinPDUBuf(pb);
@@ -230,8 +324,10 @@ load_proxy_hostspec(pmHostSpec *proxy)
     char	errmsg[PM_MAXERRMSGLEN];
     char	*envstr;
 
-    if ((envstr = getenv("PMPROXY_HOST")) != NULL) {
+    PM_LOCK(__pmLock_extcall);
+    if ((envstr = getenv("PMPROXY_HOST")) != NULL) {		/* THREADSAFE */
 	proxy->name = strdup(envstr);
+	PM_UNLOCK(__pmLock_extcall);
 	if (proxy->name == NULL) {
 	    __pmNotifyErr(LOG_WARNING,
 			  "__pmConnectPMCD: cannot save PMPROXY_HOST: %s\n",
@@ -245,6 +341,9 @@ load_proxy_hostspec(pmHostSpec *proxy)
 	    proxy->nports = __pmProxyAddPorts(&proxy->ports, proxy->nports);
 	}
     }
+    else
+	PM_UNLOCK(__pmLock_extcall);
+
 }
 
 void
