@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2013-2014 Red Hat.
+ * Copyright (c) 2013-2017 Red Hat.
  * Copyright (c) 2010 Ken McDonell.  All Rights Reserved.
  * Copyright (c) 1995-2002,2004 Silicon Graphics, Inc.  All Rights Reserved.
  * 
@@ -67,8 +67,10 @@ build_dsotab(void)
      *
      */
     char	configFileName[MAXPATHLEN];
+    char	pathbuf[MAXPATHLEN];
     FILE	*configFile;
     char	*config;
+    char	*pmdas;
     char	*p;
     char	*q;
     struct stat	sbuf;
@@ -81,7 +83,12 @@ build_dsotab(void)
     numdso = 0;
     dsotab = NULL;
 
-    strcpy(configFileName, pmGetConfig("PCP_PMCDCONF_PATH"));
+    if ((pmdas = pmGetOptionalConfig("PCP_PMDAS_DIR")) == NULL)
+	return PM_ERR_GENERIC;
+    if ((config = pmGetOptionalConfig("PCP_PMCDCONF_PATH")) == NULL)
+	return PM_ERR_GENERIC;
+    strncpy(configFileName, config, sizeof(configFileName));
+    configFileName[sizeof(configFileName) - 1] = '\0';
 #ifdef PCP_DEBUG
     if (pmDebug & DBG_TRACE_CONTEXT) {
 	fprintf(stderr, "build_dsotab: parsing %s\n", configFileName);
@@ -119,6 +126,19 @@ build_dsotab(void)
 	     */
 	    goto eatline;
 	}
+	if (strncmp(p, "linux", 5) == 0) {
+	    /*
+	     * the Linux PMDA is an exception now too ... we run it as root
+	     * (daemon) but we still want to make the DSO available for any
+	     * local context users.  We add this explicitly below.
+	     */
+	    domain = 60;
+	    init = "linux_init";
+	    snprintf(pathbuf, sizeof(pathbuf), "%s/linux/pmda_linux.so", pmdas);
+	    name = pathbuf;
+	    peekc = *p;
+	    goto dsoload;
+	}
 	/* skip the PMDA's name */
 	while (*p != '\0' && *p != '\n' && !isspace((int)*p))
 	    p++;
@@ -149,6 +169,7 @@ build_dsotab(void)
 	    p++;
 	peekc = *p;
 	*p = '\0';
+dsoload:
 #ifdef PCP_DEBUG
 	if (pmDebug & DBG_TRACE_CONTEXT) {
 	    fprintf(stderr, "[%d] domain=%d, name=%s, init=%s\n", lineno, domain, name, init);
@@ -197,7 +218,7 @@ build_dsoattrs(pmdaInterface *dispatch, __pmHashCtl *attrs)
 	__pmHashAdd(PCP_ATTR_GROUPID, namep, attrs);
 #endif
 
-    snprintf(name, sizeof(name), "%u", getpid());
+    snprintf(name, sizeof(name), "%" FMT_PID, getpid());
     name[sizeof(name)-1] = '\0';
     if ((namep = strdup(name)) != NULL)
 	__pmHashAdd(PCP_ATTR_PROCESSID, namep, attrs);
@@ -207,7 +228,8 @@ build_dsoattrs(pmdaInterface *dispatch, __pmHashCtl *attrs)
 	     node != NULL;
 	     node = __pmHashWalk(attrs, PM_HASH_WALK_NEXT)) {
 	    if ((sts = dispatch->version.six.attribute(
-				0, node->key, node->data,
+				dispatch->version.six.ext->e_context,
+				node->key, node->data,
 				node->data ? strlen(node->data)+1 : 0,
 				dispatch->version.six.ext)) < 0)
 		break;
@@ -293,6 +315,7 @@ __pmConnectLocal(__pmHashCtl *attrs)
     int			i;
     __pmDSO		*dp;
     char		pathbuf[MAXPATHLEN];
+    char		*pmdas;
     const char		*path;
 #if defined(HAVE_DLOPEN)
     unsigned int	challenge;
@@ -302,6 +325,9 @@ __pmConnectLocal(__pmHashCtl *attrs)
 #endif
 #endif
 
+    if ((pmdas = pmGetOptionalConfig("PCP_PMDAS_DIR")) == NULL)
+	return PM_ERR_GENERIC;
+
     if (numdso == -1) {
 	int	sts;
 	sts = build_dsotab();
@@ -310,20 +336,23 @@ __pmConnectLocal(__pmHashCtl *attrs)
 
     for (i = 0; i < numdso; i++) {
 	dp = &dsotab[i];
-	if (dp->domain == -1 || dp->handle != NULL)
+	if (dp->domain == -1)
 	    continue;
+	if (dp->handle != NULL)
+	    goto connected;
+
 	/*
 	 * __pmLocalPMDA() means the path to the DSO may be something
 	 * other than relative to $PCP_PMDAS_DIR ... need to try both
 	 * options and also with and without DSO_SUFFIX (so, dll, etc)
 	 */
-	snprintf(pathbuf, sizeof(pathbuf), "%s%c%s",
-		 pmGetConfig("PCP_PMDAS_DIR"), __pmPathSeparator(), dp->name);
-	if ((path = __pmFindPMDA(pathbuf)) == NULL) {
-	    snprintf(pathbuf, sizeof(pathbuf), "%s%c%s.%s",
-		 pmGetConfig("PCP_PMDAS_DIR"), __pmPathSeparator(), dp->name, DSO_SUFFIX);
+	if ((path = __pmFindPMDA(dp->name)) == NULL) {
+	    snprintf(pathbuf, sizeof(pathbuf), "%s%c%s",
+			pmdas, __pmPathSeparator(), dp->name);
 	    if ((path = __pmFindPMDA(pathbuf)) == NULL) {
-		if ((path = __pmFindPMDA(dp->name)) == NULL) {
+		snprintf(pathbuf, sizeof(pathbuf), "%s%c%s.%s",
+			    pmdas, __pmPathSeparator(), dp->name, DSO_SUFFIX);
+		if ((path = __pmFindPMDA(pathbuf)) == NULL) {
 		    snprintf(pathbuf, sizeof(pathbuf), "%s.%s", dp->name, DSO_SUFFIX);
 		    if ((path = __pmFindPMDA(pathbuf)) == NULL) {
 			pmprintf("__pmConnectLocal: Warning: cannot find DSO at \"%s\" or \"%s\"\n", 
@@ -339,8 +368,18 @@ __pmConnectLocal(__pmHashCtl *attrs)
 #if defined(HAVE_DLOPEN)
 	dp->handle = dlopen(path, RTLD_NOW);
 	if (dp->handle == NULL) {
+	    char	*errmsg;
+	    /*
+	     * Threadsafe note:
+	     *	dlerror() may not be threadsafe according to POSIX,
+	     *	but the glibc version is threadsafe.  Locking here
+	     *	is just being a little overly careful.
+	     */
+	    PM_LOCK(__pmLock_extcall);
+	    errmsg = dlerror();		/* THREADSAFE */
+	    PM_UNLOCK(__pmLock_extcall);
 	    pmprintf("__pmConnectLocal: Warning: error attaching DSO "
-		     "\"%s\"\n%s\n\n", path, dlerror());
+		     "\"%s\"\n%s\n\n", path, errmsg);
 	    pmflush();
 	    dp->domain = -1;
 	}
@@ -387,7 +426,7 @@ __pmConnectLocal(__pmHashCtl *attrs)
 	 */
 	challenge = 0xff;
 	dp->dispatch.comm.pmda_interface = challenge;
-	dp->dispatch.comm.pmapi_version = ~PMAPI_VERSION;
+	dp->dispatch.comm.pmapi_version = (~PMAPI_VERSION) & 0xff;
 	dp->dispatch.comm.flags = 0;
 	dp->dispatch.status = 0;
 
@@ -404,11 +443,12 @@ __pmConnectLocal(__pmHashCtl *attrs)
 	    dp->domain = -1;
 	}
 	else {
+connected:
 	    if (dp->dispatch.comm.pmda_interface < PMDA_INTERFACE_2 ||
 		dp->dispatch.comm.pmda_interface > PMDA_INTERFACE_LATEST) {
 		pmprintf("__pmConnectLocal: Error: Unknown PMDA interface "
 			 "version %d in \"%s\" DSO\n", 
-			 dp->dispatch.comm.pmda_interface, path);
+			 dp->dispatch.comm.pmda_interface, dp->name);
 		pmflush();
 		dlclose(dp->handle);
 		dp->domain = -1;
@@ -416,15 +456,19 @@ __pmConnectLocal(__pmHashCtl *attrs)
 	    else if (dp->dispatch.comm.pmapi_version != PMAPI_VERSION_2) {
 		pmprintf("__pmConnectLocal: Error: Unknown PMAPI version %d "
 			 "in \"%s\" DSO\n",
-			 dp->dispatch.comm.pmapi_version, path);
+			 dp->dispatch.comm.pmapi_version, dp->name);
 		pmflush();
 		dlclose(dp->handle);
 		dp->domain = -1;
 	    }
-	    else if (dp->dispatch.comm.pmda_interface >= PMDA_INTERFACE_6 &&
+	    else {
+		if (dp->dispatch.comm.pmda_interface >= PMDA_INTERFACE_5)
+		    dp->dispatch.version.four.ext->e_context = pmWhichContext();
+		if (dp->dispatch.comm.pmda_interface >= PMDA_INTERFACE_6 &&
 		    (dp->dispatch.comm.flags & PDU_FLAG_AUTH) != 0) {
-		/* Agent wants to know about connection attributes */
-		build_dsoattrs(&dp->dispatch, attrs);
+		    /* Agent wants to know about connection attributes */
+		    build_dsoattrs(&dp->dispatch, attrs);
+		}
 	    }
 	}
 #ifdef HAVE_ATEXIT
