@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2013 Red Hat.
+ * Copyright (c) 2013-2017 Red Hat.
  * Copyright (c) 1995-2002 Silicon Graphics, Inc.  All Rights Reserved.
  * 
  * This library is free software; you can redistribute it and/or modify it
@@ -18,13 +18,14 @@
 #include "fault.h"
 #include "internal.h"
 #include <stddef.h>
+#include <assert.h>
 
 /* bytes for a length field in a header/trailer, or a string length field */
 #define LENSIZE	4
 
 #ifdef PCP_DEBUG
 static void
-StrTimeval(__pmTimeval *tp)
+StrTimeval(const __pmTimeval *tp)
 {
     if (tp == NULL)
 	fprintf(stderr, "<null timeval>");
@@ -33,13 +34,57 @@ StrTimeval(__pmTimeval *tp)
 }
 #endif
 
+/*
+ * Return 1 if the indoms are the same, 0 otherwise.
+ * The time stamp does not count for this comparison.
+ */
+static int
+sameindom(const __pmLogInDom *idp1, const __pmLogInDom *idp2)
+{
+    int i, j;
+
+    if (idp1->numinst != idp2->numinst)
+	return 0; /* different */
+
+    /*
+     * Make sure that the instances and their names are the same.
+     * We can't assume that the instances are always in the same order,
+     * but we do assume that each instance occurs only once.
+     */
+    for (i = 0; i < idp1->numinst; ++i) {
+	for (j = 0; j < idp2->numinst; ++j) {
+	    if (idp1->instlist[i] == idp2->instlist[j]) {
+		/*
+		 * We found the same instance. Make sure that the names are
+		 * the same.
+		 */
+		if (strcmp(idp1->namelist[i], idp2->namelist[j]) != 0)
+		    return 0; /* different */
+		break;
+	    }
+	}
+	if (j >= idp2->numinst) {
+	    /* The current idp1 instance was not found in idp2. */
+	    return 0; /* different */
+	}
+    }
+
+    return 1; /* duplicate */
+}
+
+/*
+ * Add the given instance domain to the hashed instance domain.
+ * Filter out duplicates.
+ */
 static int
 addindom(__pmLogCtl *lcp, pmInDom indom, const __pmTimeval *tp, int numinst, 
          int *instlist, char **namelist, int *indom_buf, int allinbuf)
 {
-    __pmLogInDom	*idp;
+    __pmLogInDom	*idp, *idp_prev;
+    __pmLogInDom	*idp_cached, *idp_time;
     __pmHashNode	*hp;
-    int		sts;
+    int			timecmp;
+    int			sts;
 
 PM_FAULT_POINT("libpcp/" __FILE__ ":1", PM_FAULT_ALLOC);
     if ((idp = (__pmLogInDom *)malloc(sizeof(__pmLogInDom))) == NULL)
@@ -53,45 +98,147 @@ PM_FAULT_POINT("libpcp/" __FILE__ ":1", PM_FAULT_ALLOC);
 
 #ifdef PCP_DEBUG
     if (pmDebug & DBG_TRACE_LOGMETA) {
-	char	strbuf[20];
+	char    strbuf[20];
 	fprintf(stderr, "addindom( ..., %s, ", pmInDomStr_r(indom, strbuf, sizeof(strbuf)));
 	StrTimeval((__pmTimeval *)tp);
 	fprintf(stderr, ", numinst=%d)\n", numinst);
     }
 #endif
 
-
     if ((hp = __pmHashSearch((unsigned int)indom, &lcp->l_hashindom)) == NULL) {
 	idp->next = NULL;
 	sts = __pmHashAdd((unsigned int)indom, (void *)idp, &lcp->l_hashindom);
+	if (sts > 0) {
+	    /* __pmHashAdd returns 1 for success, but we want 0. */
+	    sts = 0;
+	}
+	return sts;
     }
-    else {
+
+    /*
+     * Filter out identical indoms. This is very common in multi-archive
+     * contexts where the individual archives almost always use the same
+     * instance domains.
+     *
+     * The indoms need to be sorted by decreasing time stamp. Before
+     * multi-archive contexts, this happened automatically. Now we
+     * must do it explicitly. Duplicates must be moved to the head of their
+     * time slot.
+     */
+    sts = 0;
+    idp_prev = NULL;
+    for (idp_cached = (__pmLogInDom *)hp->data; idp_cached; idp_cached = idp_cached->next) {
+	timecmp = __pmTimevalCmp(&idp_cached->stamp, &idp->stamp);
+
+	/*
+	 * If the time of the current cached item is before our time,
+	 * then insert here.
+	 */
+	if (timecmp < 0)
+	    break;
+
+	/*
+	 * If the time of the current cached item is the same as our time,
+	 * search for a duplicate in this time slot. If found, move it
+	 * to the head of this time slot. Otherwise insert this new item
+	 * at the head of the time slot.
+	 */
+	if (timecmp == 0) {
+	    assert(sts == 0);
+	    idp_time = idp_prev; /* just before this time slot */
+	    do {
+		/* Have we found a duplicate? */
+		if (sameindom(idp_cached, idp)) {
+		    sts = PMLOGPUTINDOM_DUP; /* duplicate */
+		    break;
+		}
+		/* Try the next one */
+		idp_prev = idp_cached;
+		idp_cached = idp_cached->next;
+		if (idp_cached == NULL)
+		    break;
+		timecmp = __pmTimevalCmp(&idp_cached->stamp, &idp->stamp);
+	    } while (timecmp == 0);
+
+	    if (sts == PMLOGPUTINDOM_DUP) {
+		/*
+		 * We found a duplicate. We can't free instlist, namelist and
+		 * indom_buf because we don't know where the storage
+		 * came from. Only the caller knows. The best we can do is to
+		 * indicate that we found a duplicate and let the caller manage
+		 * them. We do, however need to free idp.
+		 */
+		free(idp);
+		if (idp_prev == idp_time) {
+		    /* The duplicate is already in the right place. */
+		    return sts; /* ok -- duplicate */
+		}
+
+		/* Unlink the duplicate and set it up to be re-inserted. */
+		assert(idp_cached != NULL);
+		if (idp_prev)
+		    idp_prev->next = idp_cached->next;
+		else
+		    hp->data = (void *)idp_cached->next;
+		idp = idp_cached;
+	    }
+
+	    /*
+	     * Regardless of whether or not a duplicate was found, we will be
+	     * inserting the indom we have at the head of the time slot.
+	     */
+	    idp_prev = idp_time;
+	    break;
+	}
+
+	/*
+	 * The time of the current cached item is after our time.
+	 * Just keep looking.
+	 */
+	idp_prev = idp_cached;
+    }
+
+    /* Insert at the identified insertion point. */
+    if (idp_prev == NULL) {
 	idp->next = (__pmLogInDom *)hp->data;
 	hp->data = (void *)idp;
-	sts = 0;
     }
+    else {
+	idp->next = idp_prev->next;
+	idp_prev->next = idp;
+    }
+
     return sts;
 }
 
 /*
  * Load _all_ of the hashed pmDesc and __pmLogInDom structures from the metadata
  * log file -- used at the initialization (NewContext) of an archive.
- * Also load all the metric names from the metadata log file and create l_pmns.
+ * Also load all the metric names from the metadata log file and create l_pmns,
+ * if it does not already exist.
  */
 int
 __pmLogLoadMeta(__pmLogCtl *lcp)
 {
-    int		rlen;
-    int		check;
-    pmDesc	*dp;
-    int		sts = 0;
-    __pmLogHdr	h;
-    FILE	*f = lcp->l_mdfp;
-    int		numpmid = 0;
-    int		n;
+    __pmHashNode	*hp;
+    int			rlen;
+    int			check;
+    pmDesc		*dp;
+    pmDesc		*olddp;
+    int			sts = 0;
+    __pmLogHdr		h;
+    FILE		*f = lcp->l_mdfp;
+    int			numpmid = 0;
+    int			n;
+    int			numnames;
+    int			i;
+    int			len;
+    char		name[MAXPATHLEN];
     
-    if ((sts = __pmNewPMNS(&(lcp->l_pmns))) < 0)
-      goto end;
+    if (lcp->l_pmns == NULL) {
+	if ((sts = __pmNewPMNS(&(lcp->l_pmns))) < 0)
+	    goto end;
+    }
 
     fseek(f, (long)(sizeof(__pmLogLabel) + 2*sizeof(int)), SEEK_SET);
     for ( ; ; ) {
@@ -160,24 +307,78 @@ PM_FAULT_POINT("libpcp/" __FILE__ ":2", PM_FAULT_ALLOC);
 		dp->pmid = __ntohpmID(dp->pmid);
 	    }
 
-	    if ((sts = __pmHashAdd((int)dp->pmid, (void *)dp, &lcp->l_hashpmid)) < 0) {
+	    /* Add it to the hash pmid hash table. */
+	    if ((hp = __pmHashSearch((int)dp->pmid, &lcp->l_hashpmid)) != NULL) {
+		/*
+		 * This pmid is already in the hash table. Check for conflicts.
+		 */
+		olddp = (pmDesc *)hp->data;
+		if (dp->type != olddp->type) {
+		    sts = PM_ERR_LOGCHANGETYPE;
+		    free(dp);
+		    goto end;
+		}
+		if (dp->sem != olddp->sem) {
+		    sts = PM_ERR_LOGCHANGESEM;
+		    free(dp);
+		    goto end;
+		}
+		if (dp->indom != olddp->indom) {
+		    sts = PM_ERR_LOGCHANGEINDOM;
+		    free(dp);
+		    goto end;
+		}
+		if (dp->units.dimSpace != olddp->units.dimSpace ||
+		    dp->units.dimTime != olddp->units.dimTime ||
+		    dp->units.dimCount != olddp->units.dimCount ||
+		    dp->units.scaleSpace != olddp->units.scaleSpace ||
+		    dp->units.scaleTime != olddp->units.scaleTime ||
+		    dp->units.scaleCount != olddp->units.scaleCount) {
+		    sts = PM_ERR_LOGCHANGEUNITS;
+		    free(dp);
+		    goto end;
+		}
+                /*
+                 * This pmid is already known, and matches.  We can free the newly
+                 * read copy and use the one in the hash table. 
+                 */
+                free(dp);
+                dp = olddp;
+	    }
+	    else if ((sts = __pmHashAdd((int)dp->pmid, (void *)dp, &lcp->l_hashpmid)) < 0) {
 		free(dp);
 		goto end;
 	    }
 
-            else {
-                char name[MAXPATHLEN];
-                int numnames;
-		int i;
-		int len;
+	    /* read in the names & store in PMNS tree ... */
+	    if ((n = (int)fread(&numnames, 1, sizeof(numnames), f)) != 
+		sizeof(numnames)) {
+#ifdef PCP_DEBUG
+		if (pmDebug & DBG_TRACE_LOGMETA) {
+		    fprintf(stderr, "__pmLogLoadMeta: numnames read -> %d: expected: %d\n",
+			    n, (int)sizeof(numnames));
+		}
+#endif
+		if (ferror(f)) {
+		    clearerr(f);
+		    sts = -oserror();
+		}
+		else
+		    sts = PM_ERR_LOGREC;
+		goto end;
+	    }
+	    else {
+		/* swab numnames */
+		numnames = ntohl(numnames);
+	    }
 
-                /* read in the names & store in PMNS tree ... */
-		if ((n = (int)fread(&numnames, 1, sizeof(numnames), f)) != 
-                     sizeof(numnames)) {
+	    for (i = 0; i < numnames; i++) {
+		if ((n = (int)fread(&len, 1, sizeof(len), f)) != 
+		    sizeof(len)) {
 #ifdef PCP_DEBUG
 		    if (pmDebug & DBG_TRACE_LOGMETA) {
-			fprintf(stderr, "__pmLogLoadMeta: numnames read -> %d: expected: %d\n",
-				n, (int)sizeof(numnames));
+			fprintf(stderr, "__pmLogLoadMeta: len name[%d] read -> %d: expected: %d\n",
+				i, n, (int)sizeof(len));
 		    }
 #endif
 		    if (ferror(f)) {
@@ -189,70 +390,48 @@ PM_FAULT_POINT("libpcp/" __FILE__ ":2", PM_FAULT_ALLOC);
 		    goto end;
 		}
 		else {
-		    /* swab numnames */
-		    numnames = ntohl(numnames);
+		    /* swab len */
+		    len = ntohl(len);
 		}
 
- 		for (i = 0; i < numnames; i++) {
-		    if ((n = (int)fread(&len, 1, sizeof(len), f)) != 
-			 sizeof(len)) {
-#ifdef PCP_DEBUG
-			if (pmDebug & DBG_TRACE_LOGMETA) {
-			    fprintf(stderr, "__pmLogLoadMeta: len name[%d] read -> %d: expected: %d\n",
-				    i, n, (int)sizeof(len));
-			}
-#endif
-			if (ferror(f)) {
-			    clearerr(f);
-			    sts = -oserror();
-			}
-			else
-			    sts = PM_ERR_LOGREC;
-			goto end;
-		    }
-		    else {
-			/* swab len */
-			len = ntohl(len);
-		    }
-
-		    if ((n = (int)fread(name, 1, len, f)) != len) {
-#ifdef PCP_DEBUG
-			if (pmDebug & DBG_TRACE_LOGMETA) {
-			    fprintf(stderr, "__pmLogLoadMeta: name[%d] read -> %d: expected: %d\n",
-				    i, n, len);
-			}
-#endif
-			if (ferror(f)) {
-			    clearerr(f);
-			    sts = -oserror();
-			}
-			else
-			    sts = PM_ERR_LOGREC;
-			goto end;
-		    }
-                    name[len] = '\0';
+		if ((n = (int)fread(name, 1, len, f)) != len) {
 #ifdef PCP_DEBUG
 		    if (pmDebug & DBG_TRACE_LOGMETA) {
-			char	strbuf[20];
-			fprintf(stderr, "__pmLogLoadMeta: PMID: %s name: %s\n",
-				pmIDStr_r(dp->pmid, strbuf, sizeof(strbuf)), name);
+			fprintf(stderr, "__pmLogLoadMeta: name[%d] read -> %d: expected: %d\n",
+				i, n, len);
 		    }
 #endif
-
-                    if ((sts = __pmAddPMNSNode(lcp->l_pmns, dp->pmid, name)) < 0) {
-			/*
-			 * If we see a duplicate PMID, its a recoverable error.
-			 * We wont be able to see all of the data in the log, but
-			 * its better to provide access to some rather than none,
-			 * esp. when only one or two metric IDs may be corrupted
-			 * in this way (which we may not be interested in anyway).
-			 */
-			if (sts != PM_ERR_PMID)
-			    goto end;
-			sts = 0;
-		    } 
-		}/*for*/
-            }
+		    if (ferror(f)) {
+			clearerr(f);
+			sts = -oserror();
+		    }
+		    else
+			sts = PM_ERR_LOGREC;
+		    goto end;
+		}
+		name[len] = '\0';
+#ifdef PCP_DEBUG
+		if (pmDebug & DBG_TRACE_LOGMETA) {
+		    char	strbuf[20];
+		    fprintf(stderr, "__pmLogLoadMeta: PMID: %s name: %s\n",
+			    pmIDStr_r(dp->pmid, strbuf, sizeof(strbuf)), name);
+		}
+#endif
+		/* Add the new PMNS node */
+		if ((sts = __pmAddPMNSNode(lcp->l_pmns, dp->pmid, name)) < 0) {
+		    /*
+		     * If we see a duplicate name with a different PMID, its a
+		     * recoverable error.
+		     * We wont be able to see all of the data in the log, but
+		     * its better to provide access to some rather than none,
+		     * esp. when only one or two metric IDs may be corrupted
+		     * in this way (which we may not be interested in anyway).
+		     */
+		    if (sts != PM_ERR_PMID)
+			goto end;
+		    sts = 0;
+		} 
+	    }/*for*/
 	}
 	else if (h.type == TYPE_INDOM) {
 	    int			*tbuf;
@@ -326,11 +505,14 @@ PM_FAULT_POINT("libpcp/" __FILE__ ":4", PM_FAULT_ALLOC);
 		instlist = NULL;
 		namelist = NULL;
 	    }
-	    if ((sts = addindom(lcp, indom, when, numinst, instlist, namelist, tbuf, allinbuf)) < 0) {
-		free(tbuf);
-		if (allinbuf == 0)
-		    free(namelist);
+	    if ((sts = addindom(lcp, indom, when, numinst, instlist, namelist, tbuf, allinbuf)) < 0)
 		goto end;
+	    /* If this indom was a duplicate, then we need to free tbuf and
+	       namelist, as appropriate. */
+	    if (sts == PMLOGPUTINDOM_DUP) {
+		free(tbuf);
+		if (namelist != NULL && !allinbuf)
+		    free(namelist);
 	    }
 	}
 	else
@@ -417,7 +599,7 @@ __pmLogPutDesc(__pmLogCtl *lcp, const pmDesc *dp, int numnames, char **names)
             len += LENSIZE + (int)strlen(names[i]);
     }
 PM_FAULT_POINT("libpcp/" __FILE__ ":10", PM_FAULT_ALLOC);
-    if ((out = (ext_t *)malloc(len)) == NULL)
+    if ((out = (ext_t *)calloc(1, len)) == NULL)
 	return -oserror();
 
     out->hdr.len = htonl(len);
@@ -499,7 +681,7 @@ searchindom(__pmLogCtl *lcp, pmInDom indom, __pmTimeval *tp)
 	    /*
 	     * need first one at or earlier than the requested time
 	     */
-	    if (__pmTimevalSub(&idp->stamp, tp) <= 0)
+	    if (__pmTimevalCmp(&idp->stamp, tp) <= 0)
 		break;
 #ifdef PCP_DEBUG
 	    if (pmDebug & DBG_TRACE_LOGMETA) {
@@ -664,7 +846,6 @@ PM_FAULT_POINT("libpcp/" __FILE__ ":6", PM_FAULT_ALLOC);
     free(out);
 
     sts = addindom(lcp, indom, tp, numinst, instlist, namelist, NULL, 0);
-
     return sts;
 }
 
@@ -767,6 +948,52 @@ pmNameInDomArchive(pmInDom indom, int inst, char **name)
     return n;
 }
 
+/*
+ * Indoms larger than HASH_THRESHOLD will use a hash table
+ * to search the instance and name lists to be returned.
+ * Smaller indoms will use the regular linear search.
+ */
+#define HASH_THRESHOLD 16
+#define HASH_SIZE 509 /* prime */
+
+static struct {
+    int len;
+    int max;
+    int *list;
+} ihash[HASH_SIZE] = {
+    { 0, 0, NULL }
+};
+
+static int
+find_add_ihash(int id)
+{
+    int i = id % HASH_SIZE; 
+    int j;
+
+    for (j=0; j < ihash[i].len; j++) {
+    	if (ihash[i].list[j] == id)
+	    return 1;
+    }
+    ihash[i].len++;
+    if (ihash[i].len >= ihash[i].max) {
+    	ihash[i].max += 8;
+	ihash[i].list = (int *)realloc(ihash[i].list, ihash[i].max * sizeof(int));
+    }
+    ihash[i].list[ihash[i].len-1] = id;
+
+    return 0;
+}
+
+static void
+reset_ihash(void)
+{
+    int i;
+
+    /* invalidate all entries, but don't free the memory */
+    for (i=0; i < HASH_SIZE; i++)
+    	ihash[i].len = 0;
+}
+
 int
 pmGetInDomArchive(pmInDom indom, int **instlist, char ***namelist)
 {
@@ -782,6 +1009,7 @@ pmGetInDomArchive(pmInDom indom, int **instlist, char ***namelist)
     int			*ilist = NULL;
     char		**nlist = NULL;
     char		**olist;
+    int			big_indom = 0;
 
     /* avoid ambiguity when no instances or errors */
     *instlist = NULL;
@@ -803,26 +1031,43 @@ pmGetInDomArchive(pmInDom indom, int **instlist, char ***namelist)
 	    return PM_ERR_INDOM_LOG;
 	}
 
+	for (idp = (__pmLogInDom *)hp->data; idp != NULL; idp = idp->next) {
+	    if (idp->numinst > HASH_THRESHOLD) {
+		big_indom = 1;
+		reset_ihash();
+		break;
+	    }
+	}
+
 	for (idp = (__pmLogInDom *)hp->data; idp != NULL; idp = idp->next) {
 	    for (j = 0; j < idp->numinst; j++) {
-		for (i = 0; i < numinst; i++) {
-		    if (idp->instlist[j] == ilist[i])
-			break;
+		if (big_indom) {
+		    /* big indom - use a hash table */
+		    i = find_add_ihash(idp->instlist[j]) ? 0 : numinst;
 		}
-		if (i == numinst) {
-		    numinst++;
+		else {
+		    /* small indom - linear search */
+		    for (i = 0; i < numinst; i++) {
+			if (idp->instlist[j] == ilist[i])
+			    break;
+		    }
+		}
+
+		if (i < numinst)
+		    continue;
+
+		numinst++;
 PM_FAULT_POINT("libpcp/" __FILE__ ":7", PM_FAULT_ALLOC);
-		    if ((ilist = (int *)realloc(ilist, numinst*sizeof(ilist[0]))) == NULL) {
-			__pmNoMem("pmGetInDomArchive: ilist", numinst*sizeof(ilist[0]), PM_FATAL_ERR);
-		    }
-PM_FAULT_POINT("libpcp/" __FILE__ ":8", PM_FAULT_ALLOC);
-		    if ((nlist = (char **)realloc(nlist, numinst*sizeof(nlist[0]))) == NULL) {
-			__pmNoMem("pmGetInDomArchive: nlist", numinst*sizeof(nlist[0]), PM_FATAL_ERR);
-		    }
-		    ilist[numinst-1] = idp->instlist[j];
-		    nlist[numinst-1] = idp->namelist[j];
-		    strsize += strlen(idp->namelist[j])+1;
+		if ((ilist = (int *)realloc(ilist, numinst*sizeof(ilist[0]))) == NULL) {
+		    __pmNoMem("pmGetInDomArchive: ilist", numinst*sizeof(ilist[0]), PM_FATAL_ERR);
 		}
+PM_FAULT_POINT("libpcp/" __FILE__ ":8", PM_FAULT_ALLOC);
+		if ((nlist = (char **)realloc(nlist, numinst*sizeof(nlist[0]))) == NULL) {
+		    __pmNoMem("pmGetInDomArchive: nlist", numinst*sizeof(nlist[0]), PM_FATAL_ERR);
+		}
+		ilist[numinst-1] = idp->instlist[j];
+		nlist[numinst-1] = idp->namelist[j];
+		strsize += strlen(idp->namelist[j])+1;
 	    }
 	}
 PM_FAULT_POINT("libpcp/" __FILE__ ":9", PM_FAULT_ALLOC);
