@@ -1,7 +1,7 @@
 /*
  * General Utility Routines
  *
- * Copyright (c) 2012-2014 Red Hat.
+ * Copyright (c) 2012-2016 Red Hat.
  * Copyright (c) 2009 Aconex.  All Rights Reserved.
  * Copyright (c) 1995-2002,2004 Silicon Graphics, Inc.  All Rights Reserved.
  * 
@@ -19,11 +19,38 @@
  *
  * pmState - no side-effects, don't bother locking
  *
+ * dosyslog - no side-effects, other than non-determinism with concurrent
+ *	attempts to set/clear the state in __pmSyslog() which locking will
+ *	not avoid
+ *
  * pmProgname - most likely set in main(), not worth protecting here
  * 	and impossible to capture all the read uses in other places
  *
  * base (in __pmProcessDataSize) - no real side-effects, don't bother
  *	locking
+ *
+ * pmprintf_atexit_installed is protected by the util_lock mutex.
+ *
+ * done_exit is protected by the util_lock mutex.
+ *
+ * filelog[] and nfilelog are protected by the util_lock mutex.
+ *
+ * fptr, msgsize, fname and ferr are all protected by the util_lock
+ * mutex.
+ *
+ * the one-trip initialization of xconfirm is guarded by xconfirm_init
+ * ... there is no locking here as the same value would result from
+ * concurrent execution, and we don't want to hold util_lock when
+ * calling pmGetOptionalConfig()
+ *
+ * the one-trip initialization of tmpdir is guarded by tmpdir_init
+ * ... there is no locking here as the same value would result from
+ * concurrent execution, and we don't want to hold util_lock when
+ * calling pmGetOptionalConfig()
+ *
+ * errtype - no side-effects (same value would result from concurrent
+ *	execution of the initialization block, unchanged after that),
+ *	don't bother locking
  */
 
 #include <stdarg.h>
@@ -60,11 +87,33 @@ static int	nfilelog;
 static int	dosyslog;
 static int	pmState = PM_STATE_APPL;
 static int	done_exit;
+#ifdef HAVE_ATEXIT
+static int	pmprintf_atexit_installed = 0;
+#endif
+static int	xconfirm_init = 0;
+static char 	*xconfirm = NULL;
 
-INTERN char	*pmProgname = "pcp";		/* the real McCoy */
+PCP_DATA char	*pmProgname = "pcp";		/* the real McCoy */
 
 static int vpmprintf(const char *, va_list);
 
+#ifdef PM_MULTI_THREAD
+static pthread_mutex_t	util_lock = PTHREAD_MUTEX_INITIALIZER;
+#else
+void			*util_lock;
+#endif
+
+#if defined(PM_MULTI_THREAD) && defined(PM_MULTI_THREAD_DEBUG)
+/*
+ * return true if lock == util_lock
+ */
+int
+__pmIsUtilLock(void *lock)
+{
+    return lock == (void *)&util_lock;
+}
+#endif
+
 /*
  * if onoff == 1, logging is to syslog and stderr, else logging is
  * just to stderr (this is the default)
@@ -72,18 +121,15 @@ static int vpmprintf(const char *, va_list);
 void
 __pmSyslog(int onoff)
 {
-    PM_INIT_LOCKS();
-    PM_LOCK(__pmLock_libpcp);
     dosyslog = onoff;
     if (dosyslog)
 	openlog("pcp", LOG_PID, LOG_DAEMON);
     else
 	closelog();
-    PM_UNLOCK(__pmLock_libpcp);
 }
 
 /*
- * This is a wrapper around syslog(3C) that writes similar messages to stderr,
+ * This is a wrapper around syslog(3) that writes similar messages to stderr,
  * but if __pmSyslog(1) is called, the messages will really go to syslog
  */
 void
@@ -92,14 +138,13 @@ __pmNotifyErr(int priority, const char *message, ...)
     va_list		arg;
     char		*p;
     char		*level;
-    time_t		now;
+    struct timeval	tv;
+    char		ct_buf[26];
 
     va_start(arg, message);
 
-    time(&now);
+    gettimeofday(&tv, NULL);
 
-    PM_INIT_LOCKS();
-    PM_LOCK(__pmLock_libpcp);
     if (dosyslog) {
 	char	syslogmsg[2048];
 
@@ -108,7 +153,6 @@ __pmNotifyErr(int priority, const char *message, ...)
 	va_start(arg, message);
 	syslog(priority, "%s", syslogmsg);
     }
-    PM_UNLOCK(__pmLock_libpcp);
 
     /*
      * do the stderr equivalent
@@ -144,10 +188,11 @@ __pmNotifyErr(int priority, const char *message, ...)
 	    break;
     }
 
-    PM_INIT_LOCKS();
-    PM_LOCK(__pmLock_libpcp);
-    pmprintf("[%.19s] %s(%" FMT_PID ") %s: ", ctime(&now), pmProgname, getpid(), level);
-    PM_UNLOCK(__pmLock_libpcp);
+    ctime_r(&tv.tv_sec, ct_buf);
+    /* when profiling use "[%.19s.%lu]" for extra precision */
+    pmprintf("[%.19s] %s(%" FMT_PID ") %s: ", ct_buf,
+		/* (unsigned long)tv.tv_usec, */
+		pmProgname, getpid(), level);
     vpmprintf(message, arg);
     va_end(arg);
     /* trailing \n if needed */
@@ -161,27 +206,27 @@ __pmNotifyErr(int priority, const char *message, ...)
 static void
 logheader(const char *progname, FILE *log, const char *act)
 {
-    time_t	now;
     char	host[MAXHOSTNAMELEN];
+    time_t	now;
+    char	ct_buf[26];
 
     setlinebuf(log);		/* line buffering for log files */
     gethostname(host, MAXHOSTNAMELEN);
     host[MAXHOSTNAMELEN-1] = '\0';
     time(&now);
-    PM_LOCK(__pmLock_libpcp);
-    fprintf(log, "Log for %s on %s %s %s\n", progname, host, act, ctime(&now));
-    PM_UNLOCK(__pmLock_libpcp);
+    ctime_r(&now, ct_buf);
+    fprintf(log, "Log for %s on %s %s %s\n", progname, host, act, ct_buf);
 }
 
 static void
 logfooter(FILE *log, const char *act)
 {
     time_t	now;
+    char	ct_buf[26];
 
     time(&now);
-    PM_LOCK(__pmLock_libpcp);
-    fprintf(log, "\nLog %s %s", act, ctime(&now));
-    PM_UNLOCK(__pmLock_libpcp);
+    ctime_r(&now, ct_buf);
+    fprintf(log, "\nLog %s %s", act, ct_buf);
 }
 
 static void
@@ -189,16 +234,16 @@ logonexit(void)
 {
     int		i;
 
-    PM_LOCK(__pmLock_libpcp);
+    PM_LOCK(util_lock);
     if (++done_exit != 1) {
-	PM_UNLOCK(__pmLock_libpcp);
+	PM_UNLOCK(util_lock);
 	return;
     }
 
     for (i = 0; i < nfilelog; i++)
 	logfooter(filelog[i], "finished");
 
-    PM_UNLOCK(__pmLock_libpcp);
+    PM_UNLOCK(util_lock);
 }
 
 /* common code shared by __pmRotateLog and __pmOpenLog */
@@ -209,6 +254,7 @@ logreopen(const char *progname, const char *logname, FILE *oldstream,
     int		oldfd;
     int		dupoldfd;
     FILE	*dupoldstream = oldstream;
+    char	errmsg[PM_MAXERRMSGLEN];
 
     /*
      * Do our own version of freopen() because the standard one closes the
@@ -231,7 +277,7 @@ logreopen(const char *progname, const char *logname, FILE *oldstream,
 
 	oldstream = freopen(logname, "w", oldstream);
 	if (oldstream == NULL) {
-	    int	save_error = oserror();	/* need for error message */
+	    int		save_error = oserror();	/* need for error message */
 
 	    close(oldfd);
 	    if (dup(dupoldfd) != oldfd) {
@@ -266,8 +312,22 @@ logreopen(const char *progname, const char *logname, FILE *oldstream,
 		/* put oldstream back for return value */
 		oldstream = dupoldstream;
 	    }
+#ifdef HAVE_STRERROR_R_PTR
+	    {
+		char	*p;
+		p = strerror_r(save_error, errmsg, sizeof(errmsg));
+		if (p != errmsg)
+		    strncpy(errmsg, p, sizeof(errmsg));
+	    }
+#else
+	    /*
+	     * the more normal POSIX and XSI compliant variants always
+	     * fill the message buffer
+	     */
+	    strerror_r(save_error, errmsg, sizeof(errmsg));
+#endif
 	    pmprintf("%s: cannot open log \"%s\" for writing : %s\n",
-		    progname, logname, strerror(save_error));
+		    progname, logname, errmsg);
 	    pmflush();
 	}
 	else {
@@ -278,7 +338,7 @@ logreopen(const char *progname, const char *logname, FILE *oldstream,
     }
     else {
 	pmprintf("%s: cannot redirect log output to \"%s\": %s\n",
-		progname, logname, strerror(errno));
+		progname, logname, osstrerror_r(errmsg, sizeof(errmsg)));
 	pmflush();
     }
     return oldstream;
@@ -289,21 +349,22 @@ __pmOpenLog(const char *progname, const char *logname, FILE *oldstream,
 	    int *status)
 {
     oldstream = logreopen(progname, logname, oldstream, status);
-    PM_INIT_LOCKS();
     logheader(progname, oldstream, "started");
 
-    PM_LOCK(__pmLock_libpcp);
+    PM_LOCK(util_lock);
     nfilelog++;
     if (nfilelog == 1)
 	atexit(logonexit);
 
     filelog = (FILE **)realloc(filelog, nfilelog * sizeof(FILE *));
     if (filelog == NULL) {
+	PM_UNLOCK(util_lock);
 	__pmNoMem("__pmOpenLog", nfilelog * sizeof(FILE *), PM_FATAL_ERR);
+	/* NOTREACHED */
     }
     filelog[nfilelog-1] = oldstream;
 
-    PM_UNLOCK(__pmLock_libpcp);
+    PM_UNLOCK(util_lock);
     return oldstream;
 }
 
@@ -312,40 +373,47 @@ __pmRotateLog(const char *progname, const char *logname, FILE *oldstream,
 	    int *status)
 {
     int		i;
+    FILE	*newstream = oldstream;
 
-    PM_INIT_LOCKS();
-    PM_LOCK(__pmLock_libpcp);
+    PM_LOCK(util_lock);
     for (i = 0; i < nfilelog; i++) {
-	if (oldstream == filelog[i]) {
-	    logfooter(oldstream, "rotated");	/* old */
-	    oldstream = logreopen(progname, logname, oldstream, status);
-	    logheader(progname, oldstream, "rotated");	/* new */
-	    filelog[i] = oldstream;
+	if (oldstream == filelog[i])
 	    break;
+    }
+    if (i < nfilelog) {
+	PM_UNLOCK(util_lock);
+	logfooter(oldstream, "rotated");	/* old */
+	newstream = logreopen(progname, logname, oldstream, status);
+	logheader(progname, newstream, "rotated");	/* new */
+	PM_LOCK(util_lock);
+	for (i = 0; i < nfilelog; i++) {
+	    if (oldstream == filelog[i]) {
+		filelog[i] = newstream;
+		break;
+	    }
 	}
     }
-    PM_UNLOCK(__pmLock_libpcp);
-    return oldstream;
+    PM_UNLOCK(util_lock);
+    return newstream;
 }
 
 /* pmID -> string, max length is 20 bytes */
 char *
 pmIDStr_r(pmID pmid, char *buf, int buflen)
 {
-    __pmID_int*	p = (__pmID_int*)&pmid;
     if (pmid == PM_ID_NULL)
 	snprintf(buf, buflen, "%s", "PM_ID_NULL");
-    else if (p->domain == DYNAMIC_PMID && p->item == 0)
+    else if (IS_DYNAMIC_ROOT(pmid))
 	/*
 	 * this PMID represents the base of a dynamic subtree in the PMNS
 	 * ... identified by setting the domain field to the reserved
 	 * value DYNAMIC_PMID and storing the real domain of the PMDA
 	 * that can enumerate the subtree in the cluster field, while
-	 * the item field is not used
+	 * the item field is zero
 	 */
-	snprintf(buf, buflen, "%d.*.*", p->cluster);
+	snprintf(buf, buflen, "%d.*.*", pmid_cluster(pmid));
     else
-	snprintf(buf, buflen, "%d.%d.%d", p->domain, p->cluster, p->item);
+	snprintf(buf, buflen, "%d.%d.%d", pmid_domain(pmid), pmid_cluster(pmid), pmid_item(pmid));
     return buf;
 }
 
@@ -471,6 +539,36 @@ pmEventFlagsStr(int flags)
     return ebuf;
 }
 
+char *
+pmSemStr_r(int sem, char *buf, int buflen)
+{
+    switch (sem) {
+	case PM_SEM_COUNTER:
+	    strncpy (buf, "counter", buflen);
+	    break;
+	case PM_SEM_INSTANT:
+	    strncpy (buf, "instant", buflen);
+	    break;
+	case PM_SEM_DISCRETE:
+	    strncpy (buf, "discrete", buflen);
+	    break;
+	default:
+	    strncpy (buf, "???", buflen);
+	    break;
+    }
+    buf[buflen - 1] = '\0';
+    return buf;
+}
+
+const char *
+pmSemStr(int sem)
+{
+    static char ebuf[64];
+    pmSemStr_r(sem, ebuf, sizeof(ebuf));
+    return ebuf;
+}
+
+
 /*
  * Several PMAPI interfaces allocate a list of strings into a buffer
  * pointed to by (char **) which can be safely freed simply by
@@ -578,16 +676,25 @@ dump_valueset(FILE *f, pmValueSet *vsp)
     pmDesc	desc;
     char	errmsg[PM_MAXERRMSGLEN];
     char	strbuf[20];
-    char	*pmid, *p;
+    char	*pmid;
+    char	*p;
+    char	**names;
     int		have_desc = 1;
     int		n, j;
 
     pmid = pmIDStr_r(vsp->pmid, strbuf, sizeof(strbuf));
-    if ((n = pmNameID(vsp->pmid, &p)) < 0)
-	fprintf(f,"  %s (%s):", pmid, "<noname>");
+    if ((n = pmNameAll(vsp->pmid, &names)) < 0)
+	fprintf(f, "  %s (%s):", pmid, "<noname>");
     else {
-	fprintf(f,"  %s (%s):", pmid, p);
-	free(p);
+	int	j;
+	fprintf(f, "  %s (", pmid);
+	for (j = 0; j < n; j++) {
+	    if (j > 0)
+		fprintf(f, ", ");
+	    fprintf(f, "%s", names[j]);
+	}
+	fprintf(f, "):");
+	free(names);
     }
     if (vsp->numval == 0) {
 	fprintf(f, " No values returned!\n");
@@ -609,7 +716,7 @@ dump_valueset(FILE *f, pmValueSet *vsp)
     for (j = 0; j < vsp->numval; j++) {
 	pmValue	*vp = &vsp->vlist[j];
 	if (vsp->numval > 1 || vp->inst != PM_INDOM_NULL) {
-	    fprintf(f,"    inst [%d", vp->inst);
+	    fprintf(f, "    inst [%d", vp->inst);
 	    if (have_desc &&
 		pmNameInDom(desc.indom, vp->inst, &p) >= 0) {
 		fprintf(f, " or \"%s\"]", p);
@@ -1009,6 +1116,20 @@ __pmGetTimespec(struct timespec *ts)
 #endif
 }
 
+/*
+ * a : b for __pmTimeval ... <0 for a<b, ==0 for a==b, >0 for a>b
+ */
+int
+__pmTimevalCmp(const __pmTimeval *a, const __pmTimeval *b)
+{
+    int res = (int)(a->tv_sec - b->tv_sec);
+
+    if (res == 0)
+	res = (int)(a->tv_usec - b->tv_usec);
+
+    return res;
+}
+
 /*
  * Difference for two of the internal timestamps ...
  * Same as __pmtimevalSub() in tv.c, just with __pmTimeval args
@@ -1137,25 +1258,11 @@ __pmPrintDesc(FILE *f, const pmDesc *desc)
     if (type == unknownVal)
 	fprintf(f, " (%d)", desc->type);
 
-    fprintf(f,"  InDom: %s 0x%x\n", pmInDomStr_r(desc->indom, strbuf, sizeof(strbuf)), desc->indom);
-
-    switch (desc->sem) {
-	case PM_SEM_COUNTER:
-	    sem = "counter";
-	    break;
-	case PM_SEM_INSTANT:
-	    sem = "instant";
-	    break;
-	case PM_SEM_DISCRETE:
-	    sem = "discrete";
-	    break;
-	default:
-	    sem = unknownVal;
-	    break;
-    }
+    fprintf(f, "  InDom: %s 0x%x\n", pmInDomStr_r(desc->indom, strbuf, sizeof(strbuf)), desc->indom);
 
+    sem = pmSemStr_r(desc->sem, strbuf, sizeof(strbuf));
     fprintf(f, "    Semantics: %s", sem);
-    if (sem == unknownVal)
+    if (strcmp(sem, unknownVal) == 0)
 	fprintf(f, " (%d)", desc->sem);
 
     fprintf(f, "  Units: ");
@@ -1286,8 +1393,8 @@ __pmSetInternalState(int state)
  */
 
 #define MSGBUFLEN	256
-static FILE	*fptr = NULL;
-static int	msgsize = 0;
+static FILE	*fptr;
+static int	msgsize;
 static char	*fname;		/* temporary file name for buffering errors */
 static char	*ferr;		/* error output filename from PCP_STDERR */
 
@@ -1296,26 +1403,36 @@ static char	*ferr;		/* error output filename from PCP_STDERR */
 #define PM_USESTDERR       1
 #define PM_USEFILE         2
 
+/*
+ * called with util_lock held
+ */
 static int
 pmfstate(int state)
 {
-    static int	errtype = -1;
+    static int	errtype = PM_QUERYERR;
+    char	errmsg[PM_MAXERRMSGLEN];
+
+    ASSERT_IS_LOCKED(util_lock);
 
     if (state > PM_QUERYERR)
 	errtype = state;
 
-    PM_INIT_LOCKS();
-    PM_LOCK(__pmLock_libpcp);
     if (errtype == PM_QUERYERR) {
+	/* one-trip initialization */
 	errtype = PM_USESTDERR;
-	if ((ferr = getenv("PCP_STDERR")) != NULL) {
+	PM_LOCK(__pmLock_extcall);
+	ferr = getenv("PCP_STDERR");		/* THREADSAFE */
+	if (ferr != NULL)
+	    ferr = strdup(ferr);
+	PM_UNLOCK(__pmLock_extcall);
+	if (ferr != NULL) {
 	    if (strcasecmp(ferr, "DISPLAY") == 0) {
-		char * xconfirm = pmGetConfig("PCP_XCONFIRM_PROG");
-		if (access(__pmNativePath(xconfirm), X_OK) < 0) {
-		    char	errmsg[PM_MAXERRMSGLEN];
+		if (!xconfirm)
+		    fprintf(stderr, "%s: using stderr - no PCP_XCONFIRM_PROG\n",
+			    pmProgname);
+		else if (access(__pmNativePath(xconfirm), X_OK) < 0)
 		    fprintf(stderr, "%s: using stderr - cannot access %s: %s\n",
 			    pmProgname, xconfirm, osstrerror_r(errmsg, sizeof(errmsg)));
-		}
 		else
 		    errtype = PM_USEDIALOG;
 	    }
@@ -1323,22 +1440,48 @@ pmfstate(int state)
 		errtype = PM_USEFILE;
 	}
     }
-    PM_UNLOCK(__pmLock_libpcp);
     return errtype;
 }
 
+#ifdef HAVE_ATEXIT
+static void
+pmprintf_cleanup(void)
+{
+    if (fptr != NULL) {
+	unlink(fname);
+	free(fname);
+	fclose(fptr);
+	fptr = NULL;
+    }
+}
+#endif
+
 static int
 vpmprintf(const char *msg, va_list arg)
 {
     int		lsize = 0;
+    static int	tmpdir_init = 0;
+    static char	*tmpdir;
 
-    PM_INIT_LOCKS();
-    PM_LOCK(__pmLock_libpcp);
+    /* see thread-safe notes above */
+    if (!tmpdir_init) {
+	tmpdir = pmGetOptionalConfig("PCP_TMPFILE_DIR");
+	tmpdir_init = 1;
+    }
+
+    PM_LOCK(util_lock);
     if (fptr == NULL && msgsize == 0) {		/* create scratch file */
 	int	fd = -1;
-	char	*tmpdir = pmGetConfig("PCP_TMPFILE_DIR");
 
-	if (tmpdir[0] != '\0') {
+#ifdef HAVE_ATEXIT
+	if (pmprintf_atexit_installed == 0) {
+	    /* install handler for temp file cleanup */
+	    atexit(pmprintf_cleanup);
+	    pmprintf_atexit_installed = 1;
+	}
+#endif
+
+	if (tmpdir && tmpdir[0] != '\0') {
 	    mode_t cur_umask;
 
 	    /*
@@ -1394,7 +1537,7 @@ fail:
     else
 	msgsize += (lsize = vfprintf(fptr, msg, arg));
 
-    PM_UNLOCK(__pmLock_libpcp);
+    PM_UNLOCK(util_lock);
     return lsize;
 }
 
@@ -1418,15 +1561,20 @@ pmflush(void)
     int		state;
     FILE	*eptr = NULL;
     char	outbuf[MSGBUFLEN];
+    char	errmsg[PM_MAXERRMSGLEN];
 
-    PM_INIT_LOCKS();
-    PM_LOCK(__pmLock_libpcp);
+    /* see thread-safe notes above */
+    if (!xconfirm_init) {
+	xconfirm = pmGetOptionalConfig("PCP_XCONFIRM_PROG");
+	xconfirm_init = 1;
+    }
+
+    PM_LOCK(util_lock);
     if (fptr != NULL && msgsize > 0) {
 	fflush(fptr);
 	state = pmfstate(PM_QUERYERR);
 	if (state == PM_USEFILE) {
 	    if ((eptr = fopen(ferr, "a")) == NULL) {
-		char	errmsg[PM_MAXERRMSGLEN];
 		fprintf(stderr, "pmflush: cannot append to file '%s' (from "
 			"$PCP_STDERR): %s\n", ferr, osstrerror_r(errmsg, sizeof(errmsg)));
 		state = PM_USESTDERR;
@@ -1438,7 +1586,6 @@ pmflush(void)
 	    while ((len = (int)read(fileno(fptr), outbuf, MSGBUFLEN)) > 0) {
 		sts = write(fileno(stderr), outbuf, len);
 		if (sts != len) {
-		    char	errmsg[PM_MAXERRMSGLEN];
 		    fprintf(stderr, "pmflush: write() failed: %s\n", 
 			osstrerror_r(errmsg, sizeof(errmsg)));
 		}
@@ -1447,12 +1594,17 @@ pmflush(void)
 	    break;
 	case PM_USEDIALOG:
 	    /* If we're here, it means xconfirm has passed access test */
+	    if (xconfirm == NULL) {
+		fprintf(stderr, "%s: no PCP_XCONFIRM_PROG\n", pmProgname);
+		sts = PM_ERR_GENERIC;
+		break;
+	    }
 	    snprintf(outbuf, sizeof(outbuf), "%s -file %s -c -B OK -icon info"
 		    " %s -header 'PCP Information' >/dev/null",
-		    __pmNativePath(pmGetConfig("PCP_XCONFIRM_PROG")), fname,
+		    __pmNativePath(xconfirm), fname,
 		    (msgsize > 80 ? "-useslider" : ""));
-	    if (system(outbuf) < 0) {
-		char	errmsg[PM_MAXERRMSGLEN];
+	    /* no thread-safe issue here ... we're executing xconfirm */
+	    if (system(outbuf) < 0) {		/* THREADSAFE */
 		fprintf(stderr, "%s: system failed: %s\n", pmProgname,
 			osstrerror_r(errmsg, sizeof(errmsg)));
 		sts = -oserror();
@@ -1463,7 +1615,6 @@ pmflush(void)
 	    while ((len = (int)read(fileno(fptr), outbuf, MSGBUFLEN)) > 0) {
 		sts = write(fileno(eptr), outbuf, len);
 		if (sts != len) {
-		    char	errmsg[PM_MAXERRMSGLEN];
 		    fprintf(stderr, "pmflush: write() failed: %s\n", 
 			osstrerror_r(errmsg, sizeof(errmsg)));
 		}
@@ -1472,17 +1623,17 @@ pmflush(void)
 	    fclose(eptr);
 	    break;
 	}
-	fclose(fptr);
-	fptr = NULL;
 	unlink(fname);
 	free(fname);
+	fclose(fptr);
+	fptr = NULL;
 	if (sts >= 0)
 	    sts = msgsize;
     }
 
     msgsize = 0;
 
-    PM_UNLOCK(__pmLock_libpcp);
+    PM_UNLOCK(util_lock);
     return sts;
 }
 
@@ -1500,7 +1651,7 @@ __pmSetClientId(const char *id)
     char		*name = "pmcd.client.whoami";
     pmID		pmid;
     int			sts;
-    pmResult		store = { .numpmid = 1 };
+    pmResult		store;
     pmValueSet		pmvs;
     pmValueBlock	*pmvb;
     char        	host[MAXHOSTNAMELEN];
@@ -1560,7 +1711,7 @@ __pmSetClientId(const char *id)
     vblen += strlen(host) + strlen(id) + 2;
 
     /* build pmResult for pmStore() */
-    pmvb = (pmValueBlock *)malloc(PM_VAL_HDR_SIZE+vblen);
+    pmvb = (pmValueBlock *)calloc(1, PM_VAL_HDR_SIZE+vblen);
     if (pmvb == NULL) {
 	__pmNoMem("__pmSetClientId", PM_VAL_HDR_SIZE+vblen, PM_RECOV_ERR);
 	return -ENOMEM;
@@ -1583,6 +1734,8 @@ __pmSetClientId(const char *id)
     pmvs.vlist[0].value.pval = pmvb;
     pmvs.vlist[0].inst = PM_IN_NULL;
 
+    memset(&store, 0, sizeof(store));
+    store.numpmid = 1;
     store.vset[0] = &pmvs;
     sts = pmStore(&store);
     free(pmvb);
@@ -1715,7 +1868,7 @@ strchrnul(const char *s, int c)
 #ifndef HAVE_SCANDIR
 /*
  * Scan the directory dirname, building an array of pointers to
- * dirent entries using malloc(3C).  select() and compare() are
+ * dirent entries using malloc(3).  select() and compare() are
  * used to optionally filter and sort directory entries.
  */
 int
@@ -1729,26 +1882,30 @@ scandir(const char *dirname, struct dirent ***namelist,
     struct dirent	*dp;
     struct dirent	*tp;
 
-    PM_INIT_LOCKS();
-    PM_LOCK(__pmLock_libpcp);
     if ((dirp = opendir(dirname)) == NULL)
 	return -1;
 
+    /* dirp is an on-stack variable, so readdir() is ... */
+    /* THREADSAFE */
     while ((dp = readdir(dirp)) != NULL) {
 	if (select && (*select)(dp) == 0)
 	    continue;
 
 	n++;
 	if ((names = (struct dirent **)realloc(names, n * sizeof(dp))) == NULL) {
-	    PM_UNLOCK(__pmLock_libpcp);
 	    closedir(dirp);
 	    return -1;
 	}
 
 	if ((names[n-1] = tp = (struct dirent *)malloc(
 		sizeof(*dp)-sizeof(dp->d_name)+strlen(dp->d_name)+1)) == NULL) {
-	    PM_UNLOCK(__pmLock_libpcp);
 	    closedir(dirp);
+	    n--;
+	    while (n >= 1) {
+		free(names[n-1]);
+		n--;
+	    }
+	    free(names);
 	    return -1;
 	}
 
@@ -1761,7 +1918,6 @@ scandir(const char *dirname, struct dirent ***namelist,
 	memcpy(tp->d_name, dp->d_name, strlen(dp->d_name)+1);
     }
     closedir(dirp);
-    PM_UNLOCK(__pmLock_libpcp);
     *namelist = names;
 
     if (n && compare)
@@ -2120,7 +2276,7 @@ __pmProcessExists(pid_t pid)
        return 0;
     return (len > 0);
 }
-#elif defined(IS_FREEBSD)
+#elif defined(IS_FREEBSD) || defined(IS_OPENBSD)
 int 
 __pmProcessExists(pid_t pid)
 {
@@ -2160,12 +2316,12 @@ __pmProcessTerminate(pid_t pid, int force)
 int
 __pmProcessDataSize(unsigned long *size)
 {
-    static void *base;
+    static char *base;
 
     if (size && base)
-	*size = (sbrk(0) - base) / 1024;
+	*size = ((char *)sbrk(0) - base) / 1024;
     else {
-	base = sbrk(0);
+	base = (char *)sbrk(0);
 	if (size)
 	    *size = 0;
     }
@@ -2309,11 +2465,10 @@ __pmDumpStack(FILE *f)
 	call_fn[i] = names[i];
     res = trace_back_stack(MAX_DEPTH, call_addr, call_fn, MAX_DEPTH, MAX_SIZE);
     for (i = 1; i < res; i++) {
-#if defined(HAVE_64BIT_PTR)
-	fprintf(f, "  0x%016llx [%s]\n", call_addr[i], call_fn[i]);
-#else
-	fprintf(f, "  0x%08lx [%s]\n", (__uint32_t)call_addr[i], call_fn[i]);
-#endif
+	if (sizeof(void *) == sizeof(long long))
+	    fprintf(f, "  0x%016llx [%s]\n", call_addr[i], call_fn[i]);
+	else
+	    fprintf(f, "  0x%08lx [%s]\n", (__uint32_t)call_addr[i], call_fn[i]);
     }
 }
 
