@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2012-2014 Red Hat.
+ * Copyright (c) 2012-2016 Red Hat.
  * Copyright (c) 1995-2000,2003,2004 Silicon Graphics, Inc.  All Rights Reserved.
  *
  * This library is free software; you can redistribute it and/or modify it
@@ -17,6 +17,10 @@
 #include "impl.h"
 #include "pmda.h"
 #include "libdefs.h"
+#include <sys/stat.h>
+#ifdef HAVE_PWD_H
+#include <pwd.h>
+#endif
 #ifdef HAVE_SYS_UN_H
 #include <sys/un.h>
 #endif
@@ -150,6 +154,44 @@ __pmdaOpenIPv6(char *sockname, int port, int *infd, int *outfd)
 }
 
 #ifdef HAVE_STRUCT_SOCKADDR_UN
+/*
+ * Setup appropriate permissions for the Unix domain socket file
+ *
+ * pmcd must be able to talk to us on this channel, and it may be running
+ * as an unprivileged user by the time the PMDA runs this code - so, need
+ * to set permissions such that PCP_USER can access this file too.
+ *
+ * Note when run from dbpmda and/or QA these permissions might be fine to
+ * stay as regular users, so we do not fail here and simply charge ahead.
+ */
+static void
+socket_ownership(char *sockname)
+{
+    char		*username = pmGetConfig("PCP_USER");
+    char		errmsg[128];
+    struct passwd	*pw;
+    int			sts;
+
+    setoserror(0);
+    if ((pw = getpwnam(username)) == NULL)
+	__pmNotifyErr(LOG_WARNING, "__pmdaOpenUnix: getpwnam(%s) failed: %s\n",
+		username, pmErrStr_r(-oserror(), errmsg, sizeof(errmsg)));
+
+    sts = chmod(sockname, S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP);
+    if (sts == -1 && (pmDebug & DBG_TRACE_LIBPMDA))
+	__pmNotifyErr(LOG_WARNING, "__pmdaOpenUnix: chmod(%s,...) failed: %s\n",
+			sockname, osstrerror());
+    if (pw != NULL) {
+	sts = chown(sockname, pw->pw_uid, pw->pw_gid);
+	if (sts == -1 && (pmDebug & DBG_TRACE_LIBPMDA))
+	    __pmNotifyErr(LOG_WARNING, "__pmdaOpenUnix: chown(%s, ...) failed: %s\n",
+			    sockname, osstrerror());
+    }
+    else if (pmDebug & DBG_TRACE_LIBPMDA)
+	    __pmNotifyErr(LOG_WARNING, "__pmdaOpenUnix: chown(%s, ...) skipped\n",
+			    sockname);
+}
+
 /*
  * Open a unix port to PMCD
  */
@@ -195,6 +237,7 @@ __pmdaOpenUnix(char *sockname, int *infd, int *outfd)
 			netstrerror());
 	exit(1);
     }
+    socket_ownership(sockname);
 
     sts = listen(sfd, 5);	/* Max. of 5 pending connection requests */
     if (sts == -1) {
@@ -352,7 +395,7 @@ pmdaGetOptions(int argc, char *const *argv, pmdaOptions *opts, pmdaInterface *di
 	    break;
 
 	case 'h':	/* over-ride default help file */
-	    pmda->e_helptext = opts->optarg;
+	    pmda->e_helptext = strdup(opts->optarg);
 	    break;
 
 	case 'i':
@@ -369,7 +412,7 @@ pmdaGetOptions(int argc, char *const *argv, pmdaOptions *opts, pmdaInterface *di
 	    break;
 
 	case 'l':	/* over-ride default log file */
-	    pmda->e_logfile = opts->optarg;
+	    pmda->e_logfile = strdup(opts->optarg);
 	    break;
 
 	case 'p':
@@ -473,11 +516,12 @@ pmdaRehash(pmdaExt *pmda, pmdaMetric *metrics, int nmetrics)
 	    break;
 	}
     }
-    if (m == pmda->e_nmetrics)
+    if (m == pmda->e_nmetrics) {
 	pmda->e_flags |= PMDA_EXT_FLAG_HASHED;
 	if (pmDebug & DBG_TRACE_LIBPMDA)
 	    __pmNotifyErr(LOG_DEBUG, "pmdaRehash: PMDA %s: successful rebuild\n",
 			pmda->e_name);
+    }
     else {
 	pmda->e_flags &= ~PMDA_EXT_FLAG_HASHED;
 	pmdaHashDelete(hashp);
@@ -652,7 +696,7 @@ pmdaInit(pmdaInterface *dispatch, pmdaIndom *indoms, int nindoms,
     }
     else {
 	if (dispatch->version.two.text == pmdaText)
-	    __pmNotifyErr(LOG_WARNING, "pmdaInit: PMDA %s: No help text file specified", pmda->e_name); 
+	    __pmNotifyErr(LOG_WARNING, "pmdaInit: PMDA %s: No help text file specified for pmdaText", pmda->e_name); 
 #ifdef PCP_DEBUG
 	else
 	    if (pmDebug & DBG_TRACE_LIBPMDA)
@@ -696,7 +740,7 @@ pmdaInit(pmdaInterface *dispatch, pmdaIndom *indoms, int nindoms,
  */
 
 static int
-__pmdaSetupPDU(int infd, int outfd, int flags, char *agentname)
+__pmdaSetupPDU(int infd, int outfd, int flags, const char *agentname)
 {
     __pmVersionCred	handshake;
     __pmCred		*credlist = NULL;
@@ -855,7 +899,7 @@ pmdaConnect(pmdaInterface *dispatch)
  */
 
 static void
-__pmdaSetup(pmdaInterface *dispatch, int version, char *name)
+__pmdaSetup(pmdaInterface *dispatch, int version, const char *name)
 {
     pmdaExt	*pmda = NULL;
     e_ext_t	*extp;
@@ -920,7 +964,7 @@ __pmdaSetup(pmdaInterface *dispatch, int version, char *name)
 	dispatch->status = PM_ERR_GENERIC;
 	return;
     }
-    extp->pmda_interface = version;
+    extp->dispatch = dispatch;
     pmda->e_ext = (void *)extp;
 
     pmdaSetResultCallBack(dispatch, __pmFreeResultValues);
@@ -936,8 +980,8 @@ __pmdaSetup(pmdaInterface *dispatch, int version, char *name)
  */
 
 void
-pmdaDaemon(pmdaInterface *dispatch, int version, char *name, int domain, 
-	   char *logfile, char *helptext)
+pmdaDaemon(pmdaInterface *dispatch, int version, const char *name, int domain,
+	   const char *logfile, const char *helptext)
 {
     pmdaExt	*pmda;
 
@@ -948,8 +992,8 @@ pmdaDaemon(pmdaInterface *dispatch, int version, char *name, int domain,
 	return;
 
     pmda = dispatch->version.any.ext;
-    pmda->e_logfile = logfile;
-    pmda->e_helptext = helptext;
+    pmda->e_logfile = (logfile == NULL ? NULL : strdup(logfile));
+    pmda->e_helptext = (helptext == NULL ? NULL : strdup(helptext));
 
     __pmSetInternalState(PM_STATE_PMCS);
 }
@@ -967,7 +1011,7 @@ pmdaDSO(pmdaInterface *dispatch, int version, char *name, char *helptext)
     if (dispatch->status < 0)
 	return;
 
-    dispatch->version.any.ext->e_helptext = helptext;
+    dispatch->version.any.ext->e_helptext = (helptext == NULL ? NULL : strdup(helptext));
 }
 
 /*
@@ -985,3 +1029,18 @@ pmdaOpenLog(pmdaInterface *dispatch)
     __pmOpenLog(dispatch->version.any.ext->e_name, 
 		dispatch->version.any.ext->e_logfile, stderr, &c);
 }
+
+/*
+ * pmdaInterface was moved ... fix e_ext back pointer
+ */
+void
+pmdaInterfaceMoved(pmdaInterface *dispatch)
+{
+    if (dispatch->version.any.ext != NULL) {
+	if (dispatch->version.any.ext->e_ext != NULL) {
+	    e_ext_t	*extp;
+	    extp = (e_ext_t *)dispatch->version.any.ext->e_ext;
+	    extp->dispatch = dispatch;
+	}
+    }
+}
