@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2012-2013 Red Hat.
+ * Copyright (c) 2012-2014,2017 Red Hat.
  * Copyright (c) 1995-2002 Silicon Graphics, Inc.  All Rights Reserved.
  * 
  * This program is free software; you can redistribute it and/or modify it
@@ -133,42 +133,18 @@ DoText(ClientInfo *cp, __pmPDU* pb)
 int
 DoProfile(ClientInfo *cp, __pmPDU *pb)
 {
+    __pmHashCtl	*hcp;
     __pmProfile	*newProf;
     int		ctxnum, sts, i;
 
     sts = __pmDecodeProfile(pb, &ctxnum, &newProf);
     if (sts >= 0) {
-	/* Allocate more profile pointers if required */
-	if (ctxnum >= cp->szProfile) {
-	    __pmProfile	**newProfPtrs;
-	    int		need, oldSize = cp->szProfile;
-
-	    if (ctxnum - cp->szProfile < 4)
-		cp->szProfile += 4;
-	    else
-		cp->szProfile = ctxnum + 1;
-	    need = cp->szProfile * (int)sizeof(__pmProfile *);
-	    if ((newProfPtrs = (__pmProfile **)malloc(need)) == NULL) {
-		cp->szProfile = oldSize;
-		__pmNoMem("DoProfile.newProfPtrs", need, PM_RECOV_ERR);
-		__pmFreeProfile(newProf);
-		return -oserror();
-	    }
-
-	    /* Copy any old pointers and zero the newly allocated ones */
-	    if ((need = oldSize * (int)sizeof(__pmProfile *))) {
-		memcpy(newProfPtrs, cp->profile, need);
-		free(cp->profile);	/* But not the __pmProfile ptrs! */
-	    }
-	    need = (cp->szProfile - oldSize) * (int)sizeof(__pmProfile *);
-	    memset(&newProfPtrs[oldSize], 0, need);
-	    cp->profile = newProfPtrs;
+	hcp = &cp->profile;
+	if ((sts = __pmHashAdd(ctxnum, newProf, hcp)) > 0) {
+	    /* __pmHashAdd returns 1 for success, but we want zero. */
+	    sts = 0;
 	}
-	else				/* cp->profile is big enough */
-	    if (cp->profile[ctxnum] != NULL)
-		__pmFreeProfile(cp->profile[ctxnum]);
-	cp->profile[ctxnum] = newProf;
-
+    
 	/* "Invalidate" any references to the client context's profile in the
 	 * agents to which the old profile was last sent
 	 */
@@ -257,7 +233,7 @@ DoDesc(ClientInfo *cp, __pmPDU *pb)
 }
 
 int
-DoInstance(ClientInfo *cp, __pmPDU* pb)
+DoInstance(ClientInfo *cp, __pmPDU *pb)
 {
     int			sts, s;
     __pmTimeval		when;
@@ -272,15 +248,8 @@ DoInstance(ClientInfo *cp, __pmPDU* pb)
     if (sts < 0)
 	return sts;
     if (when.tv_sec != 0 || when.tv_usec != 0) {
-	/*
-	 * we have no idea how to do anything but current, yet!
-	 *
-	 * TODO EXCEPTION PCP 2.0 ...
-	 * this may be left over from the pmvcr days, and can be tossed?
-	 * ... leaving it here is benign
-	 */
 	if (name != NULL) free(name);
-	return PM_ERR_NYI;
+	return PM_ERR_IPC;
     }
     if ((ap = FindDomainAgent(((__pmInDom_int *)&indom)->domain)) == NULL) {
 	if (name != NULL) free(name);
@@ -457,8 +426,11 @@ fail:
 int
 DoPMNSNames(ClientInfo *cp, __pmPDU *pb)
 {
-    int		sts = 0;
+    int		sts;
     int		numids = 0;
+    int		numok;
+    int		lsts;
+    int		domain;
     pmID	*idlist = NULL;
     char	**namelist = NULL;
     int		i;
@@ -473,25 +445,29 @@ DoPMNSNames(ClientInfo *cp, __pmPDU *pb)
     }
 
     sts = pmLookupName(numids, namelist, idlist);
+    /*
+     * even if this fails, or looks up fewer than numids, we have to
+     * check each PMID looking for dynamic metrics and process them
+     * separately with the help of the PMDA, if possible
+     */
     for (i = 0; i < numids; i++) {
-	if (idlist[i] == PM_ID_NULL) continue;
-	if (pmid_domain(idlist[i]) == DYNAMIC_PMID && pmid_item(idlist[i]) == 0) {
-	    int		lsts;
-	    int		domain = pmid_cluster(idlist[i]);
-	    /*
-	     * don't return <domain>.*.* ... all return paths from here
-	     * must either set a valid PMID in idlist[i] or indicate
-	     * the first error in the return from pmLookupName
-	     */
-	    idlist[i] = PM_ID_NULL;	/* default case if cannot translate */
-	    if ((ap = FindDomainAgent(domain)) == NULL) {
-		if (sts > 0) sts = PM_ERR_NOAGENT;
-		continue;
-	    }
-	    if (!ap->status.connected) {
-		if (sts > 0) sts = PM_ERR_NOAGENT;
-		continue;
-	    }
+	if (idlist[i] == PM_ID_NULL || !IS_DYNAMIC_ROOT(idlist[i]))
+	    continue;
+	lsts = 0;
+	domain = pmid_cluster(idlist[i]);
+	/*
+	 * don't return <domain>.*.* ... all return paths from here
+	 * must either set a valid PMID in idlist[i] or indicate
+	 * an error via lsts
+	 */
+	idlist[i] = PM_ID_NULL;	/* default case if cannot translate */
+	if ((ap = FindDomainAgent(domain)) == NULL) {
+	    lsts = PM_ERR_NOAGENT;
+	}
+	else if (!ap->status.connected) {
+	    lsts = PM_ERR_NOAGENT;
+	}
+	else {
 	    if (ap->ipcType == AGENT_DSO) {
 		if (ap->ipc.dso.dispatch.comm.pmda_interface >= PMDA_INTERFACE_5)
 		    ap->ipc.dso.dispatch.version.four.ext->e_context = cp - client;
@@ -501,7 +477,7 @@ DoPMNSNames(ClientInfo *cp, __pmPDU *pb)
 		}
 		else {
 		    /* Not PMDA_INTERFACE_4 or later */
-		    lsts = PM_ERR_NAME;
+		    lsts = PM_ERR_NONLEAF;
 		}
 	    }
 	    else {
@@ -547,28 +523,33 @@ DoPMNSNames(ClientInfo *cp, __pmPDU *pb)
 		    fdfail != -1)
 		    CleanupAgent(ap, AT_COMM, fdfail);
 	    }
-	    /*
-	     * only set error status to the current error status
-	     * if this is the first error for this set of metrics
-	     */
-	    if (lsts < 0 && sts > 0) sts = lsts;
+	}
+	/*
+	 * only set error status to the current error status
+	 * if this is the first error for this set of metrics,
+	 * and if either it is a fatal error, or numids is 1
+	 */
+	if (lsts < 0 && sts > 0) {
+	    if ((lsts != PM_ERR_NAME && lsts != PM_ERR_NOAGENT &&
+		 lsts != PM_ERR_NONLEAF) || numids == 1)
+	    sts = lsts;
 	}
     }
 
-    if (sts < 0) {
-        /* If get an error which should be passed back along
-         * with valid data to the client
-         * then do NOT fail -> return status with the id-list.
-         */
-    	if (sts != PM_ERR_NAME && sts != PM_ERR_NONLEAF && sts != PM_ERR_NOAGENT)
-	  goto done;
+    if (sts < 0)
+	/* fatal error or explicit error in the numids == 1 case */
+	goto done;
+
+    numok = numids;
+    for (i = 0; i < numids; i++) {
+	if (idlist[i] == PM_ID_NULL)
+	    numok--;
     }
 
-    pmcd_trace(TR_XMIT_PDU, cp->fd, PDU_PMNS_IDS, numids);
-    if ((sts = __pmSendIDList(cp->fd, FROM_ANON, numids, idlist, sts)) < 0) {
+    pmcd_trace(TR_XMIT_PDU, cp->fd, PDU_PMNS_IDS, numok);
+    if ((sts = __pmSendIDList(cp->fd, FROM_ANON, numids, idlist, numok)) < 0) {
 	pmcd_trace(TR_XMIT_ERR, cp->fd, PDU_PMNS_IDS, sts);
 	CleanupClient(cp, sts);
-    	goto done;
     }
 
 done:
@@ -598,7 +579,7 @@ DoPMNSChild(ClientInfo *cp, __pmPDU *pb)
 
     namelist[0] = name;
     sts = pmLookupName(1, namelist, idlist);
-    if (sts == 1 && pmid_domain(idlist[0]) == DYNAMIC_PMID && pmid_item(idlist[0]) == 0) {
+    if (sts == 1 && IS_DYNAMIC_ROOT(idlist[0])) {
 	int		domain = pmid_cluster(idlist[0]);
 	AgentInfo	*ap = NULL;
 	if ((ap = FindDomainAgent(domain)) == NULL) {
@@ -712,16 +693,16 @@ static int travNL_i;      /* array index */
 static void
 AddLengths(const char *name)
 {
-  travNL_strlen += strlen(name) + 1;
-  travNL_num++;
+    travNL_strlen += strlen(name) + 1;
+    travNL_num++;
 }
 
 static void
 BuildNameList(const char *name)
 {
-  travNL[travNL_i++] = travNL_ptr;
-  strcpy(travNL_ptr, name);
-  travNL_ptr += strlen(name) + 1;
+    travNL[travNL_i++] = travNL_ptr;
+    strcpy(travNL_ptr, name);
+    travNL_ptr += strlen(name) + 1;
 }
 
 /*
@@ -775,7 +756,7 @@ traverse_dynamic(ClientInfo *cp, char *start, int *num_names, char ***names)
 	sts = pmLookupName(1, namelist, idlist);
 	if (sts < 1)
 	    continue;
-	if (pmid_domain(idlist[0]) == DYNAMIC_PMID && pmid_item(idlist[0]) == 0) {
+	if (IS_DYNAMIC_ROOT(idlist[0])) {
 	    int		domain = pmid_cluster(idlist[0]);
 	    AgentInfo	*ap;
 	    if ((ap = FindDomainAgent(domain)) == NULL)
@@ -1000,6 +981,69 @@ done:
 
 /*************************************************************************/
 
+static int
+GetAttribute(ClientInfo *cp, int code)
+{
+    __pmPDU	*pb;
+    char	*value;
+    int		vlen, attr;
+    int		sts, pinpdu;
+
+    /* Expecting an attribute (code) PDU from the client */
+    pinpdu = sts = __pmGetPDU(cp->fd, LIMIT_SIZE, _pmcd_timeout, &pb);
+    if (sts > 0)
+	pmcd_trace(TR_RECV_PDU, cp->fd, sts, (int)((__psint_t)pb & 0xffffffff));
+    if (sts == PDU_ATTR) {
+	if ((sts = __pmDecodeAttr(pb, &attr, &value, &vlen)) == 0) {
+	    if (code != attr) {	/* unanticipated attribute */
+		sts = PM_ERR_IPC;
+	    } else if ((value = strndup(value, vlen)) == NULL) {
+		sts = -ENOMEM;
+	    } else {	/* stash the attribute for this client */
+		sts = __pmHashAdd(attr, (void *)value, &cp->attrs);
+	    }
+	}
+    } else if (sts > 0) {	/* unexpected PDU type */
+	sts = PM_ERR_IPC;
+    }
+    if (pinpdu > 0)
+	__pmUnpinPDUBuf(pb);
+    return sts;
+}
+
+static int
+ConnectionAttributes(ClientInfo *cp, int flags)
+{
+    int sts;
+
+    if ((sts = __pmSecureServerHandshake(cp->fd, flags, &cp->attrs)) < 0) {
+	if (pmDebug & DBG_TRACE_AUTH)
+	    fprintf(stderr, "DoCreds: __pmSecureServerHandshake gave %d: %s\n",
+		    sts, pmErrStr(sts));
+	return sts;
+    }
+
+    if ((flags & PDU_FLAG_CONTAINER) &&
+	(sts = GetAttribute(cp, PCP_ATTR_CONTAINER)) < 0) {
+	if (pmDebug & DBG_TRACE_ATTR)
+	    fprintf(stderr, "DoCreds: failed GetAttribute container %d: %s\n",
+		    sts, pmErrStr(sts));
+	return sts;
+    }
+
+    return 0;
+}
+
+static int
+CheckCertRequired(ClientInfo *cp)
+{
+    if(__pmServerHasFeature(PM_SERVER_FEATURE_CERT_REQD))
+	if ( !__pmSockAddrIsLoopBack(cp->addr) && !__pmSockAddrIsUnix(cp->addr) )
+	    return 1;
+
+    return 0;
+}
+
 int
 DoCreds(ClientInfo *cp, __pmPDU *pb)
 {
@@ -1018,7 +1062,7 @@ DoCreds(ClientInfo *cp, __pmPDU *pb)
 		flags = vcp->c_flags;
 		version = vcp->c_version;
 #ifdef PCP_DEBUG
-		if (pmDebug & DBG_TRACE_AUTH) {
+		if (pmDebug & DBG_TRACE_ATTR) {
 		    static const struct {
 			int	flag;
 			char	*name;
@@ -1028,7 +1072,8 @@ DoCreds(ClientInfo *cp, __pmPDU *pb)
 			{ PDU_FLAG_AUTH,	"AUTH" },
 			{ PDU_FLAG_CREDS_REQD,	"CREDS_REQD" },
 			{ PDU_FLAG_SECURE_ACK,	"SECURE_ACK" },
-			{ PDU_FLAG_NO_NSS_INIT,	"NO_NSS_INIT" }
+			{ PDU_FLAG_NO_NSS_INIT,	"NO_NSS_INIT" },
+			{ PDU_FLAG_CONTAINER,	"CONTAINER" },
 		    };
 		    int	i;
 		    int	first = 1;
@@ -1061,6 +1106,18 @@ DoCreds(ClientInfo *cp, __pmPDU *pb)
 
     if (sts >= 0 && version)
 	sts = __pmSetVersionIPC(cp->fd, version);
+
+    /*
+     * In normal operation, some of this code is redundant. A 
+     * remote client should error out during initial handshake
+     * if it does not support client certs.
+     *
+     * We still need to check local connections and allow those through
+     * in all cases.
+     */
+    if (CheckCertRequired(cp) && (flags & PDU_FLAG_SECURE) == 0)
+	return PM_ERR_NEEDCLIENTCERT;
+
     if (sts >= 0 && flags) {
 	/*
 	 * new client has arrived; may want encryption, authentication, etc
@@ -1068,30 +1125,25 @@ DoCreds(ClientInfo *cp, __pmPDU *pb)
 	 * on to check access is allowed for the authenticated persona, and
 	 * finally notify any interested PMDAs
 	 */
-	if ((sts = __pmSecureServerHandshake(cp->fd, flags, &cp->attrs)) < 0) {
-#ifdef PCP_DEBUG
-	    if (pmDebug & DBG_TRACE_AUTH)
-		fprintf(stderr, "DoCreds: __pmSecureServerHandshake returns %d: %s\n", sts, pmErrStr(sts));
-#endif
+	if ((sts = ConnectionAttributes(cp, flags)) < 0)
 	    return sts;
-	}
     }
     if ((sts = CheckAccountAccess(cp)) < 0) {	/* host access done already */
-#ifdef PCP_DEBUG
 	if (pmDebug & DBG_TRACE_AUTH)
-	    fprintf(stderr, "DoCreds: CheckAccountAccess returns %d: %s\n", sts, pmErrStr(sts));
-#endif
+	    fprintf(stderr, "DoCreds: CheckAccountAccess returns %d: %s\n",
+		    sts, pmErrStr(sts));
 	return sts;
     }
-    else if (sts > 0)	/* account authentication successful - inform PMDAs */ {
-	sts = AgentsAuthentication(cp - client);
-#ifdef PCP_DEBUG
-	if (sts < 0 && (pmDebug & DBG_TRACE_AUTH))
-	    fprintf(stderr, "DoCreds: AgentsAuthentication returns %d: %s\n", sts, pmErrStr(sts));
-#endif
+    /*
+     * account authentication successful (if needed) and/or other attributes
+     * have been given - in these cases, we need to inform interested PMDAs.
+     */
+    else if (sts > 0 || (flags & PDU_FLAG_CONTAINER)) {
+	sts = AgentsAttributes(cp - client);
+	if (sts < 0 && (pmDebug & (DBG_TRACE_AUTH|DBG_TRACE_ATTR)))
+	    fprintf(stderr, "DoCreds: AgentsAttributes returns %d: %s\n",
+		    sts, pmErrStr(sts));
     }
 
-    /* else: no account-based authentication needed, so finish successfully */
-
     return sts;
 }
