@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2014 Red Hat.
+ * Copyright (c) 2014-2017 Red Hat.
  * Copyright (c) 1995-2001 Silicon Graphics, Inc.  All Rights Reserved.
  * 
  * This program is free software; you can redistribute it and/or modify it
@@ -17,6 +17,10 @@
 
 int	last_log_offset;
 
+#define IS_DERIVED_LOGGED(x) (pmid_domain(x) == DYNAMIC_PMID && (pmid_cluster(x) & 2048) == 2048 && pmid_item(x) != 0)
+#define SET_DERIVED_LOGGED(x) pmid_build(pmid_domain(x), 2048 | pmid_cluster(x), pmid_item(x))
+#define CLEAR_DERIVED_LOGGED(x) pmid_build(pmid_domain(x), ~2048 & pmid_cluster(x), pmid_item(x))
+
 /*
  * pro tem, we have a single context with the pmcd providing the
  * results, hence need to send the profile each time
@@ -312,28 +316,61 @@ lookupTaskCacheIndex(task_t *tp, pmID pmid)
 static int
 lookupTaskCacheNames(pmID pmid, char ***namesptr)
 {
-    int i, numnames = 0, len = 0;
-    char *data, **names = NULL;
-    task_t *tp;
+    int		i;
+    int		j;
+    int		numnames;
+    int		str_len = 0;
+    char	*data;
+    char	**names;
+    task_t	*tp;
+
+    /*
+     * start with names (including duplicate, if any) from the PMNS
+     */
+    numnames = pmNameAll(pmid, &names);
+    if (numnames < 0) {
+	numnames = 0;
+	names = (char **)malloc(0);
+    }
+    for (i = 0; i < numnames; i++) {
+	str_len += strlen(names[i]) + 1;
+    }
 
     for (tp = tasklist; tp != NULL; tp = tp->t_next) {
 	for (i = 0; i < tp->t_numpmid; i++) {
 	    if (tp->t_pmidlist[i] != pmid)
 		continue;
-	    len += strlen(tp->t_namelist[i]) + 1;
-	    numnames++;
-	}
-    }
-
-    names = (char **)malloc(numnames * sizeof(names[0]) + len);
-    data = (char *)names + (numnames * sizeof(names[0]));
-    for (tp = tasklist, len = 0; tp != NULL; tp = tp->t_next) {
-	for (i = 0; i < tp->t_numpmid; i++) {
-	    if (tp->t_pmidlist[i] != pmid)
-		continue;
-	    names[len++] = data;
-	    strcpy(data, tp->t_namelist[i]);
-	    data += (strlen(tp->t_namelist[i]) + 1);
+	    for (j = 0; j < numnames; j++) {
+		if (strcmp(names[j], tp->t_namelist[i]) == 0) {
+		    /* in task list, and already in names[] ... skip */
+		    break;
+		}
+	    }
+	    if (j == numnames) {
+		/*
+		 * need to append this one ... this is rare!
+		 * only known case where this happens is when the
+		 * pmlogger configuration file names a dynamic metric
+		 * that is mapped to an existing PMID which is in
+		 * the PMNS, but the PMID->name service knows nothing
+		 * about the name of the dynamic metric
+		 */
+		int	old_str_len = str_len;
+		str_len += strlen(tp->t_namelist[i]) + 1;
+		numnames++;
+		names = (char **)realloc(names, numnames * sizeof(names[0]) + str_len);
+		data = (char *)names + ((numnames-1) * sizeof(names[0]));
+		/* relocate strings */
+		memmove(data+sizeof(names[0]), data, old_str_len);
+		data += sizeof(names[0]);
+		for (j = 0; j < numnames; j++) {
+		    names[j] = data;
+		    if (j == numnames - 1) {
+			strcpy(data, tp->t_namelist[i]);
+		    }
+		    data += (strlen(names[j]) + 1);
+		}
+	    }
 	}
     }
 
@@ -341,14 +378,32 @@ lookupTaskCacheNames(pmID pmid, char ***namesptr)
     return numnames;
 }
 
+/*
+ * Warning: called in signal handler context ... be careful
+ */
 void
 log_callback(int afid, void *data)
+{
+    task_t		*tp;
+    for (tp = tasklist; tp != NULL; tp = tp->t_next) {
+	if (tp->t_afid == afid) {
+	    tp->t_alarm = 1;
+	    log_alarm = 1;
+	    break;
+	}
+    }
+}
+
+/*
+ * do real work from callback ...
+ */
+void
+do_work(task_t *tp)
 {
     int			i;
     int			j;
     int			k;
     int			sts;
-    task_t		*tp = (task_t *)data;
     fetchctl_t		*fp;
     indomctl_t		*idp;
     pmResult		*resp;
@@ -371,13 +426,23 @@ log_callback(int afid, void *data)
     __pmTimeval		resp_tval;
     unsigned long	peek_offset;
 
+#ifdef PCP_DEBUG
+    if ((pmDebug & DBG_TRACE_APPL2) && (pmDebug & DBG_TRACE_DESPERATE)) {
+	struct timeval	now;
+
+	__pmtimevalNow(&now);
+	__pmPrintStamp(stderr, &now);
+	fprintf(stderr, " do_work(tp=%p): afid=%d parse_done=%d exit_samples=%d\n", tp, tp->t_afid, parse_done, exit_samples);
+    }
+#endif
+
     if (!parse_done)
 	/* ignore callbacks until all of the config file has been parsed */
 	return;
 
     /* find AFctl_t for this afid */
     for (acp = achead; acp != (AFctl_t *)0; acp = acp->ac_next) {
-	if (acp->ac_afid == afid)
+	if (acp->ac_afid == tp->t_afid)
 	    break;
     }
     if (acp == (AFctl_t *)0) {
@@ -386,7 +451,7 @@ log_callback(int afid, void *data)
 	    __pmNoMem("log_callback: new AFctl_t entry calloc",
 		     sizeof(AFctl_t), PM_FATAL_ERR);
 	}
-	acp->ac_afid = afid;
+	acp->ac_afid = tp->t_afid;
 	acp->ac_next = achead;
 	achead = acp;
     }
@@ -446,11 +511,19 @@ log_callback(int afid, void *data)
 	clearavail(fp);
 
 	if ((sts = myFetch(fp->f_numpmid, fp->f_pmidlist, &pb)) < 0) {
+	    if (sts == -EINTR) {
+		/* disconnect() already done in myFetch() */
+		return;
+	    }
+	    if (sts != -ETIMEDOUT) {
+		/* optionally report and disconnect() the first time thru */
 #ifdef PCP_DEBUG
-	    if (pmDebug & DBG_TRACE_APPL2)
-		fprintf(stderr, "callback: disconnecting because myFetch failed: %s\n", pmErrStr(sts));
+		if (pmDebug & DBG_TRACE_APPL2)
+		    fprintf(stderr, "callback: disconnecting because myFetch failed: %s\n", pmErrStr(sts));
 #endif
-	    disconnect(sts);
+		disconnect(sts);
+	    }
+	    continue;
 	}
 #ifdef PCP_DEBUG
 	if (pmDebug & DBG_TRACE_APPL2)
@@ -459,7 +532,7 @@ log_callback(int afid, void *data)
 
 	/*
 	 * hook to rewrite PDU buffer ...
-	lfp */
+	 */
 	pb = rewrite_pdu(pb, archive_version);
 
 	if (rflag) {
@@ -489,18 +562,24 @@ log_callback(int afid, void *data)
 	}
 
 	/*
-	 * would prefer to save this up until after any meta data and/or
+	 * Would prefer to save this up until after any meta data and/or
 	 * temporal index writes, but __pmDecodeResult changes the pointers
 	 * in the pdu buffer for the non INSITU values ... sigh
+	 * Unfortunately, if we have derived metrics we need to rewrite
+	 * the PMIDs, and this can't be done until after the pmResult
+	 * is decoded ... so we have 2 "write" paths for the PDU buffer
+	 * ... more sighing
 	 */
 	last_log_offset = ftell(logctl.l_mfp);
 	assert(last_log_offset >= 0);
-	if ((sts = __pmLogPutResult2(&logctl, pb)) < 0) {
-	    fprintf(stderr, "__pmLogPutResult2: %s\n", pmErrStr(sts));
-	    exit(1);
+	if (tp->t_dm == 0) {
+	    if ((sts = __pmLogPutResult2(&logctl, pb)) < 0) {
+		fprintf(stderr, "__pmLogPutResult2: %s\n", pmErrStr(sts));
+		exit(1);
+	    }
+	    __pmOverrideLastFd(fileno(logctl.l_mfp));
 	}
-
-	__pmOverrideLastFd(fileno(logctl.l_mfp));
+	resp = NULL; /* silence coverity */
 	if ((sts = __pmDecodeResult(pb, &resp)) < 0) {
 	    fprintf(stderr, "__pmDecodeResult: %s\n", pmErrStr(sts));
 	    exit(1);
@@ -509,6 +588,39 @@ log_callback(int afid, void *data)
 	resp_tval.tv_sec = resp->timestamp.tv_sec;
 	resp_tval.tv_usec = resp->timestamp.tv_usec;
 
+	if (tp->t_dm != 0) {
+	    /*
+	     * pmResult contains at least one derived metric, rewrite
+	     * the cluster field of the PMID(s) (set the top bit), then
+	     * then output the buffer then restore the PMID(s).
+	     *
+	     * This forces the PMID in the archive to NOT look like
+	     * the PMID of a derived metric, which is need to replay
+	     * the archive correctly.
+	     */
+	    __pmPDU	*pdubuf;
+	    for (i = 0; i < resp->numpmid; i++) {
+		pmValueSet	*vsp = resp->vset[i];
+		if (IS_DERIVED(vsp->pmid))
+		    vsp->pmid = SET_DERIVED_LOGGED(vsp->pmid);
+	    }
+	    if ((sts = __pmEncodeResult(fileno(logctl.l_mfp), resp, &pdubuf)) < 0) {
+		fprintf(stderr, "__pmEncodeResult: %s\n", pmErrStr(sts));
+		exit(1);
+	    }
+	    if ((sts = __pmLogPutResult2(&logctl, pdubuf)) < 0) {
+		fprintf(stderr, "__pmLogPutResult2: %s\n", pmErrStr(sts));
+		exit(1);
+	    }
+	    __pmUnpinPDUBuf(pdubuf);
+	    __pmOverrideLastFd(fileno(logctl.l_mfp));
+	    for (i = 0; i < resp->numpmid; i++) {
+		pmValueSet	*vsp = resp->vset[i];
+		if (IS_DERIVED_LOGGED(vsp->pmid))
+		    vsp->pmid = CLEAR_DERIVED_LOGGED(vsp->pmid);
+	    }
+	}
+
 	needti = 0;
 	old_meta_offset = ftell(logctl.l_mdfp);
 	assert(old_meta_offset >= 0);
@@ -529,11 +641,21 @@ log_callback(int afid, void *data)
 		}
 		desc = tp->t_desclist[taskindex];
 		numnames = lookupTaskCacheNames(vsp->pmid, &names);
+		if (numnames < 1) {
+		    fprintf(stderr, "lookupTaskCacheNames(%s, ...): %s\n", pmIDStr(vsp->pmid), pmErrStr(sts));
+		    exit(1);
+		}
+		if (IS_DERIVED(desc.pmid))
+		    /* derived metric, rewrite cluster field ... */
+		    desc.pmid = SET_DERIVED_LOGGED(desc.pmid);
 		if ((sts = __pmLogPutDesc(&logctl, &desc, numnames, names)) < 0) {
 		    fprintf(stderr, "__pmLogPutDesc: %s\n", pmErrStr(sts));
 		    exit(1);
 		}
-		if (numnames) {
+		if (IS_DERIVED_LOGGED(desc.pmid))
+		    /* derived metric, restore cluster field ... */
+		    desc.pmid = CLEAR_DERIVED_LOGGED(desc.pmid);
+		if (numnames > 0) {
 		    free(names);
 		}
 	    }
@@ -586,7 +708,8 @@ log_callback(int afid, void *data)
 		if (needindom) {
 		    /*
 		     * Note.  We do NOT free() instlist and namelist allocated
-		     *	  here ... look for magic below log{Put,Get}InDom ...
+		     *	  here unless this indom is a duplicate.
+		     *    Look for magic below log{Put,Get}InDom.
 		     */
 		    if ((numinst = pmGetInDom(desc.indom, &instlist, &namelist)) < 0) {
 			fprintf(stderr, "pmGetInDom(%s): %s\n", pmInDomStr(desc.indom), pmErrStr(numinst));
@@ -598,6 +721,10 @@ log_callback(int afid, void *data)
 			fprintf(stderr, "__pmLogPutInDom: %s\n", pmErrStr(sts));
 			exit(1);
 		    }
+		    if (sts == PMLOGPUTINDOM_DUP) {
+			free(instlist);
+			free(namelist);
+		    }
 		    needti = 1;
 #ifdef PCP_DEBUG
 		    if (pmDebug & DBG_TRACE_APPL2)
@@ -668,8 +795,12 @@ log_callback(int afid, void *data)
 
 	tp->t_size = pdu_bytes;
 
-	if (pdu_metrics > 1)
-	    fprintf(stderr, "\nGroup [%d metrics] {\n", pdu_metrics);
+	if (pdu_metrics > 1) {
+	    fprintf(stderr, "\nGroup [%d metrics", pdu_metrics);
+	    if (tp->t_dm > 0)
+		fprintf(stderr, ", %d derived", tp->t_dm);
+	    fprintf(stderr, "] {\n");
+	}
 	else
 	    fprintf(stderr, "\nMetric ");
         
@@ -685,9 +816,9 @@ log_callback(int afid, void *data)
                 if (pdu_metrics > 1)
                     fprintf(stderr, "\n");
             } 
-            if (pdu_metrics > 1)
-                fprintf(stderr, "}");
         }
+	if (pdu_metrics > 1)
+	    fprintf(stderr, "}");
 	fprintf(stderr, " logged ");
 
 	if (tp->t_delta.tv_sec == 0 && tp->t_delta.tv_usec == 0)
