@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2012-2014 Red Hat.
+ * Copyright (c) 2012-2016 Red Hat.
  * Copyright (c) 1995-2001,2003 Silicon Graphics, Inc.  All Rights Reserved.
  * 
  * This program is free software; you can redistribute it and/or modify it
@@ -17,8 +17,9 @@
 #include <limits.h>
 #include <sys/stat.h>
 #include "logger.h"
+#include <errno.h>
 
-char		*configfile;
+char		*configfile;		/* current config filename, must be *alloc()d */
 __pmLogCtl	logctl;
 int		exit_samples = -1;       /* number of samples 'til exit */
 __int64_t	exit_bytes = -1;         /* number of bytes 'til exit */
@@ -30,21 +31,26 @@ struct timeval	vol_switch_time;         /* time interval 'til vol switch */
 int		vol_samples_counter;     /* Counts samples - reset for new vol*/
 int		vol_switch_afid = -1;    /* afid of event for vol switch */
 int		vol_switch_flag;         /* sighup received - switch vol now */
+int		vol_switch_alarm;	 /* vol_switch_callback() called */
+int		run_done_alarm;		 /* run_done_callback() called */
+int		log_alarm;	 	 /* log_callback() called */
 int		parse_done;
 int		primary;		/* Non-zero for primary pmlogger */
 char	    	*archBase;		/* base name for log files */
 char		*pmcd_host;
 char		*pmcd_host_conn;
-struct timeval	epoch;
+int		host_context = PM_CONTEXT_HOST;	 /* pmcd / local context mode */
 int		archive_version = PM_LOG_VERS02; /* Type of archive to create */
-int		linger;			/* linger with no tasks/events */
+int		linger = 0;		/* linger with no tasks/events */
 int		rflag;			/* report sizes */
+int		Cflag;			/* parse config and exit */
+struct timeval	epoch;
 struct timeval	delta = { 60, 0 };	/* default logging interval */
 int		exit_code;		/* code to pass to exit (zero/signum) */
 int		qa_case;		/* QA error injection state */
 char		*note;			/* note for port map file */
 
-static int 	    pmcdfd;		/* comms to pmcd */
+static int 	    pmcdfd = -1;	/* comms to pmcd */
 static __pmFdSet    fds;		/* file descriptors mask for select */
 static int	    numfds;		/* number of file descriptors in mask */
 
@@ -54,6 +60,7 @@ static time_t	rsc_start;
 static char	*rsc_prog = "<unknown>";
 static char	*folio_name = "<unknown>";
 static char	*dialog_title = "PCP Archive Recording Session";
+static int	sep;
 
 void
 run_done(int sts, char *msg)
@@ -81,16 +88,22 @@ run_done(int sts, char *msg)
     exit(sts);
 }
 
+/*
+ * Warning: called in signal handler context ... be careful
+ */
 static void
 run_done_callback(int i, void *j)
 {
-    run_done(0, NULL);
+    run_done_alarm = 1;
 }
 
+/*
+ * Warning: called in signal handler context ... be careful
+ */
 static void
 vol_switch_callback(int i, void *j)
 {
-    newvolume(VOL_SW_TIME);
+    vol_switch_alarm = 1;
 }
 
 static int
@@ -311,9 +324,12 @@ do_dialog(char cmd)
 
     snprintf(lbuf, sizeof(lbuf), "PCP recording for the archive folio \"%s\" and the host", folio_name);
     nchar = add_msg(&p, nchar, lbuf);
-    snprintf(lbuf, sizeof(lbuf), " \"%s\" has been in progress for %ld %s",
-	pmcd_host,
-	now < 240 ? now : now/60, now < 240 ? "seconds" : "minutes");
+    if (now < 240)
+	snprintf(lbuf, sizeof(lbuf), " \"%s\" has been in progress for %ld seconds",
+	pmcd_host, (long)now);
+    else
+	snprintf(lbuf, sizeof(lbuf), " \"%s\" has been in progress for %ld minutes",
+	pmcd_host, (long)((now + 30)/60));
     nchar = add_msg(&p, nchar, lbuf);
     nchar = add_msg(&p, nchar, " and in that time the pmlogger process has created an");
     nchar = add_msg(&p, nchar, " archive of ");
@@ -435,6 +451,7 @@ failed:
 
     if (cmd == 'Q' || (cmd == 'X' && strcmp(lbuf, "Yes") == 0)) {
 	run_done(0, "Recording session terminated");
+	/*NOTREACHED*/
     }
 
     if (cmd != '?') {
@@ -448,12 +465,16 @@ failed:
 static pmLongOptions longopts[] = {
     PMAPI_OPTIONS_HEADER("Options"),
     { "config", 1, 'c', "FILE", "file to load configuration from" },
+    { "check", 0, 'C', 0, "parse configuration and exit" },
     PMOPT_DEBUG,
     PMOPT_HOST,
     { "log", 1, 'l', "FILE", "redirect diagnostics and trace output" },
     { "linger", 0, 'L', 0, "run even if not primary logger instance and nothing to log" },
     { "note", 1, 'm', "MSG", "descriptive note to be added to the port map file" },
+    PMOPT_SPECLOCAL,
+    { "local-PMDA", 0, 'o', 0, "metrics sourced without connecting to pmcd" },
     PMOPT_NAMESPACE,
+    { "PID", 1, 'p', "PID", "Log specified metric for the lifetime of the pid" },
     { "primary", 0, 'P', 0, "execute as primary logger instance" },
     { "report", 0, 'r', 0, "report record sizes and archive growth rate" },
     { "size", 1, 's', "SIZE", "terminate after endsize has been accumulated" },
@@ -470,17 +491,57 @@ static pmLongOptions longopts[] = {
 };
 
 static pmOptions opts = {
-    .short_options = "c:D:h:l:Lm:n:Prs:T:t:uU:v:V:x:y?",
+    .short_options = "c:CD:h:l:K:Lm:n:op:Prs:T:t:uU:v:V:x:y?",
     .long_options = longopts,
     .short_usage = "[options] archive",
 };
 
+static FILE *
+do_pmcpp(char *configfile)
+{
+    FILE	*f;
+    char	cmd[3*MAXPATHLEN+80];
+    char	*bin_dir = pmGetConfig("PCP_BINADM_DIR");
+    char	*lib_dir = pmGetConfig("PCP_VAR_DIR");
+
+    if (configfile != NULL) {
+	if ((f = fopen(configfile, "r")) == NULL) {
+	    fprintf(stderr, "%s: Cannot open config file \"%s\": %s\n",
+		pmProgname, configfile, osstrerror());
+	    exit(1);
+	}
+	fclose(f);
+    }
+
+    if (bin_dir == NULL) {
+	fprintf(stderr, "%s: pmGetConfig: cannot get $PCP_BINADM_DIR value\n",
+		pmProgname);
+	exit(1);
+    }
+    if (lib_dir == NULL) {
+	fprintf(stderr, "%s: pmGetConfig: cannot get $PCP_VAR_DIR value\n",
+		pmProgname);
+	exit(1);
+    }
+
+    snprintf(cmd, sizeof(cmd), "%s%cpmcpp -rs %s -I %s%cconfig%cpmlogger",
+	bin_dir, sep, configfile == NULL ? "" : configfile, lib_dir, sep, sep);
+    fprintf(stderr, "preprocessor cmd: %s\n", cmd);
+
+    if ((f = popen(cmd, "r")) == NULL) {
+	fprintf(stderr, "%s: popen(\"%s\", \"r\") failed: %s\n",
+		pmProgname, cmd, osstrerror());
+	exit(1);
+    }
+
+    return f;
+}
+
 int
 main(int argc, char **argv)
 {
     int			c;
     int			sts;
-    int			sep = __pmPathSeparator();
     int			use_localtime = 0;
     int			isdaemon = 0;
     char		*pmnsfile = PM_NS_DEFAULT;
@@ -496,8 +557,13 @@ main(int argc, char **argv)
     char		*runtime = NULL;
     int	    		ctx;		/* handle corresponding to ctxp below */
     __pmContext  	*ctxp;		/* pmlogger has just this one context */
+    int			niter;
+    pid_t               target_pid = 0;
 
     __pmGetUsername(&username);
+    sep = __pmPathSeparator();
+    if ((endnum = getenv("PMLOGGER_INTERVAL")) != NULL)
+	delta.tv_sec = atoi(endnum);
 
     /*
      * Warning:
@@ -510,7 +576,7 @@ main(int argc, char **argv)
 
 	case 'c':		/* config file */
 	    if (access(opts.optarg, F_OK) == 0)
-		configfile = opts.optarg;
+		configfile = strdup(opts.optarg);
 	    else {
 		/* does not exist as given, try the standard place */
 		char *sysconf = pmGetConfig("PCP_VAR_DIR");
@@ -523,11 +589,15 @@ main(int argc, char **argv)
 		if (access(configfile, F_OK) != 0) {
 		    /* still no good, error handling happens below */
 		    free(configfile);
-		    configfile = opts.optarg;
+		    configfile = strdup(opts.optarg);
 		}
 	    }
 	    break;
 
+	case 'C':		/* parse config and exit */
+	    Cflag = 1;
+	    break;
+
 	case 'D':	/* debug flag */
 	    sts = __pmParseDebug(opts.optarg);
 	    if (sts < 0) {
@@ -547,6 +617,14 @@ main(int argc, char **argv)
 	    logfile = opts.optarg;
 	    break;
 
+	case 'K':
+	    if ((endnum = __pmSpecLocalPMDA(opts.optarg)) != NULL) {
+		pmprintf("%s: __pmSpecLocalPMDA failed: %s\n",
+			pmProgname, endnum);
+		opts.errors++;
+	    }
+	    break;
+
 	case 'L':		/* linger if not primary logger */
 	    linger = 1;
 	    break;
@@ -561,6 +639,30 @@ main(int argc, char **argv)
 	    pmnsfile = opts.optarg;
 	    break;
 
+	case 'o':		/* local context mode, no pmcd */
+	    /*
+	     * Note, using Lflag here because this has the same
+	     * semantics as -L for all the other PCP commands, but
+	     * -L was already taken (for "linger") in pmlogger, so
+	     * we're forced to use -o on the command line.
+	     */
+	    host_context = PM_CONTEXT_LOCAL;
+	    opts.Lflag = 1;
+	    break;
+
+	case 'p':
+	    target_pid = (int)strtol(opts.optarg, &endnum, 10);
+	    if (*endnum != '\0') {
+		pmprintf("%s: invalid process identifier (%s)\n",
+			 pmProgname, opts.optarg);
+		opts.errors++;
+	    } else if (!__pmProcessExists(target_pid)) {
+		pmprintf("%s: PID error - no such process (%d)\n",
+			 pmProgname, target_pid);
+		opts.errors++;
+	    }
+	    break;
+
 	case 'P':		/* this is the primary pmlogger */
 	    primary = 1;
 	    isdaemon = 1;
@@ -651,7 +753,7 @@ main(int argc, char **argv)
 	}
     }
 
-    if (primary && pmcd_host != NULL) {
+    if (pmcd_host_conn != NULL && primary) {
 	pmprintf(
 	    "%s: -P and -h are mutually exclusive; use -P only when running\n"
 	    "%s on the same (local) host as the PMCD to which it connects.\n",
@@ -659,6 +761,14 @@ main(int argc, char **argv)
 	opts.errors++;
     }
 
+    if (pmcd_host_conn != NULL && host_context == PM_CONTEXT_LOCAL) {
+	pmprintf(
+	    "%s: -o and -h are mutually exclusive; use -o only when running\n"
+	    "%s on the same (local) host as the DSO PMDA(s) being used.\n",
+		pmProgname, pmProgname);
+	opts.errors++;
+    }
+
     if (!opts.errors && opts.optind != argc - 1) {
 	pmprintf("%s: insufficient arguments\n", pmProgname);
 	opts.errors++;
@@ -680,18 +790,17 @@ main(int argc, char **argv)
     if (isdaemon)
 	__pmSetProcessIdentity(username);
 
-    __pmOpenLog("pmlogger", logfile, stderr, &sts);
-    if (sts != 1) {
-	fprintf(stderr, "%s: Warning: log file (%s) creation failed\n", pmProgname, logfile);
-	/* continue on ... writing to stderr */
+    if (Cflag == 0) {
+	__pmOpenLog("pmlogger", logfile, stderr, &sts);
+	if (sts != 1) {
+	    fprintf(stderr, "%s: Warning: log file (%s) creation failed\n", pmProgname, logfile);
+	    /* continue on ... writing to stderr */
+	}
     }
 
     /* base name for archive is here ... */
     archBase = argv[opts.optind];
 
-    if (pmcd_host_conn == NULL)
-	pmcd_host_conn = "local:";
-
     /* initialise access control */
     if (__pmAccAddOp(PM_OP_LOG_ADV) < 0 ||
 	__pmAccAddOp(PM_OP_LOG_MAND) < 0 ||
@@ -701,13 +810,18 @@ main(int argc, char **argv)
     }
 
     if (pmnsfile != PM_NS_DEFAULT) {
-	if ((sts = pmLoadNameSpace(pmnsfile)) < 0) {
+	if ((sts = pmLoadASCIINameSpace(pmnsfile, 1)) < 0) {
 	    fprintf(stderr, "%s: Cannot load namespace from \"%s\": %s\n", pmProgname, pmnsfile, pmErrStr(sts));
 	    exit(1);
 	}
     }
 
-    if ((ctx = pmNewContext(PM_CONTEXT_HOST, pmcd_host_conn)) < 0) {
+    if (host_context == PM_CONTEXT_LOCAL)
+	pmcd_host_conn = "local context";
+    else if (pmcd_host_conn == NULL)
+	pmcd_host_conn = "local:";
+
+    if ((ctx = pmNewContext(host_context, pmcd_host_conn)) < 0) {
 	fprintf(stderr, "%s: Cannot connect to PMCD on host \"%s\": %s\n", pmProgname, pmcd_host_conn, pmErrStr(ctx));
 	exit(1);
     }
@@ -718,7 +832,7 @@ main(int argc, char **argv)
 	exit(1);
     }
 
-    if (rsc_fd == -1) {
+    if (rsc_fd == -1 && host_context != PM_CONTEXT_LOCAL) {
 	/* no -x, so register client id with pmcd */
 	__pmSetClientIdArgv(argc, argv);
     }
@@ -726,24 +840,19 @@ main(int argc, char **argv)
     /*
      * discover fd for comms channel to PMCD ... 
      */
-    if ((ctxp = __pmHandleToPtr(ctx)) == NULL) {
-	fprintf(stderr, "%s: botch: __pmHandleToPtr(%d) returns NULL!\n", pmProgname, ctx);
-	exit(1);
-    }
-    pmcdfd = ctxp->c_pmcd->pc_fd;
-    PM_UNLOCK(ctxp->c_lock);
-
-    if (configfile != NULL) {
-	if ((yyin = fopen(configfile, "r")) == NULL) {
-	    fprintf(stderr, "%s: Cannot open config file \"%s\": %s\n",
-		pmProgname, configfile, osstrerror());
+    if (host_context != PM_CONTEXT_LOCAL) {
+	if ((ctxp = __pmHandleToPtr(ctx)) == NULL) {
+	    fprintf(stderr, "%s: botch: __pmHandleToPtr(%d) returns NULL!\n", pmProgname, ctx);
 	    exit(1);
 	}
+	pmcdfd = ctxp->c_pmcd->pc_fd;
+	PM_UNLOCK(ctxp->c_lock);
     }
-    else {
-	/* **ANY** Lex would read from stdin automagically */
-	configfile = "<stdin>";
-    }
+
+    yyin = do_pmcpp(configfile);
+    /* do not return unless yyin is valid */
+    if (configfile == NULL)
+	configfile = strdup("<stdin>");
 
     __pmOptFetchGetParams(&ocp);
     ocp.c_scope = 1;
@@ -754,17 +863,13 @@ main(int argc, char **argv)
 
     if (yyparse() != 0)
 	exit(1);
-    if (configfile != NULL)
-	fclose(yyin);
+    fclose(yyin);
     yyend();
 
 #ifdef PCP_DEBUG
     fprintf(stderr, "Config parsed\n");
 #endif
 
-    fprintf(stderr, "Starting %slogger for host \"%s\" via \"%s\"\n",
-            primary ? "primary " : "", pmcd_host, pmcd_host_conn);
-
 #ifdef PCP_DEBUG
     if (pmDebug & DBG_TRACE_LOG) {
 	fprintf(stderr, "optFetch Cost Parameters: pmid=%d indom=%d fetch=%d scope=%d\n",
@@ -780,7 +885,7 @@ main(int argc, char **argv)
 		PMLC_GET_MAND(tp->t_state) ? "mand" : "adv",
 		PMLC_GET_ON(tp->t_state) ? "on" : "off");
 	    fprintf(stderr, " delta: %ld usec", 
-			(long)1000 * tp->t_delta.tv_sec + tp->t_delta.tv_usec);
+			(long)(1000 * tp->t_delta.tv_sec + tp->t_delta.tv_usec));
 	    fprintf(stderr, " numpmid: %d\n", tp->t_numpmid);
 	    for (i = 0; i < tp->t_numpmid; i++) {
 		fprintf(stderr, "  %s (%s):\n", pmIDStr(tp->t_pmidlist[i]), tp->t_namelist[i]);
@@ -790,6 +895,12 @@ main(int argc, char **argv)
     }
 #endif
 
+    if (Cflag)
+	exit(0);
+
+    fprintf(stderr, "Starting %slogger for host \"%s\" via \"%s\"\n",
+            primary ? "primary " : "", pmcd_host, pmcd_host_conn);
+
     if (!primary && tasklist == NULL && !linger) {
 	fprintf(stderr, "Nothing to log, and not the primary logger instance ... good-bye\n");
 	exit(1);
@@ -880,7 +991,8 @@ main(int argc, char **argv)
 	    __pmFD_SET(ctlfds[i], &fds);
     }
 #ifndef IS_MINGW
-    __pmFD_SET(pmcdfd, &fds);
+    if (pmcdfd != -1)
+	__pmFD_SET(pmcdfd, &fds);
 #endif
     if (rsc_fd != -1)
 	__pmFD_SET(rsc_fd, &fds);
@@ -897,10 +1009,59 @@ main(int argc, char **argv)
 
     for ( ; ; ) {
 	int		nready;
+
+#ifdef PCP_DEBUG
+	if ((pmDebug & DBG_TRACE_APPL2) && (pmDebug & DBG_TRACE_DESPERATE)) {
+	    fprintf(stderr, "before __pmSelectRead(%d,...): run_done_alarm=%d vol_switch_alarm=%d log_alarm=%d\n", numfds, run_done_alarm, vol_switch_alarm, log_alarm);
+	}
+#endif
+
+	niter = 0;
+	while (log_alarm && niter++ < 10) {
+	    __pmAFblock();
+	    log_alarm = 0;
+#ifdef PCP_DEBUG
+	    if (pmDebug & DBG_TRACE_APPL2)
+		fprintf(stderr, "delayed callback: log_alarm\n");
+#endif
+	    for (tp = tasklist; tp != NULL; tp = tp->t_next) {
+		if (tp->t_alarm) {
+		    tp->t_alarm = 0;
+		    do_work(tp);
+		}
+	    }
+	    __pmAFunblock();
+	}
+
+	if (vol_switch_alarm) {
+	    __pmAFblock();
+	    vol_switch_alarm = 0;
+#ifdef PCP_DEBUG
+	    if (pmDebug & DBG_TRACE_APPL2)
+		fprintf(stderr, "delayed callback: vol_switch_alarm\n");
+#endif
+	    newvolume(VOL_SW_TIME);
+	    __pmAFunblock();
+	}
+
+	if (run_done_alarm) {
+#ifdef PCP_DEBUG
+	    if (pmDebug & DBG_TRACE_APPL2)
+		fprintf(stderr, "delayed callback: run_done_alarm\n");
+#endif
+	    run_done(0, NULL);
+	    /*NOTREACHED*/
+	}
+
 	__pmFD_COPY(&readyfds, &fds);
 	nready = __pmSelectRead(numfds, &readyfds, NULL);
 
-	/* block signals to simplify IO handling */
+#ifdef PCP_DEBUG
+	if ((pmDebug & DBG_TRACE_APPL2) && (pmDebug & DBG_TRACE_DESPERATE)) {
+	    fprintf(stderr, "__pmSelectRead(%d,...) done: nready=%d run_done_alarm=%d vol_switch_alarm=%d log_alarm=%d\n", numfds, nready, run_done_alarm, vol_switch_alarm, log_alarm);
+	}
+#endif
+
 	__pmAFblock();
 	if (nready > 0) {
 
@@ -1041,6 +1202,9 @@ main(int argc, char **argv)
 
 	__pmAFunblock();
 
+	if (target_pid && !__pmProcessExists(target_pid))
+	    exit(EXIT_SUCCESS);
+
 	if (exit_code)
 	    break;
     }
@@ -1094,14 +1258,16 @@ newvolume(int vol_switch_type)
 	    __pmLogWriteLabel(logctl.l_mfp, &logctl.l_label);
 	    logctl.l_state = PM_LOG_STATE_INIT;
 	}
-#if 0
-	if (last_stamp.tv_sec != 0) {
-	    __pmTimeval	tmp;
-	    tmp.tv_sec = (__int32_t)last_stamp.tv_sec;
-	    tmp.tv_usec = (__int32_t)last_stamp.tv_usec;
-	    __pmLogPutIndex(&logctl, &tmp);
-	}
-#endif
+
+	/*
+	 * Note:
+	 * 	No need to update the temporal index here.
+	 *	The new volume is empty (other than the label record)
+	 *	at this stage, but when the first data record is
+	 *	added we update the temporal index at that time ...
+	 *	this happens in do_work() over in callback.c
+	 */
+
 	fclose(logctl.l_mfp);
 	logctl.l_mfp = newfp;
 	logctl.l_label.ill_vol = logctl.l_curvol = nextvol;
@@ -1119,63 +1285,101 @@ void
 disconnect(int sts)
 {
     time_t  		now;
-#if CAN_RECONNECT
     int			ctx;
     __pmContext		*ctxp;
-#endif
 
-    time(&now);
-    if (sts != 0)
-	fprintf(stderr, "%s: Error: %s\n", pmProgname, pmErrStr(sts));
-    fprintf(stderr, "%s: Lost connection to PMCD on \"%s\" at %s",
-	    pmProgname, pmcd_host, ctime(&now));
-#if CAN_RECONNECT
-    if (primary) {
-	fprintf(stderr, "This is fatal for the primary logger.");
-	exit(1);
-    }
-    close(pmcdfd);
-    __pmFD_CLR(pmcdfd, &fds);
-    pmcdfd = -1;
-    numfds = maxfd() + 1;
     if ((ctx = pmWhichContext()) >= 0)
 	ctxp = __pmHandleToPtr(ctx);
     if (ctx < 0 || ctxp == NULL) {
 	fprintf(stderr, "%s: disconnect botch: cannot get context: %s\n", pmProgname, pmErrStr(ctx));
 	exit(1);
     }
-    ctxp->c_pmcd->pc_fd = -1;
+
+    if (pmcdfd != -1) {
+	/*
+	 * only do this once per disconnect event ... because there
+	 * multiple callbacks and PDU interchange attempts after a
+	 * disconnect and before a successful pmReconnectContext(),
+	 * only need to shut down the channel and report once.
+	 */
+	if (sts != -EINTR) {
+	    time(&now);
+	    if (sts != 0)
+		fprintf(stderr, "%s: Error: %s\n", pmProgname, pmErrStr(sts));
+	    fprintf(stderr, "%s: Lost connection to PMCD on \"%s\" at %s",
+		pmProgname, pmcd_host, ctime(&now));
+	}
+	if (pmcdfd != -1) {
+	    close(pmcdfd);
+	    __pmFD_CLR(pmcdfd, &fds);
+	    pmcdfd = -1;
+	}
+	numfds = maxfd() + 1;
+	ctxp->c_pmcd->pc_fd = -1;
+	/*
+	 * since there is only one context in play here, pc_refcnt is not
+	 * really in play, but reset it for consistency
+	 */
+	ctxp->c_pmcd->pc_refcnt = 0;
+    }
+
+    /*
+     * for repeated calls here, c_lock may have been locked again,
+     * so although there is no channel cleanup to be done, release
+     * the lock
+     */
     PM_UNLOCK(ctxp->c_lock);
-#else
-    exit(1);
-#endif
 }
 
-#if CAN_RECONNECT
 int
 reconnect(void)
 {
     int	    		sts;
     int			ctx;
+    time_t		now;
     __pmContext		*ctxp;
 
     if ((ctx = pmWhichContext()) >= 0)
 	ctxp = __pmHandleToPtr(ctx);
     if (ctx < 0 || ctxp == NULL) {
-	fprintf(stderr, "%s: reconnect botch: cannot get context: %s\n", pmProgname, pmErrStr(ctx));
+	fprintf(stderr, "%s: reconnect botch: cannot get context: %s\n",
+		pmProgname, pmErrStr(ctx));
 	exit(1);
     }
     sts = pmReconnectContext(ctx);
     if (sts >= 0) {
-	time_t      now;
-	time(&now);
-	fprintf(stderr, "%s: re-established connection to PMCD on \"%s\" at %s\n",
-		pmProgname, pmcd_host, ctime(&now));
 	pmcdfd = ctxp->c_pmcd->pc_fd;
 	__pmFD_SET(pmcdfd, &fds);
 	numfds = maxfd() + 1;
     }
     PM_UNLOCK(ctxp->c_lock);
-    return sts;
+    if (sts < 0)
+	return sts;
+
+    time(&now);
+    fprintf(stderr, "%s: re-established connection to PMCD on \"%s\" at %s",
+	    pmProgname, pmcd_host, ctime(&now));
+
+    /*
+     * Metrics may have changed while PMCD was unreachable, so we
+     * need to recheck each metric to make sure that its PMID and
+     * semantics have not changed.  We cannot recover if there is
+     * an incompatible change - must defer to controlling scripts
+     * or processes (a new-named archive will have to be created,
+     * from a new pmlogger process, and pmlogrewrite/pmlogextract
+     * will need to become involved if they need to be merged).
+     */
+    validate_metrics();
+
+    /*
+     * All metrics have been validated, however, this state change
+     * represents a potential gap in the stream of metrics.  So we
+     * must store a <mark> record at this point.
+     */
+     if ((sts = putmark()) < 0) {
+	fprintf(stderr, "putmark: %s\n", pmErrStr(sts));
+	exit(1);
+    }
+
+    return 0;
 }
-#endif
