@@ -1,7 +1,7 @@
 # pylint: disable=C0103
 """ Wrapper module for LIBPCP - the core Performace Co-Pilot API """
 #
-# Copyright (C) 2012-2015 Red Hat
+# Copyright (C) 2012-2017 Red Hat
 # Copyright (C) 2009-2012 Michael T. Werner
 #
 # This file is part of the "pcp" module, the python interfaces for the
@@ -21,7 +21,7 @@
 # Additional Information:
 #
 # Performance Co-Pilot Web Site
-# http://www.performancecopilot.org
+# http://www.pcp.io
 #
 # Performance Co-Pilot Programmer's Guide
 # cf. Chapter 3. PMAPI - The Performance Metrics API
@@ -67,10 +67,33 @@
                 print "load average 1=",atom.f
             elif results.contents.get_inst(i, j) == inst5:
                 print "load average 5=",atom.f
+
+
+    # ... or, using the fetchgroup interface:
+
+    from pcp import pmapi
+    import cpmapi as c_api
+
+    pmfg = pmapi.fetchgroup(c_api.PM_CONTEXT_HOST, "local:")
+    v = pmfg.extend_item("hinv.ncpu")
+    vv = pmfg.extend_indom("kernel.all.load", c_api.PM_TYPE_FLOAT)
+    vvv = pmfg.extend_event("systemd.journal.records", field="systemd.journal.field.string")
+    t = pmfg.extend_timestamp()
+
+    pmfg.fetch()
+    print("time: %s" % t())
+    print("number of cpus: %d" % v())
+    for icode, iname, value in vv():
+        print("load average %s: %f" % (iname, value()))
+    for ts, line in vvv():
+        print("%s : %s" % (ts, line()))
+
 """
 
-# for reporting on times from pmLocaltime function
-from time import mktime
+import sys
+import time
+import errno
+import datetime
 
 # constants adapted from C header file <pcp/pmapi.h>
 import cpmapi as c_api
@@ -78,7 +101,7 @@ import cpmapi as c_api
 # for interfacing with LIBPCP - the client-side C API
 import ctypes
 from ctypes import c_char, c_int, c_uint, c_long, c_char_p, c_void_p
-from ctypes import c_longlong, c_ulonglong, c_float, c_double
+from ctypes import c_float, c_double, c_int32, c_uint32, c_int64, c_uint64
 from ctypes import CDLL, POINTER, CFUNCTYPE, Structure, Union
 from ctypes import addressof, pointer, sizeof, cast, byref
 from ctypes import create_string_buffer, memmove
@@ -89,16 +112,15 @@ from ctypes.util import find_library
 #
 # dynamic library loads
 #
-
 LIBPCP = CDLL(find_library("pcp"))
-LIBC = CDLL(find_library("c"))
+libc_name = "c" if sys.platform != "win32" else "msvcrt"
+LIBC = CDLL(find_library(libc_name))
 
 
 ##############################################################################
 #
 # python version information and compatibility
 #
-import sys
 
 if sys.version >= '3':
     integer_types = (int,)
@@ -124,10 +146,14 @@ def pyFileToCFile(fileObj):
 #
 
 class pmErr(Exception):
+    def __init__(self, *args):
+        super(pmErr, self).__init__(*args)
+        self.errno = args[0]
+
     def __str__(self):
         errSym = None
         try:
-            errSym = c_api.pmErrSymDict[self.args[0]]
+            errSym = c_api.pmErrSymDict[self.errno]
         except KeyError:
             pass
         if errSym == None:
@@ -136,10 +162,11 @@ class pmErr(Exception):
 
     def message(self):
         errStr = create_string_buffer(c_api.PM_MAXERRMSGLEN)
-        errStr = LIBPCP.pmErrStr_r(self.args[0], errStr, c_api.PM_MAXERRMSGLEN)
+        errStr = LIBPCP.pmErrStr_r(self.errno, errStr, c_api.PM_MAXERRMSGLEN)
+        result = str(errStr.decode())
         for index in range(1, len(self.args)):
-            errStr += b" " + str(self.args[index]).encode('utf-8')
-        return str(errStr.decode())
+            result += " " + str(self.args[index])
+        return result
 
     def progname(self):
         return str(c_char_p.in_dll(LIBPCP, "pmProgname").value.decode())
@@ -164,10 +191,28 @@ class pmUsageErr(Exception):
 # Section 3.5 - Performance Metric Values
 #
 
-# these hardcoded decls should be derived from <sys/time.h>
+# These hardcoded decls should be derived from <sys/time.h>, but no such
+# ctypes facility exists.  Particularly problematic is the tv_usec field
+# (which POSIX defines as having type suseconds_t) - this can be 32 bits
+# on some 64 bit platforms (hence c_long is not always correct).
+
+if c_api.PM_SIZEOF_SUSECONDS_T == 4:
+    c_suseconds_t = c_int32
+elif c_api.PM_SIZEOF_SUSECONDS_T == 8:
+    c_suseconds_t = c_int64
+else:
+    raise pmErr(c_api.PM_ERR_CONV, "Unexpected suseconds_t size")
+
+if c_api.PM_SIZEOF_TIME_T == 4:
+    c_time_t = c_int32
+elif c_api.PM_SIZEOF_TIME_T == 8:
+    c_time_t = c_int64
+else:
+    raise pmErr(c_api.PM_ERR_CONV, "Unexpected time_t size")
+
 class timeval(Structure):
-    _fields_ = [("tv_sec", c_long),
-                ("tv_usec", c_long)]
+    _fields_ = [("tv_sec", c_time_t),
+                ("tv_usec", c_suseconds_t)]
 
     def __init__(self, sec = 0, usec = 0):
         Structure.__init__(self)
@@ -200,9 +245,21 @@ class timeval(Structure):
 
     def sleep(self):
         """ Delay for the amount of time specified by this timeval. """
-        c_api.pmtimevalSleep(self.tv_sec, self.tv_usec)
+        time.sleep(float(self))
         return None
 
+class timespec(Structure):
+    _fields_ = [("tv_sec", c_time_t),
+                ("tv_nsec", c_long)]
+
+    def __init__(self, sec = 0, nsec = 0):
+        Structure.__init__(self)
+        self.tv_sec = sec
+        self.tv_nsec = nsec
+
+    def __str__(self):
+        return "%.3f" % (float(self.tv_sec)+(float(self.tv_nsec)/1000000000.0))
+
 class tm(Structure):
     _fields_ = [("tm_sec", c_int),
                 ("tm_min", c_int),
@@ -216,26 +273,44 @@ class tm(Structure):
                 ("tm_gmtoff", c_long),	# glibc/bsd extension
                 ("tm_zone", c_char_p)]	# glibc/bsd extension
 
+    def struct_time(self):
+        # convert POSIX representations - see mktime(3) - to python:
+        # https://docs.python.org/3/library/time.html#time.struct_time
+        pywday = self.tm_wday - 1
+        if pywday < 0:
+            pywday = 6
+        stlist = [self.tm_year + 1900, self.tm_mon + 1, self.tm_mday,
+                  self.tm_hour, self.tm_min, self.tm_sec,
+                  pywday, self.tm_yday - 1, self.tm_isdst ]
+        return time.struct_time(stlist)
+
     def __str__(self):
-        timetuple = (self.tm_year+1900, self.tm_mon, self.tm_mday,
-                     self.tm_hour, self.tm_min, self.tm_sec,
-                     self.tm_wday, self.tm_yday, self.tm_isdst)
-        inseconds = 0.0
-        try:
-            inseconds = mktime(timetuple)
-        except:
-            pass
-        return "%s %s" % (inseconds.__str__(), timetuple)
+        # For debugging this, the timetuple is possibly more useful
+        # timetuple = (self.tm_year+1900, self.tm_mon, self.tm_mday,
+        #              self.tm_hour, self.tm_min, self.tm_sec,
+        #              self.tm_wday, self.tm_yday, self.tm_isdst)
+        # return str(timetuple)
+
+        # For regular users manipulating struct tm, pretty-print it
+        result = ctypes.create_string_buffer(32)
+        second = c_api.pmMktime(
+                     self.tm_sec, self.tm_min, self.tm_hour,
+                     self.tm_mday, self.tm_mon, self.tm_year,
+                     self.tm_wday, self.tm_yday, self.tm_isdst,
+                     self.tm_gmtoff, str(self.tm_zone))
+        timetp = c_long(long(second))
+        LIBPCP.pmCtime(byref(timetp), result)
+        return str(result.value.decode()).rstrip()
 
 class pmAtomValue(Union):
     """Union used for unpacking metric values according to type
 
     Constants for specifying metric types are defined in module pmapi
     """
-    _fields_ = [("l", c_int),
-                ("ul", c_uint),
-                ("ll", c_longlong),
-                ("ull", c_ulonglong),
+    _fields_ = [("l", c_int32),
+                ("ul", c_uint32),
+                ("ll", c_int64),
+                ("ull", c_uint64),
                 ("f", c_float),
                 ("d", c_double),
                 ("cp", c_char_p),
@@ -247,9 +322,11 @@ class pmAtomValue(Union):
                   c_api.PM_TYPE_U64 : lambda x: x.ull,
                   c_api.PM_TYPE_FLOAT : lambda x: x.f,
                   c_api.PM_TYPE_DOUBLE : lambda x: x.d,
-                  c_api.PM_TYPE_STRING : lambda x: x.cp,
+                  c_api.PM_TYPE_STRING : lambda x: str(x.cp.decode('utf-8')),
                   c_api.PM_TYPE_AGGREGATE : lambda x: None,
                   c_api.PM_TYPE_AGGREGATE_STATIC : lambda x: None,
+                  c_api.PM_TYPE_EVENT : lambda x: None,
+                  c_api.PM_TYPE_HIGHRES_EVENT : lambda x: None,
                   c_api.PM_TYPE_NOSUPPORT : lambda x: None,
                   c_api.PM_TYPE_UNKNOWN : lambda x: None
             }
@@ -486,6 +563,14 @@ class pmLogLabel(Structure):
                  ("hostname", c_char * c_api.PM_LOG_MAXHOSTLEN),
                  ("tz", c_char * c_api.PM_TZ_MAXLEN) ]
 
+    def get_hostname(self):
+        """ Return the hostname from the structure as native str """
+        return str(self.hostname.decode())
+
+    def get_timezone(self):
+        """ Return the timezone from the structure as native str """
+        return str(self.tz.decode())
+
 
 ##############################################################################
 #
@@ -524,14 +609,12 @@ LIBPCP.pmTraversePMNS.argtypes = [c_char_p, traverseCB_type]
 LIBPCP.pmUnloadNameSpace.restype = c_int
 LIBPCP.pmUnloadNameSpace.argtypes = []
 
-LIBPCP.pmRegisterDerived.restype = c_int
-LIBPCP.pmRegisterDerived.argtypes = [c_char_p, c_char_p]
+LIBPCP.pmRegisterDerivedMetric.restype = c_int
+LIBPCP.pmRegisterDerivedMetric.argtypes = [c_char_p, c_char_p, POINTER(c_char_p)]
 
 LIBPCP.pmLoadDerivedConfig.restype = c_int
 LIBPCP.pmLoadDerivedConfig.argtypes = [c_char_p]
 
-LIBPCP.pmDerivedErrStr.restype = c_char_p
-LIBPCP.pmDerivedErrStr.argtypes = []
 
 ##
 # PMAPI Metrics Description Services
@@ -656,7 +739,6 @@ LIBPCP.pmFetchArchive.argtypes = [POINTER(POINTER(pmResult))]
 ##
 # PMAPI Ancilliary Support Services
 
-
 LIBPCP.pmGetOptionalConfig.restype = c_char_p
 LIBPCP.pmGetOptionalConfig.argtypes = [c_char_p]
 
@@ -690,6 +772,9 @@ LIBPCP.pmTypeStr_r.argtypes = [c_int, c_char_p, c_int]
 LIBPCP.pmAtomStr_r.restype = c_char_p
 LIBPCP.pmAtomStr_r.argtypes = [POINTER(pmAtomValue), c_int, c_char_p, c_int]
 
+LIBPCP.pmSemStr_r.restype = c_char_p
+LIBPCP.pmSemStr_r.argtypes = [c_int, c_char_p, c_int]
+
 LIBPCP.pmPrintValue.restype = None
 LIBPCP.pmPrintValue.argtypes = [c_void_p, c_int, c_int, POINTER(pmValue), c_int]
 
@@ -731,12 +816,17 @@ class pmOptions(object):
         return self._mode
     def _R_delta(self):
         return self._delta
+    def _R_need_reset(self):
+        return self._need_reset
+    def _W_need_reset(self, value):
+        self._need_reset = value
 
     ##
     # property definitions
 
     mode = property(_R_mode, None, None, None)
     delta = property(_R_delta, None, None, None)
+    need_reset = property(_R_need_reset, _W_need_reset, None, None)
 
     ##
     # creation and destruction
@@ -753,9 +843,10 @@ class pmOptions(object):
             c_api.pmSetOptionFlags(c_api.PM_OPTFLAG_BOUNDARIES)
         self._delta = 1			# default archive pmSetMode delta
         self._mode = c_api.PM_MODE_INTERP # default pmSetMode access mode
+        self._need_reset = False	# flag for __del__ memory reclaim
 
     def __del__(self):
-        if c_api:
+        if LIBPCP and self._need_reset != False:
             c_api.pmResetAllOptions()
 
     ##
@@ -764,6 +855,9 @@ class pmOptions(object):
     def pmGetOptionFlags(self):
         return c_api.pmGetOptionFlags()
 
+    def pmSetOptionContext(self, context):
+        return c_api.pmSetOptionContext(context)
+
     def pmSetOptionFlags(self, flags):
         return c_api.pmSetOptionFlags(flags)
 
@@ -788,6 +882,19 @@ class pmOptions(object):
         """ Set sampling interval (pmParseInterval string) """
         return c_api.pmSetOptionInterval(interval)
 
+    def pmGetOperands(self):
+        """ After a pmGetOptions(3) call has been made this method
+            returns a list of any remaining parameters which were
+            not parsed as command line options, aka "operands".
+            http://pubs.opengroup.org/onlinepubs/009695399/basedefs/xbd_chap03.html#tag_03_254
+        """
+        return c_api.pmGetOperands()
+
+    # Deprecated, use pmGetOperands() above instead
+    def pmGetNonOptionsFromList(self, argv):
+        return c_api.pmGetNonOptionsFromList(argv)
+
+    # Deprecated, use pmGetOperands() above instead
     def pmNonOptionsFromList(self, argv):
         return c_api.pmGetNonOptionsFromList(argv)
 
@@ -851,8 +958,8 @@ class pmOptions(object):
         return c_api.pmSetLongOptionDebug()
 
     def pmSetLongOptionGuiMode(self):
-        """ Add support for -g/--guimode into PMAPI monitor tool """
-        return c_api.pmSetLongOptionGuiMode()
+        """ Unimplemented """
+        return None
 
     def pmSetLongOptionHost(self):
         """ Add support for -h/--host into PMAPI monitor tool """
@@ -875,8 +982,8 @@ class pmOptions(object):
         return c_api.pmSetLongOptionOrigin()
 
     def pmSetLongOptionGuiPort(self):
-        """ Add support for -p/--guiport into PMAPI monitor tool """
-        return c_api.pmSetLongOptionGuiPort()
+        """ Unimplemented """
+        return None
 
     def pmSetLongOptionStart(self):
         """ Add support for -S/--start into PMAPI monitor tool """
@@ -918,6 +1025,10 @@ class pmOptions(object):
         """ Add support for --archive-folio into PMAPI monitor tool """
         return c_api.pmSetLongOptionArchiveFolio()
 
+    def pmSetLongOptionContainer(self):
+        """ Add support for --container into PMAPI monitor tool """
+        return c_api.pmSetLongOptionContainer()
+
     def pmSetLongOptionHostList(self):
         """ Add support for --host-list into PMAPI monitor tool """
         return c_api.pmSetLongOptionHostList()
@@ -931,11 +1042,11 @@ class pmOptions(object):
     def pmGetOptionArchives(self):	# str list
         return c_api.pmGetOptionArchives()
 
-    def pmGetOptionAlignment(self):		# timeval
-        sec = c_api.pmGetOptionAlignment_sec()
-        if sec == None:
+    def pmGetOptionAlignment(self):	# timeval
+        alignment = c_api.pmGetOptionAlign_optarg()
+        if alignment == None:
             return None
-        return timeval(sec, c_api.pmGetOptionAlignment_sec())
+        return timeval.fromInterval(alignment)
 
     def pmGetOptionStart(self):		# timeval
         sec = c_api.pmGetOptionStart_sec()
@@ -943,6 +1054,9 @@ class pmOptions(object):
             return None
         return timeval(sec, c_api.pmGetOptionStart_usec())
 
+    def pmGetOptionAlignOptarg(self):	# string
+        return c_api.pmGetOptionAlign_optarg()
+
     def pmGetOptionFinishOptarg(self):	# string
         return c_api.pmGetOptionFinish_optarg()
 
@@ -967,18 +1081,46 @@ class pmOptions(object):
     def pmGetOptionSamples(self):	# int
         return c_api.pmGetOptionSamples()
 
+    def pmGetOptionHostZone(self):	# boolean
+        if c_api.pmGetOptionHostZone() == 0:
+            return False
+        return True
+
     def pmGetOptionTimezone(self):	# str
         return c_api.pmGetOptionTimezone()
 
+    def pmGetOptionContainer(self):	# str
+        return c_api.pmGetOptionContainer()
+
+    def pmGetOptionLocalPMDA(self):        # boolean
+        if c_api.pmGetOptionLocalPMDA() == 0:
+            return False
+        return True
+
+    def pmSetOptionArchive(self, archive):	# str
+        return c_api.pmSetOptionArchive(archive)
+
     def pmSetOptionArchiveList(self, archives):	# str
         return c_api.pmSetOptionArchiveList(archives)
 
     def pmSetOptionArchiveFolio(self, folio):	# str
         return c_api.pmSetOptionArchiveFolio(folio)
 
+    def pmSetOptionContainer(self, container):	# str
+        return c_api.pmSetOptionContainer(container)
+
+    def pmSetOptionHost(self, host):	# str
+        return c_api.pmSetOptionHost(host)
+
     def pmSetOptionHostList(self, hosts):	# str
         return c_api.pmSetOptionHostList(hosts)
 
+    def pmSetOptionSpecLocal(self, spec):        # str
+        return c_api.pmSetOptionSpecLocal(spec)
+
+    def pmSetOptionLocalPMDA(self):
+        return c_api.pmSetOptionLocalPMDA()
+
 
 ##############################################################################
 #
@@ -1035,7 +1177,7 @@ class pmContext(object):
         self._type = typed                              # the context type
         self._target = target                            # the context target
         self._ctx = c_api.PM_ERR_NOCONTEXT                # init'd pre-connect
-        if type(target) != type(b''):
+        if target and type(target) != type(b''):
             source = target.encode('utf-8')
         else:
             source = target
@@ -1064,7 +1206,8 @@ class pmContext(object):
             the typed/index parameters can be used to setup several
             contexts based on the given command line parameters.
         """
-        if (typed <= 0):
+        if typed == None or typed <= 0:
+            options.need_reset = True
             if c_api.pmGetOptionsFromList(argv):
                 raise pmUsageErr
             typed = options.pmGetOptionContext()
@@ -1259,13 +1402,16 @@ class pmContext(object):
             name = name.encode('utf-8')
         if type(expr) != type(b''):
             expr = expr.encode('utf-8')
-        status = LIBPCP.pmRegisterDerived(name, expr)
-        if status != 0:
-            raise pmErr(status)
+        errmsg = c_char_p()
+        result = LIBPCP.pmRegisterDerivedMetric(name, expr, byref(errmsg))
+        if result != 0:
+            text = str(errmsg.value.decode())
+            LIBC.free(errmsg)
+            raise pmErr(c_api.PM_ERR_CONV, text)
         status = LIBPCP.pmReconnectContext(self.ctx)
         if status < 0:
             raise pmErr(status)
-        
+
     def pmLoadDerivedConfig(self, fname):
         """PMAPI - Register derived metric names and definitions from a file
         pm.pmLoadDerivedConfig("FileName")
@@ -1279,14 +1425,18 @@ class pmContext(object):
         if status < 0:
             raise pmErr(status)
 
+    # Deprecated, no longer needed, py wrapper uses pmRegisterDerivedMetric(3)
+    # and the exception encodes a more complete error message as a result.
     @staticmethod
     def pmDerivedErrStr():
-        """PMAPI - Return an error message if the pmRegisterDerived metric
+        """PMAPI - Return an error message if the pmRegisterDerived(3) metric
         definition cannot be parsed
         pm.pmRegisterDerived()
         """
         result = LIBPCP.pmDerivedErrStr()
-        return str(result.decode())
+        if result != None:
+            return str(result.decode())
+        return None
 
     ##
     # PMAPI Metrics Description Services
@@ -1390,7 +1540,7 @@ class pmContext(object):
         if status < 0:
             raise pmErr(status)
         if status > 0:
-            nameL = list(map(lambda x: str(nameA_p[x].decode()), range(status)))
+            nameL = list(map(lambda x: str(nameA_p[x].decode('ascii', 'ignore')), range(status)))
             instL = list(map(lambda x: int(instA_p[x]), range(status)))
             LIBC.free(instA_p)
             LIBC.free(nameA_p)
@@ -1430,7 +1580,7 @@ class pmContext(object):
             raise pmErr(status)
         result = name_p.value
         LIBC.free(name_p)
-        return str(result.decode())
+        return str(result.decode('ascii', 'ignore'))
 
     ##
     # PMAPI Context Services
@@ -1536,7 +1686,10 @@ class pmContext(object):
         status = LIBPCP.pmUseContext(self.ctx)
         if status < 0:
             raise pmErr(status)
-        status = LIBPCP.pmSetMode(mode, pointer(timeVal), delta)
+        when = None
+        if timeVal != None:
+            when = pointer(timeVal)
+        status = LIBPCP.pmSetMode(mode, when, delta)
         if status < 0:
             raise pmErr(status)
         return status
@@ -1709,7 +1862,7 @@ class pmContext(object):
         if status < 0:
             raise pmErr(status)
         if status > 0:
-            nameL = list(map(lambda x: str(nameA_p[x].decode()), range(status)))
+            nameL = list(map(lambda x: str(nameA_p[x].decode('ascii', 'ignore')), range(status)))
             instL = list(map(lambda x: int(instA_p[x]), range(status)))
             LIBC.free(instA_p)
             LIBC.free(nameA_p)
@@ -1748,7 +1901,7 @@ class pmContext(object):
             raise pmErr(status)
         result = name_p.value
         LIBC.free(name_p)
-        return str(result.decode())
+        return str(result.decode('ascii', 'ignore'))
 
     def pmFetchArchive(self):
         """PMAPI - Fetch measurements from the target source
@@ -1780,7 +1933,7 @@ class pmContext(object):
 
     @staticmethod
     def pmErrStr(code):
-        """PMAPI - Return value from environment or pcp config file """
+        """PMAPI - Convert an error code to a readable string  """
         errstr = ctypes.create_string_buffer(c_api.PM_MAXERRMSGLEN)
         result = LIBPCP.pmErrStr_r(code, errstr, c_api.PM_MAXERRMSGLEN)
         return str(result.decode())
@@ -1880,6 +2033,15 @@ class pmContext(object):
         result = LIBPCP.pmAtomStr_r(byref(atom), typed, atomstr, 96)
         return str(result.decode())
 
+    @staticmethod
+    def pmSemStr(sem):
+        """PMAPI - Convert a performance metric semantic to a readable string
+        "string" = pmSemStr(c_api.PM_SEM_COUNTER)
+        """
+        semstr = ctypes.create_string_buffer(32)
+        result = LIBPCP.pmSemStr_r(sem, semstr, 32)
+        return str(result.decode())
+
     @staticmethod
     def pmPrintValue(fileObj, result, ptype, vset_idx, vlist_idx, min_width):
         """PMAPI - Print the value of a metric
@@ -1950,3 +2112,348 @@ class pmContext(object):
         """
         return tvp.sleep()
 
+    @staticmethod
+    def pmProgname():
+        return str(c_char_p.in_dll(LIBPCP, "pmProgname").value.decode())
+
+    @staticmethod
+    def pmDebug(flags):
+        if (c_int.in_dll(LIBPCP, "pmDebug").value & flags):
+            return True
+        return False
+
+
+# ----- fetchgroup API
+
+LIBPCP.pmCreateFetchGroup.restype = c_int
+LIBPCP.pmCreateFetchGroup.argtypes = [POINTER(c_void_p), c_int, c_char_p]
+LIBPCP.pmGetFetchGroupContext.restype = c_int
+LIBPCP.pmGetFetchGroupContext.argtypes = [c_void_p]
+LIBPCP.pmDestroyFetchGroup.restype = c_int
+LIBPCP.pmDestroyFetchGroup.argtypes = [c_void_p]
+LIBPCP.pmExtendFetchGroup_item.restype = c_int
+LIBPCP.pmExtendFetchGroup_item.argtypes = [c_void_p, c_char_p, c_char_p, c_char_p,
+                                           POINTER(pmAtomValue), c_int, POINTER(c_int)]
+LIBPCP.pmExtendFetchGroup_indom.restype = c_int
+LIBPCP.pmExtendFetchGroup_indom.argtypes = [c_void_p, c_char_p, c_char_p,
+                                            POINTER(c_int),
+                                            POINTER(c_char_p),
+                                            POINTER(pmAtomValue),
+                                            c_int,
+                                            POINTER(c_int),
+                                            c_uint,
+                                            POINTER(c_uint),
+                                            POINTER(c_int)]
+LIBPCP.pmExtendFetchGroup_event.restype = c_int
+LIBPCP.pmExtendFetchGroup_event.argtypes = [c_void_p, c_char_p, c_char_p, c_char_p, c_char_p,
+                                            POINTER(timespec),
+                                            POINTER(pmAtomValue),
+                                            c_int,
+                                            POINTER(c_int),
+                                            c_uint,
+                                            POINTER(c_uint),
+                                            POINTER(c_int)]
+LIBPCP.pmExtendFetchGroup_timestamp.restype = c_int
+LIBPCP.pmExtendFetchGroup_timestamp.argtypes = [c_void_p, POINTER(timeval)]
+LIBPCP.pmFetchGroup.restype = c_int
+LIBPCP.pmFetchGroup.argtypes = [c_void_p]
+
+
+class fetchgroup(object):
+    """Defines a PMAPI fetchgroup.
+
+    This class wraps the pmFetchGroup set of C PMAPI functions (q.v.)
+    in an object-oriented manner.
+
+    Each instance of this class represents one fetchgroup, in which
+    interest in several metrics (individual or indoms) is registered.
+    Each registration results in an function-like object that may be
+    called to decode that metric's value(s).  Errors are signalled
+    with exceptions rather than result integers.  Strings are all
+    UTF-8 encoded.
+    """
+
+    class fetchgroup_item(object):
+        """
+        An internal class to receive value/status for a single item.
+        It may be called as if it were a function object to decode
+        the embedded pmAtomValue, which was set at the most recent
+        .fetch() call.
+        """
+
+        def __init__(self, pmtype):
+            """Allocate a single instance to receive a fetchgroup item."""
+            self.sts = c_int(c_api.PM_ERR_VALUE)
+            self.pmtype = pmtype
+            self.value = pmAtomValue()
+
+        def __call__(self):
+            """Retrieve a converted value of a fetchgroup item, if available."""
+            if self.sts.value < 0:
+                raise pmErr(self.sts.value)
+            return self.value.dref(self.pmtype)
+
+
+    class fetchgroup_timestamp(object):
+        """
+        An internal class to receive value for a single timestamp.
+        It may be called as if it were a function object to decode
+        the timestamp, which was set at the most recent
+        .fetch() call, into a datetime object.
+        """
+
+        def __init__(self, ctx):
+            """Allocate a single instance to receive a fetchgroup timestamp."""
+            self.value = timeval()
+            self.ctx = ctx
+
+        def __call__(self):
+            """
+            Retrieve a converted value of a timestamp, if available.  Use
+            pmLocaltime() to convert to a datetime object.
+            """
+            ts = self.ctx.pmLocaltime(self.value.tv_sec)
+            us = int(self.value.tv_usec)
+            dt = datetime.datetime(ts.tm_year+1900, ts.tm_mon+1, ts.tm_mday,
+                                   ts.tm_hour, ts.tm_min, ts.tm_sec, us, None)
+            return dt
+
+
+    class fetchgroup_indom(object):
+        """
+        An internal class to receive value/status for an indom of
+        items.  It may be called as if it were a function object to
+        create an list of tuples containing instance-code/-name/value
+        information.  Each value is a function object that decodes
+        the embedded pmAtomValue, which was set at the most recent
+        fetch() call.
+        """
+
+        def __init__(self, pmtype, num):
+            """Allocate a single instance to receive a fetchgroup item."""
+            stss_t = c_int * num
+            values_t = pmAtomValue * num
+            icodes_t = c_uint * num
+            inames_t = c_char_p * num
+            self.sts = c_int()
+            self.stss = stss_t()
+            self.pmtype = pmtype
+            self.values = values_t()
+            self.icodes = icodes_t()
+            self.inames = inames_t()
+            self.num = c_uint()
+
+        def __call__(self):
+            """Retrieve a converted value of a fetchgroup item, if available."""
+            vv = []
+            if self.sts.value < 0:
+                raise pmErr(self.sts.value)
+            for i in range(self.num.value):
+                def decode_one(self, i):
+                    if self.stss[i] < 0:
+                        raise pmErr(self.stss[i])
+                    return self.values[i].dref(self.pmtype)
+                vv.append((self.icodes[i],
+                           self.inames[i].decode('utf-8') if self.inames[i] else None,
+                           # nested lambda for proper i capture
+                           (lambda i: (lambda: decode_one(self, i)))(i)))
+            return vv
+
+
+    class fetchgroup_event(object):
+        """
+        An internal class to receive value/status for an
+        event record field.  It may be called as if it were a function
+        object to create an list of tuples containing timestamp/value
+        information.  Each value is a function object that decodes
+        the embedded pmAtomValue, which was set at the most recent
+        fetch() call.
+        """
+
+        def __init__(self, pmtype, num, ctx):
+            """Allocate a single instance to receive a fetchgroup item."""
+            stss_t = c_int * num
+            values_t = pmAtomValue * num
+            timespec_t = timespec * num
+            self.sts = c_int()
+            self.stss = stss_t()
+            self.pmtype = pmtype
+            self.times = timespec_t()
+            self.values = values_t()
+            self.num = c_uint()
+            self.ctx = ctx
+
+        def __call__(self):
+            """Retrieve a converted value of a fetchgroup item, if available."""
+            vv = []
+            if self.sts.value < 0:
+                raise pmErr(self.sts.value)
+            # print ([self.values[i].dref(self.pmtype) for i in range(self.num.value)])
+            for i in range(self.num.value):
+                def decode_one(self, i):
+                    if self.stss[i] < 0:
+                        raise pmErr(self.stss[i])
+                    return self.values[i].dref(self.pmtype)
+
+                ts = self.ctx.pmLocaltime(self.times[i].tv_sec)
+                us = int(self.times[i].tv_nsec)//1000
+                dt = datetime.datetime(ts.tm_year+1900, ts.tm_mon+1, ts.tm_mday,
+                                       ts.tm_hour, ts.tm_min, ts.tm_sec, us, None)
+                # nested lambda for proper i capture
+                vv.append((dt,
+                           (lambda i: (lambda: decode_one(self, i)))(i)))
+            return vv
+
+
+    class pmContext_borrowed(pmContext):
+        """
+        An internal class for accessing the private PMAPI context
+        belonging to a fetchgroup.  It works just like a pmContext,
+        except it overrides the constructor/destructor to reflect
+        the "borrowed" state of the context.
+        """
+        def __init__(self, ctx, typed, target):
+            """Override pmContext ctor to eschew pmNewContext."""
+            self._ctx = ctx
+            self._type = typed
+            self._target = target
+            # NB: don't call pmContext.__init__!
+
+        def __del__(self):
+            """Override pmContext ctor to eschew pmDestroyContext."""
+            self._ctx = c_api.PM_ERR_NOCONTEXT
+            
+    def __init__(self, typed=c_api.PM_CONTEXT_HOST, target="local:"):
+        """Create a fetchgroup from a pmContext."""
+
+        self.pmfg = c_void_p()
+        self.items = []
+        if typed == c_api.PM_CONTEXT_LOCAL and target == None:
+            target = "" # Ignored
+        sts = LIBPCP.pmCreateFetchGroup(byref(self.pmfg), typed, target.encode('utf-8'))
+        if sts < 0:
+            raise pmErr(sts)
+        self.ctx = fetchgroup.pmContext_borrowed(LIBPCP.pmGetFetchGroupContext(self.pmfg),
+                                                 typed, target)
+        
+    def __del__(self):
+        """Destroy the fetchgroup.  Drop references to fetchgroup_* items."""
+
+        if LIBPCP != None and self.pmfg.value != None:
+            sts = LIBPCP.pmDestroyFetchGroup(self.pmfg)
+            if sts < 0:
+                raise pmErr(sts)
+        del self.items[:]
+
+    def get_context(self):
+        """
+        Return the private pmContext used by the fetchgroup.
+        WARNING: mutation of this context by other PMAPI functions
+        may disrupt fetchgroup functionality.
+        """
+        
+        return self.ctx
+        
+    def extend_item(self, metric=None, mtype=None, scale=None, instance=None):
+        """Extend the fetchgroup with a single metric.  Infer type if
+        necessary.  Convert scale/rate if appropriate/requested.
+        Requires a specified instance if metric has an instance
+        domain.
+
+        """
+
+        if metric is None:
+            raise pmErr(-errno.EINVAL)
+        if mtype is None: # a special service to dynamically-typed python; not accepted at the C level
+            pmids = self.ctx.pmLookupName(metric)
+            descs = self.ctx.pmLookupDescs(pmids)
+            mtype = descs[0].type
+        v = fetchgroup.fetchgroup_item(mtype)
+        sts = LIBPCP.pmExtendFetchGroup_item(self.pmfg,
+                                             c_char_p(metric.encode('utf-8') if metric else None),
+                                             c_char_p(instance.encode('utf-8') if instance else None),
+                                             c_char_p(scale.encode('utf-8') if scale else None),
+                                             pointer(v.value), c_int(mtype),
+                                             pointer(v.sts))
+        if sts < 0:
+            raise pmErr(sts)
+        self.items.append(v) # remember to keep registered pmAtomValue/etc. alive
+        return v
+
+
+    def extend_indom(self, metric=None, mtype=None, scale=None, maxnum=100):
+        """Extend the fetchgroup with up to @maxnum instances of a metric.
+        (Metrics without instances are also accepted.)  Infer type if
+        necessary.  Convert scale/rate if appropriate/requested.
+
+        """
+
+        if metric is None or maxnum < 0:
+            raise pmErr(-errno.EINVAL)
+        if mtype is None: # a special service to dynamically-typed python; not accepted at the C level
+            pmids = self.ctx.pmLookupName(metric)
+            descs = self.ctx.pmLookupDescs(pmids)
+            mtype = descs[0].type
+        vv = fetchgroup.fetchgroup_indom(mtype, maxnum)
+        sts = LIBPCP.pmExtendFetchGroup_indom(self.pmfg,
+                                              c_char_p(metric.encode('utf-8') if metric else None),
+                                              c_char_p(scale.encode('utf-8') if scale else None),
+                                              cast(pointer(vv.icodes), POINTER(c_int)),
+                                              cast(pointer(vv.inames), POINTER(c_char_p)),
+                                              cast(pointer(vv.values), POINTER(pmAtomValue)),
+                                              c_int(mtype),
+                                              cast(pointer(vv.stss), POINTER(c_int)),
+                                              c_uint(maxnum), pointer(vv.num), pointer(vv.sts))
+        if sts < 0:
+            raise pmErr(sts)
+        self.items.append(vv) # remember to keep registered pmAtomValue/etc. alive
+        return vv
+
+
+    def extend_timestamp(self):
+        """Extend the fetchgroup with a timestamp query. """
+
+        v = fetchgroup.fetchgroup_timestamp(self.ctx)
+        sts = LIBPCP.pmExtendFetchGroup_timestamp(self.pmfg,
+                                                  pointer(v.value))
+        if sts < 0:
+            raise pmErr(sts)
+        self.items.append(v) # remember to keep registered timeval alive
+        return v
+
+
+    def extend_event(self, metric=None, field=None, ftype=None, scale=None, instance=None, maxnum=100):
+        """Extend the fetchgroup with up to @maxnum instances of the given field of
+        the given event metric's records.  Infer type if necessary.  Convert scale
+        if appropriate/requested.
+        """
+
+        if metric is None or maxnum < 0:
+            raise pmErr(-errno.EINVAL)
+        if ftype is None: # a special service to dynamically-typed python; not accepted at the C level
+            pmids = self.ctx.pmLookupName(field)
+            descs = self.ctx.pmLookupDescs(pmids)
+            ftype = descs[0].type
+        vv = fetchgroup.fetchgroup_event(ftype, maxnum, self.ctx)
+        sts = LIBPCP.pmExtendFetchGroup_event(self.pmfg,
+                                              c_char_p(metric.encode('utf-8') if metric else None),
+                                              c_char_p(instance.encode('utf-8') if instance else None),
+                                              c_char_p(field.encode('utf-8') if field else None),
+                                              c_char_p(scale.encode('utf-8') if scale else None),
+                                              cast(pointer(vv.times), POINTER(timespec)),
+                                              cast(pointer(vv.values), POINTER(pmAtomValue)),
+                                              c_int(ftype),
+                                              cast(pointer(vv.stss), POINTER(c_int)),
+                                              c_uint(maxnum), pointer(vv.num), pointer(vv.sts))
+        if sts < 0:
+            raise pmErr(sts)
+        self.items.append(vv) # remember to keep registered pmAtomValue/etc. alive
+        return vv
+
+
+    def fetch(self):
+        """Fetch all the metrics in this fetchgroup and update all values."""
+
+        sts = LIBPCP.pmFetchGroup(self.pmfg)
+        if sts < 0:
+            raise pmErr(sts)
